#include <stdio.h>

/* TREE STRUCT: GENERIC BINARY TREE NODE */
typedef struct treenode {
	struct treenode *left;
	struct treenode *right;
	/* Data is unimportant ... */
} TREENODE;

/* STATIC VARIABLES */
static int list_num;
static TREENODE list_base;
static TREENODE *list_tail;
static TREENODE *root;
static int left;

/*
   list_num  Is simply a counter of how many items have been
             added to the list.
   list_base Is used as a dummy node at the head of the list
             to prevent testing for a null root every time a
             node is added to the list.
   list_tail Is used as a pointer to the last node in the
             linked list.
   root      Is used as a pointer to the node at the head
             of the list.
   left      Is a boolean to keep track of which way we are
             going in the tree so that we divide uneven lists
             correctly. Although it would be more clear to use
             a parameter, the routine is finished with left
             before it recurses (it is no longer using it) and
             making it static reduces the stack size required
             for the recursion.
*/

/* Function prototypes */
void *optree(void *);
static void list_build(TREENODE *);
static TREENODE *form_tree(int);

/* OPTIMIZE A TREE */
void *optree(void *opt_root)
{
	if (opt_root == NULL) return (NULL);

	list_tail = &list_base;
	list_num = 0;

	list_build(opt_root);

	root = list_base.right;

	left = 0;
	return ((void *) form_tree(list_num));
}

/* FORM AN OPTIMIZED TREE FROM LIST */
TREENODE *form_tree(int num);
{
	int middle;
	TREENODE *ptr;

	middle = (num >> 1);

	/* SPECIAL 5-NODE CASE */
	if (num == 5) middle++;

	/* LEAN BRANCH TO CENTER OF TREE */
	if (left) middle = num - middle - 1;

	/* REMOVE LEFT SUBTREE FROM LIST */
	left = 1;
	ptr = (middle > 0) ? form_tree(middle) : NULL;
	root->left = ptr;

	/* REMOVE THIS NODE FROM LIST */
	ptr = root;
	root = root->right;

	/* REMOVE RIGHT SUBTREE FROM LIST */
	left = 0;
	middle = num - middle - 1;
	ptr->right = (middle > 0) ? form_tree(middle) : NULL;
	return ptr;
}

/* FLATTEN TREE INTO LINKED LIST */
void list_build(TREENODE *node)
{
	if (node->left) list_build(node->left);

	list_tail->right = node;
	list_tail = node;
	list_num++;

	if (node->right) list_build(node->right);
}
/* End of File */
