#include <math.h>
#define	ANSI
/*
	Adaptive quadrature routine.
	Called with
		lft:  left endpoint
		rt:   right endpoint
		f:    pointer to function to integrate
		err:  pointer to allowable error
	Returns
		adaptive    evaluated integral
		err         pointer to upper limit of actual error
*/
#if	defined(ANSI)
double adaptive(double lft, double rt, double (*f)(double x), double *err)
#else
double adaptive(lft, rt, f, err)
double lft, rt, (*f)(), *err;
#endif
{
	double h;	/* width of this interval */
	double p1, p2, p3, p4, p5;	/* value of function at the 5 points */
	double error;	/* actual error on this interval */
	double err1, err2;	/* errors on the sub-intervals */
	double int3, int5;	/* integral with Simpson's 3- and 5-point approx. */
	h = rt - lft;	/* get the width of the interval */

	/* get the 2 approximations to the integral */

	p1 = (*f)(lft);
	p2 = (*f)(lft + h/4.0);
	p3 = (*f)(lft + h/2.0);
	p4 = (*f)(rt - h/4.0);
	p5 = (*f)(rt);

	int3 = h*(p1 + 4.0*p3 + p5)/6.0;
	int5 = h*(p1 + 4.0*p2 + 2.0*p3 + 4.0*p4 + p5)/12.0;

	error = fabs(int3 - int5)/15.0;	/* find the actual error */
	if (error < *err) {	/* if within tolerance, */
		*err = error;	/* copy the actual error to caller, */
		return int5;	/* return the more accurate approximation */
	} else {	/* otherwise, */
		err1 = err2 = *err/2.0;	/* cut the allowable error in half, */
		int5 = adaptive(lft, lft+h/2.0, f, &err1);	/* integrate the two */
		int5 += adaptive(lft+h/2.0, rt, f, &err2);	/* half-regions, */
		*err = err1 + err2;	/* sum the real errors, */
		return int5;	/* and return the integral. */
	}
}
