/*
	String handling support by Martin Weitzel.
	Interface macros to "pstr_x" for fixed number of arguments.
	Using these macros instead of direct calls to the function
	helps to make the source more readable.
	Strings that are used only once may be build with help of the
	"tmpstr" function or (preferably) with one of the "TMPSTR" macros.
	The space occupied by a string build in this way is freed with
	the next call to one of these macros, so its content should be
	accessed IMMEDIATLY. To reduce the risk of inappropriate use,
	the result of calls to "tmpstr" should NEVER be stored into a
	pointer, and when calling a function that needs several string
	arguments, AT MOST ONE may be initialized with the result of a
	call to "tmpstr". In both cases this includes indirect calls to
	"tmpstr" through one of the "TMPSTR" macros.
*/
#if	!defined(PSTRING_H)
#define PSTRING_H
/* interface to pstring module */
extern char *pstr_x(const char *str, ...);
extern char *tmpstr(char *str);
/* some macros */
#define pstr_1(a)	(pstr_x((a), (const char *)0))
#define pstr_2(a,b)	(pstr_x((a),(b),(const char *)0))
#define pstr_3(a,b,c)	(pstr_x((a),(b),(c),(const char *)0))
#define pstr_4(a,b,c,d)	(pstr_x((a),(b),(c),(d),(const char *)0))
/* .... may be extended .... */
/* more macros */
#define TMPSTR1(a)	(tmpstr(pstr_1((a))))
#define TMPSTR2(a,b)	(tmpstr(pstr_2((a),(b))))
#define TMPSTR3(a,b,c)	(tmpstr(pstr_3((a),(b),(c))))
#define TMPSTR4(a,b,c,d)	(tmpstr(pstr_4((a),(b),(c),(d))))
/* ....  may be extended ....
#endif
