/*
	dirret.c MS-DOS and UNIX portable directory builder.

		Routines:

			DIRRET *dir_retrieve (path, dir_entries);

			void dir_free (dir_ptr);

		Usage:	dir_retrieve ("/usr/include", &count);

		Possible errno:	ENOENT, ENOTDIR and ENOMEM
*/
/*--------------------------------------------------------------------*/
#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD and SYSV */
#include <stdio.h>	/* stream I/O functions */
#include <errno.h>	/* constants used by functions to set errno */
/*--------------------------------------------------------------------*/
#if	!defined(HC1)	/* not High-C V1.6 */
#include <sys/types.h>	/* defines types used in "sys/stat.h" */
#include <sys/stat.h>	/* defines type returned by stat() */
#endif
/*--------------------------------------------------------------------*/
#if	defined(SYSV) || defined(BSD)	/* UNIX-C */
#include <dirent.h>	/* directory header */
#endif
/*--------------------------------------------------------------------*/
#if	defined(BSD)	/* BSD-UNIX */
#include <strings.h>	/* string header */
extern int errno;
#else	/* not BSD-UNIX */
#include <string.h>	/* string header */
#endif
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
#include <stddef.h>	/* defines useful macros, errno declared */
#include <stdlib.h>	/* errno declared */
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#if	defined(HC1)
/*-------------------------High-C-V1.6--------------------------------*/
#include <system.cf>	/* MS-DOS specific functions */
#endif
#if	defined(MSC) || defined(HC2) || defined(WC)
/*------------MS-C,Zortech-C,High-C-V2.3,WATCOM-C-9.0/386-------------*/
#include <dos.h>	/* MS-DOS specific functions */
#endif
/*--------------------------------------------------------------------*/
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "dirret.h"	/* the dir header file */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define SLOTSIZE	50	/* allocate at least this much */
#define PATHSIZE	128	/* max size of full path name */
/*--------------------------------------------------------------------*/
#if	defined(SYSV) || defined(BSD)	/* UNIX-C */
static struct dirent *find_buf;	/* pointer to file information */
static DIR *dir_ptr;	/* pointer to directory stream */
#endif
/*--------------------------------------------------------------------*/
#if	defined(HC1)
/*-------------------------High-C-V1.6--------------------------------*/
static Find_buffer find_buf;	/* file information structure */
#endif
#if	defined(MSC) || defined(HC2) || defined(WC)	
/*------------MS-C,Zortech-C,High-C-V2.3,WATCOM-C-9.0/386-------------*/
static struct find_t find_buf;	/* file information structure */
#endif
/*--------------------------------------------------------------------*/
/*
	dir_prep(path) readies the directory interface for retrieving
	names from the directory. If this fails, indicated by return 0,
	it means that the directory referenced is un-readable.
*/
#if	defined(ANSI)	/* ANSI-C */
static int dir_prep(char *path)
#else	/* old style */
static int dir_prep(path)
char *path;
#endif
{
#if	!(defined(BSD) || defined(SYSV))
/*-------------------------MS-DOS-Section-----------------------------*/
	char path_name[PATHSIZE];	/* to hold a path name */
	int n = strlen(path);	/* length of directory path */
	strcpy(path_name, path);
	if (n > 0)
		if (path[n-1] != '/')
			strcat(path_name, "/");
	strcat(path_name, "*.*");
	/*
		_A_SUBDIR (Attr_directory) means read all files,
		both regular files and sub-directories.
		_dos_findfirst returns 0 if the first file is found.
	*/
#if	defined(HC1)
/*-------------------------High-C-V1.6--------------------------------*/
	c_find_first(path_name, &find_buf, Attr_directory);
	if(errno == Error_file_not_found || errno == Error_no_more_files) {
#endif
#if	defined(MSC) || defined(HC2) || defined(WC)
/*------------MS-C,Zortech-C,High-C-V2.3,WATCOM-C-9.0/386-------------*/
	if(_dos_findfirst(path_name, _A_SUBDIR, &find_buf) != 0) {
#endif
/*-------------------------MS-DOS-Section-----------------------------*/
		errno = ENOENT;
		return 0;
	}
#else
/*--------------------------UNIX-Section------------------------------*/
	/* open the directory */
	if((dir_ptr = opendir(path)) == (DIR *)0) {
		errno = ENOTDIR;
		return 0;
	}
#endif
/*---------------------------MS-DOS+UNIX------------------------------*/
	return 1;
}
/*--------------------------------------------------------------------*/
/*
	dir_next(path, transfer, dir_hold) return the next directory entry.
	If it returns 0, it indicates no more entries.
*/
#if	defined(ANSI)	/* ANSI-C */
static int dir_next(char *path, DIRRET *transfer, int *dir_hold)
#else	/* old style */
static int dir_next(path, transfer, dir_hold)
char *path;
DIRRET *transfer;
int *dir_hold;
#endif
{
#if	defined(BSD) || defined(SYSV)
/*--------------------------UNIX-Section------------------------------*/
	char path_name[PATHSIZE];	/* to hold a path name */
	int n = strlen(path);	/* length of directory path */
	struct stat stat_buf;
	while (1) {
		if((find_buf = readdir(dir_ptr)) == (struct dirent *)0)
			return 0;
		else if(strlen(find_buf->d_name) <= MAXLEN)
			break;
	}
	strcpy(transfer->name, find_buf->d_name);
	strcpy(path_name, path);
	if (n > 0)
		if (path[n-1] != '/')
			strcat(path_name, "/");
	strcat(path_name, transfer->name);
	/* is it a sub-directory? */
	stat(path_name, &stat_buf);
	transfer->subdir = ((stat_buf.st_mode & S_IFDIR)? 1: 0);
#else
/*-------------------------MS-DOS-Section-----------------------------*/
	if(*dir_hold == 0) {	/* we are not holding data from _dos_findfirst */
#if	defined(HC1)
/*-------------------------High-C-V1.6--------------------------------*/
		find_next(&find_buf);
		if(errno == Error_no_more_files) {
#endif
#if	defined(MSC) || defined(HC2) || defined(WC)
/*------------MS-C,Zortech-C,High-C-V2.3,WATCOM-C-9.0/386-------------*/
		/* _dos_findnext return of 0 is good */
		if(_dos_findnext(&find_buf) != 0) {
#endif
/*-------------------------MS-DOS-Section-----------------------------*/
			return 0;
		}
	} else	/* we are holding data from _dos_findfirst */
		*dir_hold = 0;	/* next call will be _dos_findnext() */
#if	defined(HC1)
/*-------------------------High-C-V1.6--------------------------------*/
	strcpy(transfer->name, find_buf.Name);
	transfer->subdir = (find_buf.Attr & Attr_directory)? 1: 0;
#endif
#if	defined(MSC) || defined(HC2) || defined(WC)
/*------------MS-C,Zortech-C,High-C-V2.3,WATCOM-C-9.0/386-------------*/
	strcpy(transfer->name, find_buf.name);
	transfer->subdir = (find_buf.attrib & _A_SUBDIR)? 1: 0;
#endif
#endif
/*---------------------------MS-DOS+UNIX------------------------------*/
	return 1;
}
/*--------------------------------------------------------------------*/
/*
	dir_retrieve builds a sorted directory table of the directory
	referenced by path. A pointer to an array of structs of file name
	strings is returned. The end of table is marked by a '\0' byte in
	the first element of the name.
*/
#if	defined(ANSI)	/* ANSI-C */
DIRRET *dir_retrieve(char *path, int *dir_entries)
#else	/* old style */
DIRRET *dir_retrieve(path, dir_entries)
char *path;
int *dir_entries;
#endif
{
/*---------------------------MS-DOS+UNIX------------------------------*/
	DIRRET *dir, *tmp;
	int dir_hold = 1;	/* for MS-DOS, hold data from _dos_findfirst */
	int slots = SLOTSIZE, nfiles = 0;

	/* prepare the directory for reading */
	if (dir_prep(path) == 0)
		return (DIRRET *)0;

	/* start with a minimum table size */
	dir = (DIRRET *)malloc(slots*sizeof(DIRRET));

	if(dir == (DIRRET *)0) {	/* table allocation failed */
#if	defined(BSD) || defined(SYSV)
/*---------------------------UNIX-Section-----------------------------*/
		closedir(dir_ptr);	/* close the directory stream */
#endif
/*---------------------------MS-DOS+UNIX------------------------------*/
		errno = ENOMEM;
		return (DIRRET *)0;
	}

	while (1) {
		if(nfiles == slots) {	/* check for table overflow */
			slots <<= 1;	/* double the table size */

			/* re-allocate larger table */
			tmp = (DIRRET *)realloc(dir, slots*sizeof(DIRRET));

			if(tmp == (DIRRET *)0) {	/* table allocation failed */
				free(dir);
#if	defined(BSD) || defined(SYSV)
/*---------------------------UNIX-Section-----------------------------*/
				closedir(dir_ptr);	/* close the directory stream */
#endif
/*---------------------------MS-DOS+UNIX------------------------------*/
				errno = ENOMEM;
				return (DIRRET *)0;
			} else	/* larger table allocated */
				dir = tmp;	/* new table pointer */
		}

		/* get the next file name, if it exists */
		if(dir_next(path, &dir[nfiles], &dir_hold) == 0) {
			dir[nfiles].name[0] = '\0';	/* finish table */
#if	defined(BSD) || defined(SYSV)
/*---------------------------UNIX-Section-----------------------------*/
			closedir(dir_ptr);	/* close the directory stream */
#endif
/*---------------------------MS-DOS+UNIX------------------------------*/
			break;
		}
		++nfiles;
	}
#if	defined(ANSI)
/*------------------------------ANSI-C--------------------------------*/
	qsort(dir, nfiles, sizeof(DIRRET),
			(int(*)(const void *, const void *))strcmp);
#else
/*-------------------------------K&R-C--------------------------------*/
	qsort(dir, nfiles, sizeof(DIRRET), (int(*)())strcmp);
#endif
/*---------------------------MS-DOS+UNIX------------------------------*/
	if(dir_entries != (int *)0)
		*dir_entries = nfiles;
	return dir;
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void dir_free(DIRRET *dir)
#else	/* old style */
void dir_free(dir)
DIRRET *dir;
#endif
{
	free(dir);
}
