#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <math.h>
/*----------------------------------------------------------------------*/
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "fft2.h"	/* function prototypes for fft2.c */
/*----------------------------------------------------------------------*/
#define	MAXDIM	1024
#define	MAXDIM2	2*MAXDIM
/*----------------------------------------------------------------------*/
static float CoSi[MAXDIM];	/* cos(), sin() table */
static int Irev[MAXDIM2];	/* bit-reversal index-table */
static float Data[MAXDIM2];
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int max_dim(void)
#else	/* old style */
int max_dim()
#endif
{
	return MAXDIM;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Trig(int n)
#else	/* old style */
static void Trig(n)
int n;
#endif
{
	double theta = 4.0*atan(1.0)/n;
	int i;
	for (i = 0; i < n; i += 2) {
		double thi = i*theta;
		CoSi[i] = (float)cos(thi);
		CoSi[i+1] = (float)sin(thi);
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Perm(int n)
#else	/* old style */
static void Perm(n)
int n;
#endif
{
	int i, j;
	n <<= 1;
	for (i = 0; i < n; ++i)
		Irev[i] = i;
	for (i = 0, j = 0; i < n; i += 2) {
		int m;
		if (i < j) {
			Irev[i] = j;
			Irev[i+1] = j+1;
			Irev[j] = i;
			Irev[j+1] = i+1;
		}
		for (m = n >> 1; m <= j && m > 1; m >>= 1)
			j -= m;
		j += m;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Four(int n, int m, int is)
#else	/* old style */
static void Four(n, m, is)
int n, m, is;
#endif
{
	int i, kmax, istep, jstep;
	int imax = m << 1;

	for (i = 0; i < imax; ++i) {
		int j = Irev[i];
		if (i < j) {
			float tmp = Data[j];
			Data[j] = Data[i];
			Data[i] = tmp;
		}
	}

	for (jstep = n, kmax = 2; kmax < imax; jstep >>= 1, kmax = istep) {
		int j = 0;
		int k = 0;
		istep = kmax << 1;
		for (; k < kmax; j += jstep, k += 2) {
			float wr = CoSi[j];
			float wi = CoSi[j+1]*is;
			for (i = k; i < imax; i += istep) {
				int j = i + kmax;
				float tr = wr*Data[j] - wi*Data[j+1];
				float ti = wr*Data[j+1] + wi*Data[j];
				Data[j] = Data[i] - tr;
				Data[j+1] = Data[i+1] - ti;
				Data[i] += tr;
				Data[i+1] += ti;
			}
		}
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Fxr1(int n, int is)
#else	/* old style */
static void Fxr1(n, is)
int n, is;
#endif
{
	int i, j;
	const float One = (float)1;
	const float Half = (float)0.5;
	Data[n] = Data[0];
	Data[n+1] = Data[1];
	for (i = 0, j = n; i <= n/2; i += 2, j -= 2) {
		float zr = (One - is*CoSi[i+1])*Half;
		float zi = Half*CoSi[i];
		float difr = Data[i] - Data[j];
		float difi = Data[i+1] + Data[j+1];
		float tr = zr*difr - zi*difi;
		float ti = zr*difi + zi*difr;
		Data[i] = Data[i] - tr;
		Data[i+1] = Data[i+1] - ti;
		if (i < j) {
			Data[j] = Data[j] + tr;
			Data[j+1] = Data[j+1] - ti;
		}
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void Fxr2(int n, int is)
#else	/* old style */
static void Fxr2(n, is)
int n, is;
#endif
{
	int i, j;
	const float Two = (float)2;
	const float One = (float)1;
	const float Half = (float)0.5;
	for (i = 0, j = n; i <= n/2 ; i += 2, j -= 2) {
		float zr = (One + is*CoSi[i+1])*Half;
		float zi = -Half*CoSi[i];
		float difr = Data[i] - Data[j];
		float difi = Data[i+1] + Data[j+1];
		float tr = zr*difr - zi*difi;
		float ti = zr*difi + zi*difr;
		Data[i] = (Data[i] - tr)*Two;
		Data[i+1] = (Data[i+1] - ti)*Two;
		if (i < j) {
			Data[j] = (Data[j] + tr)*Two;
			Data[j+1] = (Data[j+1] - ti)*Two;
		}
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void fft2(int nc, float a[], int n, int is, int direction)
#else	/* old style */
void fft2(nc, a, n, is, direction)
float a[];
int nc, n, is, direction;
#endif
{
	int i, j, ir, ii, jr, ji;
	int m = n >> 1;
	Trig(n);
	if (direction >= 0) {
		Perm(m);
		for (j = 0; j < (n*nc); j += nc) {
			for (i = 0; i < n; ++i)
				Data[i] = a[j+i];
			Four(n, m, is);
			Fxr1(n, is);
			for (i = 0; i < (n+2); ++i)
				a[j+i] = Data[i]/n;
		}
		Perm(n);
		for (jr=0, ji=1; jr <= n; jr+=2, ji+=2) {
			for (ir=0, ii=1, i=0; ir < (2*n); ir+=2, ii+=2, i+=nc) {
				Data[ir] = a[i+jr];
				Data[ii] = a[i+ji];
			}
			Four(n, n, is);
			for (ir=0, ii=1, i=0; ir < (2*n); ir+=2, ii+=2, i+=nc) {
				a[i+jr] = Data[ir];
				a[i+ji] = Data[ii];
			}
		}
	}
	if (direction <= 0) {
		Perm(n);
		for (jr=0, ji=1; jr <= n; jr+=2, ji+=2) {
			for (ir=0, ii=1, i=0; ir < (2*n); ir+=2, ii+=2, i+=nc) {
				Data[ir] = a[i+jr];
				Data[ii] = a[i+ji];
			}
			Four(n, n, -is);
			for (ir=0, ii=1, i=0; ir < (2*n); ir+=2, ii+=2, i+=nc) {
				a[i+jr] = Data[ir];
				a[i+ji] = Data[ii];
			}
		}
		Perm(m);
		for (j = 0; j < (n*nc); j += nc) {
			for (i = 0; i < (n+2); ++i)
				Data[i] = a[j+i]/n;
			Fxr2(n, -is);
			Four(n, m, -is);
			for (i = 0; i < n; ++i)
				a[j+i] = Data[i];
		}
	}
}
