#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD and SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <ctype.h>	/* macros used in character classification */
#include <errno.h>	/* constants used by functions to set errno */
/*----------------------------------------------------------------------*/
#if	defined(HC1)	/* High-C V1.6 */
#if	!defined(ENOMEM)
#define	ENOMEM	Error_not_enough_memory
#endif
#endif
/*----------------------------------------------------------------------*/
#if	defined(BSD)	/* BSD-UNIX */
#include <sys/types.h>	/* size_t defined */
#include <strings.h>	/* string header */
extern int errno;
#else	/* not BSD-UNIX */
#include <string.h>	/* string header */
#endif
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
#include <stddef.h>	/* defines useful macros, errno declared */
#include <stdlib.h>	/* errno declared */
#else	/* not ANSI-C */
#define remove unlink	/* UNIX unlink is equivalent to ANSI remove */
extern char *getenv();
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*----------------------------------------------------------------------*/
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "dirret.h"	/* the dir header file */
#include "imhdr.h"	/* function prototypes for imhdr.c */
/*----------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*----------------------------------------------------------------------*/
#define	LINSIZE	256	/* line buffer size */
#define	PATHSIZ	128	/* path buffer size */
#define	ENVSIZE	128	/* environment buffer size */
#define	PTABSIZ	16	/* path table size */
#define	ITABSIZ	64	/* image table starting size */
#define	KTABSIZ	32	/* key table starting size */
/*----------------------------------------------------------------------*/
#if	defined(BSD)	/* BSD-UNIX */
#define strchr	index	/* BSD name */
#endif
/*----------------------------------------------------------------------*/
#if	defined(BSD) || defined(SYSV)	/* UNIX-C */
#define DELIMIT	':'	/* environment delimitor for UNIX */
#else	/* MS-DOS */
#define DELIMIT	';'	/* environment delimitor for MS-DOS */
#if	defined(strcmp)
#undef	strcmp
#endif
#if	defined(HC2) || defined(WC)	/* High-C V2.3 or WATCOM-C 9.0/386 */
#define strcmp strcmpi	/* case insensitive library function */
#else
#define strcmp	StrCmp	/* MS-DOS is case insensitive */
#endif
#endif
/*----------------------------------------------------------------------*/
typedef struct knode {	/* structure for keynode */
	char *key;
	char *text;
	struct knode *pred;
	struct knode *succ;
} KNODE;
/*----------------------------------------------------------------------*/
typedef struct inode {	/* structure for image node */
	char *name;
	int flag;
	KNODE *head;
	KNODE *tail;
	struct inode *pred;
	struct inode *succ;
} INODE;
/*----------------------------------------------------------------------*/
#if	!(defined(BSD) || defined(SYSV))	/* not UNIX-C */
#if	!(defined(HC2) || defined(WC))	/* not High-C V2.3 or WATCOM-C */
#if	defined(ANSI)	/* ANSI-C */
static int StrCmp(
const char *s,
const char *t)
#else	/* old style */
static int StrCmp(s, t)	/* case insensitive compare */
const char *s;
const char *t;
#endif
{
	for (; toupper(*s) == toupper(*t); s++, t++)
		if (*s == '\0')
			return 0;
	if (toupper((int)*s) > toupper((int)*t))
		return 1;
	else
		return -1;
}
#endif
#endif
/*----------------------------------------------------------------------*/
static INODE *head = (INODE *)0;	/* first image node */
static INODE *tail = (INODE *)0;	/* last image node */
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static INODE *FindImageNode(
const char *name)
#else	/* old style */
static INODE *FindImageNode(name)	/* find image node from name */
const char *name;	/* image name */
#endif
{
	INODE *i = head;	/* first node in image list */
	while (i != (INODE *)0 && strcmp(name, i->name) != 0)
		i = i->succ;
	return i;	/* NULL pointer if not found */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static INODE *InsertImageNode(
const char *name)
#else	/* old style */
static INODE *InsertImageNode(name)	/* insert image node with name */
const char *name;	/* image name */
#endif
{
	INODE *i = (INODE *)malloc(sizeof(INODE));
	if (i == (INODE*)0) {	/* node allocation failed */
		errno = ENOMEM;
		return (INODE *)0;	/* NULL pointer */
	}
	i->name = (char *)malloc(strlen(name) + sizeof(char));
	if (i->name == (char *)0) {	/* name allocation failed */
		free(i);	/* deallocate image node */
		errno = ENOMEM;
		return (INODE *)0;	/* NULL pointer */
	}
	strcpy(i->name, name);	/* save image name */
	i->flag = -1;	/* header file has not been read */
	i->head = (KNODE *)0;	/* empty key link */
	i->tail = (KNODE *)0;
	i->pred = tail;	/* connect backward link */
	i->succ = (INODE *)0;	/* terminate forward link */
	if (i->pred == (INODE *)0)	/* first image node */
		head = i;	/* start forward link */
	else
		tail->succ = i;	/* connect forward link */
	tail = i;	/* start backward link */
	return i;	/* pointer to image node */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static KNODE *FindKeyNode(
INODE *iptr,
const char *key)
#else	/* old style */
static KNODE *FindKeyNode(iptr, key)	/* search for key node */
INODE *iptr;	/* pointer to image node */
const char *key;	/* keyword */
#endif
{
	KNODE *k = iptr->head;	/* first node in keylist */
	while (k != (KNODE *)0 && strcmp(key, k->key) != 0)
		k = k->succ;
	return k;	/* NULL pointer if not found */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static KNODE *InsertKeyNode(
INODE *iptr,
const char *key)
#else	/* old style */
static KNODE *InsertKeyNode(iptr, key)	/* insert key node */
INODE *iptr;	/* pointer to image node */
const char *key;	/* keyword */
#endif
{
	KNODE *k = (KNODE *)malloc(sizeof(KNODE));
	if (k == (KNODE *)0) {	/* keynode allocation failed */
		errno = ENOMEM;
		return (KNODE *)0;	/* NULL pointer */
	}
	k->key = (char *)malloc(strlen(key) + sizeof(char));
	if(k->key == (char *)0) {	/* key allocation failed */
		free(k);	/* deallocate keynode */
		errno = ENOMEM;
		return (KNODE *)0;	/* NULL pointer */
	}
	k->text = (char *)0;	/* no text link */
	strcpy(k->key, key);	/* save keyword */
	k->succ = (KNODE *)0;	/* terminate forward link */
	k->pred = iptr->tail;	/* connect backward link */
	if(k->pred == (KNODE *)0)	/* first key node */
		iptr->head = k;	/* start forward link */
	else
		k->pred->succ = k;	/* connect forward link */
	iptr->tail = k;	/* start backward link */
	return k;	/* pointer to keynode */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static char *SaveKeyText(
KNODE *kptr,
const char *text)
#else	/* old style */
static char *SaveKeyText(kptr, text)	/* save text in key node */
KNODE *kptr;	/* pointer to key node */
const char *text;	/* keyword text */
#endif
{
	if (kptr->text != (char *)0)	/* text already allocated */
		kptr->text = (char *)realloc((void *)kptr->text,
			strlen(text) + sizeof(char));
	else
		kptr->text = (char *)malloc(strlen(text) + sizeof(char));
	if (kptr->text == (char *)0) {	/* text allocation failed */
		errno = ENOMEM;
		return (char *)0;	/* NULL pointer */
	}
	return strcpy(kptr->text, text);	/* save text string */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void RemoveKeyText(
KNODE *kptr)
#else	/* old style */
static void RemoveKeyText(kptr)	/* remove text link from key node */
KNODE *kptr;	/* pointer to key node */
#endif
{
	if (kptr->text != (char *)0)
		free(kptr->text);	/* deallocate text string */
	kptr->text = (char *)0;	/* no keytext */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void RemoveKeyNode(
INODE *iptr,
KNODE *kptr)
#else	/* old style */
static void RemoveKeyNode(iptr, kptr)	/* remove key node from link */
INODE *iptr;	/* pointer to image node */
KNODE *kptr;	/* pointer to key node */
#endif
{
	if (kptr->pred == (KNODE *)0)	/* first keynode */
		iptr->head = kptr->succ;	/* start forward link */
	else	/* not first node */
		kptr->pred->succ = kptr->succ;	/* connect forward link */
	if (kptr->succ == (KNODE *)0)	/* last node */
		iptr->tail = kptr->pred;	/* start backward link */
	else	/* not last node */
		kptr->succ->pred = kptr->pred;	/* connect backward link */
	RemoveKeyText(kptr);	/* deallocate text string */
	free(kptr->key);	/* deallocate keyword */
	free(kptr);	/* deallocate keynode */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void RemoveKeyList(
INODE *iptr)
#else	/* old style */
static void RemoveKeyList(iptr)	/* remove key nodes from link */
INODE *iptr;	/* pointer to image node */
#endif
{
	KNODE *k;
	while ((k = iptr->tail) != (KNODE *)0)
		RemoveKeyNode(iptr, k);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void RemoveImageNode(
INODE *iptr)
#else	/* old style */
static void RemoveImageNode(iptr)	/* remove image node */
INODE *iptr;	/* pointer to image node */
#endif
{
	RemoveKeyList(iptr);
	if (iptr->pred == (INODE *)0)	/* first image node */
		head = iptr->succ;
	else	/* not first node */
		iptr->pred->succ = iptr->succ;
	if (iptr->succ == (INODE *)0)	/* last image node */
		tail = iptr->pred;
	else
		iptr->succ->pred = iptr->pred;
	free(iptr->name);	/* deallocate image name */
	free(iptr);	/* deallocate image node */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void RemoveImageList(void)
#else	/* old style */
static void RemoveImageList()	/* remove image nodes from link */
#endif
{
	INODE *i;
	while ((i = tail) != (INODE *)0)
		RemoveImageNode(i);
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void KeyText(
char **pkey,
char **ptext,
char *s)
#else	/* old style */
static void KeyText(pkey, ptext, s)	/* ptr to ptr to key and text */
char **pkey;	/* pointer to keyword string in s */
char **ptext;	/* pointer to keytext string in s */
char *s;	/* string containing keyword followed by text string */
#endif
{
	char *t = s;
	while (*t != '\0' && *t != '\n')	/* skip until '\0' or '\n' */
		t++;
	if (*t == '\n')	/* if '\n', replace with '\0' */
		*t = '\0';
	while (isspace(*s))	/* skip until non-space */
		s++;
	*pkey = s;	/* keyword address */
	while (*s != '\0' && !isspace(*s))	/* skip keyword */
		s++;
	if (*s != '\0') {	/* if space */
		*s++ = '\0';	/* terminate keyword with '\0' */
		while (isspace(*s))	/* skip until non-space */
			s++;
	}
	*ptext = s;	/* address of text string */
}
/*----------------------------------------------------------------------*/
static char hpath[PATHSIZ] = {(char)0};	/* directory path */
static char hfile[PATHSIZ] = {(char)0};	/* header file path */
#if	defined(ANSI)	/* ANSI-C */
static void HeaderFile(
const char *name)
#else	/* old style */
static void HeaderFile(name)	/* cat header name to path name */
const char *name;	/* image name */
#endif
{
	strcpy(hfile, hpath);	/* header directory path */
	if (strlen(hfile) > 0)
		strcat(hfile, "/");	/* append '/' */
	strncat(hfile, name, MAXLEN - 4);	/* append header name */
	strcat(hfile, ".HDR");	/* append ".HDR" */
}
/*----------------------------------------------------------------------*/
static char  line[LINSIZE] = {(char)0};	/* line buffer */
#if	defined(ANSI)	/* ANSI-C */
static int ReadHeader(
INODE *iptr)
#else	/* old style */
static int ReadHeader(iptr)	/* read keywords from header file */
INODE *iptr;	/* pointer to image node */
#endif
{
	FILE *fp;
	KNODE *k;
	char *s, *key, *text;
	HeaderFile(iptr->name);
	if ((fp = fopen(hfile, "r")) == (FILE *)0) {
		fprintf(stderr,
			"ReadHeader: %s: Can't open header file\n", iptr->name);
		return -1;	/* cannot open header file */
	}
	while (fgets(line, LINSIZE, fp) != (char *)0) {	/* next line */
		KeyText(&key, &text, line);
		if (strlen(key) > 0 && strlen(text) > 0) {
			k = FindKeyNode(iptr, key);
			if (k == (KNODE *)0) {
				k = InsertKeyNode(iptr, key);
				if(k == (KNODE *)0) {
					RemoveKeyList(iptr);
					fclose(fp);
					return 0;
				} else {
					s = SaveKeyText(k, text);
					if(s == (char *)0) {
						RemoveKeyList(iptr);
						fclose(fp);
						return 0;
					}
				}
			}
		}
	}
	fclose(fp);
	return 1;	/*  image header allocation is ok */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static int WriteHeader(
INODE *iptr)
#else	/* old style */
static int WriteHeader(iptr)	/* write header file to directory */
INODE *iptr;	/* pointer to image node */
#endif
{
	FILE *fp;
	KNODE *k;
	HeaderFile(iptr->name);
	if ((fp = fopen(hfile, "w")) == (FILE *)0)
		return 0;	/* cannot open header file */
	for (k = iptr->head; k != (KNODE *)0; k = k->succ){
		sprintf(line, "%s\t%s\n", k->key, k->text);
		fputs(line, fp);
	}
	fclose(fp);
	return 1;	/* keywords are written onto header file */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *strdel(
char *s,
const char c)
#else	/* old style */
char *strdel(s, c)	/* delete all c in string s */
char *s;	/* string with characters to be deleted */
const char c;	/* character to be deleted */
#endif
{
	char  *t, *u, *v;
	for (t = s; *t != '\0';)
		if (*t == c)
			for (v = t, u = v++; *u != '\0'; u++, v++)
				*u = *v;
		else
			t++;
	return s;
}
/*----------------------------------------------------------------------*/
static char *token = (char *)0;	/* pointer to next token field in s */
#if	defined(ANSI)	/* ANSI-C */
char *stoken(
char *s,
const char c)
#else	/* old style */
char *stoken(s, c)	/* find next token in s */
char *s;	/* string containing a sequence of tokens */
const char c;	/* delimiter character */
#endif
{
	char *t;
	if (s == (char *)0)	/* get token from old string */
		if ((s = token) == (char *)0)	/* no old string */
			return (char *)0;
		else
			s = token;	/* next token field in string */
	for (; isspace(*s) || (*s == c); s++)
		;	/* skip delimiters */
	for (t = s; (*t != '\0') && (*t != c) && !isspace(*t); t++)
		;	/* token */
	if (*t != '\0')	/* terminate token if not '\0' */
		*t++ = '\0';	/* advance pointer to next field */
	if (*t == '\0')	/* no more token fields */
		token = (char *)0;
	else
		token = t;	/* next token field */
	return s;
}
/*----------------------------------------------------------------------*/
static char  envstr[ENVSIZE] = {(char)0};	/* environment string */
static char *envtab[PTABSIZ] = {(char *)0};	/* path table */
#if	defined(ANSI)	/* ANSI-C */
char **getpath(
const char *envname,
int *pathcount)
#else	/* old style */
char **getpath(envname, pathcount)	/* get table with path names */
const char *envname;	/* environment keyword */
int *pathcount;	/* number of paths found in the environment string */
#endif
{
	char *s;
	int n;
	*pathcount = 0;
	if ((s = getenv(envname)) == (char *)0)
		return (char **)0;	/* no environment string */
	envstr[0] = '\0';
	strncat(envstr, s, ENVSIZE);	/* cat environment string */
	for (s = envstr; *s != '\0'; s++)	/* change '\\' to '/' */
		if (*s == '\\')
			*s = '/';
	if ((s = stoken(envstr, DELIMIT)) == (char *)0)
		return (char **)0;	/* string is empty */
	for (n = 0; s != (char *)0 && n < PTABSIZ; n++) {
		envtab[n] = s;
		s = stoken((char *)0, DELIMIT);	/* next token */
	}
	*pathcount = n;	/* number of entries */
	return envtab;	/* environment table */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void HeaderPath(
const int i)
#else	/* old style */
static void HeaderPath(i)	/* assign header path from environment */
const int i;	/* select i'th path name */
#endif
{
	char **p;
	int n;
	if ((p = getpath("HDRPATH", &n)) == (char **)0)
		strcpy(hpath, ".");	/* copy current directory path */
	else if (i < n)
		strcpy(hpath, p[i]);	/* copy i'th environment path */
	else
		strcpy(hpath, ".");	/* copy current directory path */
}
/*----------------------------------------------------------------------*/
static unsigned int itabsize = ITABSIZ;	/* starting image table size */
static char **itab = (char **)0;	/* image table pointer */
#if	defined(ANSI)	/* ANSI-C */
char **getitab(int *icount)
#else	/* old style */
char **getitab(icount)	/* returns table with image names */
int *icount;	/* number of entries in table */
#endif
{
	INODE *i = head;
	char **tmp;
	unsigned int n = 0;
	if (itab == (char **)0)	/* image table has not been allocated */
		if ((itab = (char**)malloc(itabsize*sizeof(char*))) == (char**)0) {
			*icount = 0;
			return (char **)0;	/* no space for table */
		}
	while (i != (INODE *)0) {
		if (n >= itabsize) {	/* ask for more table space */
			itabsize <<= 1;	/* double table size */
			tmp = (char **)realloc(itab, itabsize*sizeof(char *));
			if (tmp != (char **)0)
				itab = tmp;	/* new space successfully allocated */
			else
				break;	/* no space for larger table */
		}
		itab[n] = i->name;	/* pointer to image name */
		n++;
		i = i->succ;
	}
	if ((*icount = n) == 0)
		return (char **)0;	/* empty table */
	else
		return itab;	/* return pointer to table */
}
/*----------------------------------------------------------------------*/
static unsigned int ktabsize = KTABSIZ;	/* starting size of key table */
static char **ktab = (char **)0;	/* key table pointer */
#if	defined(ANSI)	/* ANSI-C */
char **getktab(const char *name, int *kcount)
#else	/* old style */
char **getktab(name, kcount)	/* returns table with key names */
const char *name;	/* image name */
int *kcount;	/* number of entries in table */
#endif
{
	INODE *i;
	KNODE *k;
	char **tmp;
	unsigned int n = 0;
	if (ktab == (char **)0)	/* key table has not been allocated */
		if ((ktab = (char**)malloc(ktabsize*sizeof(char *))) == (char**)0) {
			*kcount = 0;
			return (char **)0;	/* no space for table */
		}
	if ((i = FindImageNode(name)) != (INODE *)0) {	/* image found */
		if (i->flag < 0)	/* first access of image */
			if (ReadHeader(i) > 0)	/* read keywords */
				i->flag = 0;	/* key list allocated */
		k = i->head;
		while (k != (KNODE *)0) {
			if (n >= ktabsize) {	/* ask for more table space */
				ktabsize <<= 1;	/* double table size */
				tmp = (char **)realloc(ktab, ktabsize*sizeof(char *));
				if (tmp != (char **)0)
					ktab = tmp;	/* new space successfully allocated */
				else
					break;	/* no space for larger table */
			}
			ktab[n] = k->key;	/* pointer to key name */
			n++;
			k = k->succ;
		}
	}
	if ((*kcount = n) == 0)
		return (char **)0;	/* empty table */
	else
		return ktab;	/* return pointer to table */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *truncname(char *name)
#else	/* old style */
char *truncname(name)	/* truncate name to (MAXLEN-4) characters */
char *name;	/* image name */
#endif
{
	int i;
	for (i = 0; i < (MAXLEN-4); ++i)
		if (name[i] == '\0')
			return name;
	name[MAXLEN-4] = '\0';
	return name;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void rmi(
const char *name)
#else	/* old style */
void rmi(name)	/* remove image node from link and header from directory */
const char *name;	/* image name */
#endif
{
	INODE *i;
	if ((i = FindImageNode(name)) != (INODE *)0) {
		HeaderFile(i->name);
		remove(hfile);	/* remove header file */
		RemoveImageNode(i);
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void rmk(
const char *name,
const char *key)
#else	/* old style */
void rmk(name, key)	/* remove key word from image node */
const char *name;	/* image name */
const char *key;	/* keyword */
#endif
{
	INODE *i;
	KNODE *k;
	if ((i = FindImageNode(name)) != (INODE *)0) {	/* image found */
		if (i->flag < 0)	/* first access of image */
			if (ReadHeader(i) > 0)	/* read keywords */
				i->flag = 0;	/* key list allocated */
		if ((k = FindKeyNode(i, key)) != (KNODE *)0) {
			RemoveKeyNode(i, k);
			i->flag = 1;	/* key list has been modified */
		}
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int xist(
const char *name)
#else	/* old style */
int xist(name)	/* does image node exist? */
const char *name;	/* image name */
#endif
{
	if (FindImageNode(name) == (INODE *)0)
		return 0;	/* image node not found */
	else
		return 1;	/* found it */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int cpi(
const char *old,
const char *new)
#else	/* old style */
int cpi(old, new)	/* copy old image node to new image node */
const char *old;	/* old image name */
const char *new;	/* new image name */
#endif
{
	INODE *i1, *i2;
	KNODE *k1, *k2;
	if ((i1 = FindImageNode(old)) == (INODE *)0)
		return -1;	/* old image is not in list */
	if (i1->flag < 0)	/* first access of old image */
		if (ReadHeader(i1) > 0)	/* read keywords */
			i1->flag = 0;	/* key list allocated */
	if ((i2 = FindImageNode(new)) != (INODE *)0)
		return 0;	/* new image is already in list */
	if ((i2 = InsertImageNode(new)) == (INODE *)0)
		return 0;	/* image node allocation failed */
	for (k1 = i1->head; k1 != (KNODE *)0; k1 = k1->succ) {
		if ((k2 = InsertKeyNode(i2, k1->key)) == (KNODE *)0) {
			RemoveImageNode(i2);
			return 0;	/* key node allocation failed */
		}
		if (SaveKeyText(k2, k1->text) == (char *)0) {
			RemoveImageNode(i2);
			return 0;	/* text allocation failed */
		}
	}
	i2->flag = 1;	/* key list has been modified */
	return 1;	/* all keywords copied */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *putk(
const char *name,
const char *key,
const char *text)
#else	/* old style */
char *putk(name, key, text)	/* save key text for image name */
const char *name;	/* image name */
const char *key;	/* keyword */
const char *text;	/* keytext */
#endif
{
	INODE *i;
	KNODE *k;
	if ((i = FindImageNode(name)) == (INODE *)0) {	/* new image */
		if ((i = InsertImageNode(name)) == (INODE *)0)
			return (char *)0;	/* allocation failed */
	} else if (i->flag < 0)	/* first access of image */
		if (ReadHeader(i) > 0)	/* read keywords */
			i->flag = 0;	/* key list allocated */
	if ((k = FindKeyNode(i, key)) == (KNODE *)0)	/* new key */
		if ((k = InsertKeyNode(i, key)) == (KNODE *)0)
			return (char *)0;	/* allocation failed */
	if (SaveKeyText(k, text) == (char *)0)
		return (char *)0;	/* allocation failed */
	i->flag = 1;	/* key list has been modified */
	return k->text;	/* pointer to key text */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
char *getk(
const char *name,
const char *key)
#else	/* old style */
char *getk(name, key)	/* get key text for image name */
const char *name;	/* image name */
const char *key;	/* keyword */
#endif
{
	INODE *i;
	KNODE *k;
	if ((i = FindImageNode(name)) == (INODE *)0)
		return (char *)0;	/* image not found */
	else if (i->flag < 0)	/* first access of image */
		if (ReadHeader(i) > 0)	/* read keywords */
			i->flag = 0;	/* key list allocated */
	if ((k = FindKeyNode(i, key)) == (KNODE *)0)
		return (char *)0;	/*  keyword not found */
	return k->text;	/* pointer to key text */
}
/*----------------------------------------------------------------------*/
static char iname[MAXLEN+1] = {(char)0};	/* image name buffer */
#if	defined(ANSI)	/* ANSI-C */
int prolog(
const int m)
#else	/* old style */
int prolog(m)	/* allocate image list */
const int m;	/* select the m'th path from environment string */
#endif
{
	DIRRET *dir;
	char *s, *t;
	int dir_entries, n;
	HeaderPath(m);
	if ((dir = dir_retrieve(hpath, &dir_entries)) == (DIRRET *)0)
		return -1;	/* cannot read header directory */
	for (n = 0; n < dir_entries; ++n) {
		s = dir[n].name;	/* next entry name */
		if (!dir[n].subdir) {	/* not sub-directory */
			t = strchr(s, '.');	/* find first '.' */
			if (t != (char *)0)	/* found it */
				if (strcmp(t, ".HDR") == 0 || strcmp(t, ".hdr") == 0) {
					strcpy(iname, s);
					stoken(iname, '.');
					if (InsertImageNode(iname) == (INODE *)0) {
						dir_free(dir);
						return 0;
					}
				}
		}
	}
	dir_free(dir);
	return 1;	/*  image list allocated */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int update(void)
#else	/* old style */
int update()
#endif
{
	INODE *i;
	int ok = 1;
	for (i = head; i != (INODE *)0; i = i->succ)
		if (i->flag > 0)
			if (WriteHeader(i) > 0)
				i->flag = 0;
			else
				ok = 0;
	return ok;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int epilog(void)
#else	/* old style */
int epilog()
#endif
{
	int ok;
	if ((ok = update()) > 0)
		RemoveImageList();
	return ok;
}
/*----------------------------------------------------------------------*/
static char tbuf[LINSIZE] = {(char)0};	/* text buffer */
#if	defined(ANSI)	/* ANSI-C */
int rddir(
const char *dir)
#else	/* old style */
int rddir(dir)	/* read directory file */
const char *dir;	/* directory file name */
#endif
{
	FILE *fp;
	char *key, *text;
	int exist = 1;
	strcpy(iname, "error");	/* default name */
	if ((fp = fopen(dir, "r")) == (FILE *)0)
		return -1;	/* cannot open directory file */
	while (fgets(line, LINSIZE, fp) != (char *)0) {	/* next line */
		KeyText(&key, &text, line);
		if (strlen(key) > 0 && strlen(text) > 0) {
			if (strcmp("BEGIN", key) == 0) {	/* new image */
				strcpy(tbuf, text);	/* copy of text */
				stoken(tbuf, '\\');	/* first token */
				strdel(tbuf, '.');	/* delete '.' */ 
				*iname = '\0';	/* image name */
				strncat(iname, tbuf, MAXLEN - 4);
				if ((exist = xist(iname)) != 0)
					printf("rddir: can't enter %s\n", iname);
			} else if (!exist)	/* image does not exist */
				if (putk(iname, key, text) == (char *)0) {
					fclose(fp);
					return 0;	/* insert failed */
				}
		}
	}
	fclose(fp);
	return 1;	/* all headers installed */
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int wrdir(
const char *dir)
#else	/* old style */
int wrdir(dir)	/* write directory file */
const char *dir;	/* directory file name */
#endif
{
	INODE *i;
	KNODE *k;
	FILE *fp;
	if ((fp = fopen(dir, "w")) == (FILE *)0)
		return -1;	/* cannot open directory file */
	for (i = head; i != (INODE *)0; i = i->succ) {	/* new image */
		sprintf(line, "BEGIN\t%s\n", i->name);
		fputs(line, fp);
		if (i->flag < 0)	/* first access of image */
			if (ReadHeader(i) > 0)	/* read keywords */
				i->flag = 0;	/* key list allocated */
		for (k = i->head; k != (KNODE *)0; k = k->succ) {
			sprintf(line, "%s\t%s\n", k->key, k->text);
			fputs(line, fp);
		}
		sprintf(line, "END\n");
		fputs(line, fp);
		HeaderFile(i->name);
		remove(hfile);	/* remove header file */
		RemoveImageNode(i);
	}
	fclose(fp);
	return 1;	/* all headers written */
}
/*----------------------------------------------------------------------*/
