#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD and SYSV */
/*----------------------------------------------------------------------*/
#include <math.h>
/*----------------------------------------------------------------------*/
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "opconst.h"
/*----------------------------------------------------------------------*/
#define  NINT(x)  ((LONG) (((x) < 0)? (x) - 0.5 : (x) + 0.5))
#define  ABS(x)   (((x) < 0)? -(x): (x))
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void const_uchar(void *buffer,
const int size,
const int op,
const COMPLEX value)
#else	/* old style */
void const_uchar(buffer, size, op, value)
void *buffer;
const int size;
const int op;
const COMPLEX value;
#endif
{
	switch (op) {
		case '+':
			{
				const int v = (int) NINT(value.x);
				const int h = (int) UCHAR_Max;
				UCHAR *p = (UCHAR *) buffer;
				int i, d;
				for (i = 0; i < size; ++i) {
					d = p[i] + v;
					d = (d < 0)? 0: d;
					d = (d > h)? h: d;
					p[i] = (UCHAR) d;
				}
			}
			break;
		case '-':
			{
				const int v = (int) NINT(value.x);
				const int h = (int) UCHAR_Max;
				UCHAR *p = (UCHAR *) buffer;
				int i, d;
				for (i = 0; i < size; ++i) {
					d = p[i] - v;
					d = (d < 0)? 0: d;
					d = (d > h)? h: d;
					p[i] = (UCHAR) d;
				}
			}
			break;
		case '*':
			{
				const float v = ABS(value.x);
				const int h = (int) UCHAR_Max;
				UCHAR *p = (UCHAR *) buffer;
				int i, d;
				for (i = 0; i < size; ++i) {
					d = (int) (p[i]*v + 0.5);
					d = (d > h)? h: d;
					p[i] = (UCHAR) d;
				}
			}
			break;
		case '/':
			{
				const float v = ABS(value.x);
				const int h = (int) UCHAR_Max;
				UCHAR *p = (UCHAR *) buffer;
				int i, d;
				for (i = 0; i < size; ++i) {
					d = (int) (p[i]/v + 0.5);
					d = (d > h)? h: d;
					p[i] = (UCHAR) d;
				}
			}
			break;
		case '@':
			{
				const double v = ABS(value.x);
				const int h = (int) UCHAR_Max;
				UCHAR *p = (UCHAR *) buffer;
				int i, d;
				for (i = 0; i < size; ++i) {
					d = (int) (sqrt((double)p[i])*v + 0.5);
					d = (d > h)? h: d;
					p[i] = (UCHAR) d;
				}
			}
			break;
		case '=':
			{
				const UCHAR v = (UCHAR) NINT(value.x);
				UCHAR *p = (UCHAR *) buffer;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = v;
			}
			break;
		case '>':
			{
				const int v = (int) NINT(value.x);
				UCHAR *p = (UCHAR *) buffer;
				int i;
				for (i = 0; i < size; ++i)
					p[i] >>= v;
			}
			break;
		case '<':
			{
				const int v = (int) NINT(value.x);
				UCHAR *p = (UCHAR *) buffer;
				int i;
				for (i = 0; i < size; ++i)
					p[i] <<= v;
			}
			break;
		default:
			break;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void const_short(void *buffer,
const int size,
const int op,
const COMPLEX value,
const double blank)
#else	/* old style */
void const_short(buffer, size, op, value, blank)
void *buffer;
const int size;
const int op;
const COMPLEX value;
const double blank;
#endif
{
	switch (op) {
		case '+':
			{
				const short b = (short) blank;
				const LONG v = NINT(value.x);
				const LONG h = (LONG) SHORT_Max;
				const LONG l = (LONG) SHORT_Min;
				short *p = (short *) buffer;
				LONG d;
				short s;
				int i;
				for (i = 0; i < size; ++i) {
					d = (LONG) p[i];
					d += v;
					s = (short) d;
					p[i] = (p[i] == b || d > h || d < l)? b: s;
				}
			}
			break;
		case '-':
			{
				const short b = (short) blank;
				const LONG v = NINT(value.x);
				const LONG h = (LONG) SHORT_Max;
				const LONG l = (LONG) SHORT_Min;
				short *p = (short *) buffer;
				LONG d;
				short s;
				int i;
				for (i = 0; i < size; ++i) {
					d = (LONG) p[i];
					d -= v;
					s = (short) d;
					p[i] = (p[i] == b || d < l || d > h)? b: s;
				}
			}
			break;
		case '*':
			{
				const short b = (short) blank;
				const float v = value.x;
				const float h = (float) SHORT_Max;
				const float l = (float) SHORT_Min;
				short *p = (short *) buffer;
				double d;
				short s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d *= v;
					s = (short) (d + 0.5);
					t = (short) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i] == b || d > h || d < l)? b: s;
				}
			}
			break;
		case '/':
			{
				const short b = (short) blank;
				const float v = value.x;
				const float h = (float) SHORT_Max;
				const float l = (float) SHORT_Min;
				short *p = (short *) buffer;
				double d;
				short s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d /= v;
					s = (short) (d + 0.5);
					t = (short) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i] == b || d > h || d < l)? b: s;
				}
			}
			break;
		case '@':
			{
				const short b = (short) blank;
				const double v = (double) value.x;
				const double h = (double) SHORT_Max;
				const double l = (double) SHORT_Min;
				short *p = (short *) buffer;
				double d;
				short s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (p[i] < 0)? 0.0: (double) p[i];
					d = sqrt(d);
					d *= v;
					s = (short) (d + 0.5);
					t = (short) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i]==b || p[i]<0 || d>h || d<l)? b: s;
				}
			}
			break;
		case '=':
			{
				const short b = (short) blank;
				const short v = (short) NINT(value.x);
				short *p = (short *) buffer;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = (p[i] == b)? b: v;
			}
			break;
		case '>':
			{
				const int v = (int) NINT(value.x);
				short *p = (short *) buffer;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = (unsigned short)p[i] >> v;
			}
			break;
		case '<':
			{
				const int v = (int) NINT(value.x);
				short *p = (short *) buffer;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = (unsigned short)p[i] << v;
			}
			break;
		default:
			break;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void const_long(void *buffer,
const int size,
const int op,
const COMPLEX value,
const double blank)
#else	/* old style */
void const_long(buffer, size, op, value, blank)
void *buffer;
const int size;
const int op;
const COMPLEX value;
const double blank;
#endif
{
	switch (op) {
		case '+':
			{
				const LONG b = (LONG) blank;
				const float v = (float) NINT(value.x);
				const float h = (float) LONG_Max;
				const float l = (float) LONG_Min;
				LONG *p = (LONG *) buffer;
				double d;
				LONG s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d += v;
					s = (LONG) (d + 0.5);
					t = (LONG) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i] == b || d > h || d < l)? b: s;
				}
			}
			break;
		case '-':
			{
				const LONG b = (LONG) blank;
				const float v = (float) NINT(value.x);
				const float h = (float) LONG_Max;
				const float l = (float) LONG_Min;
				LONG *p = (LONG *) buffer;
				double d;
				LONG s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d -= v;
					s = (LONG) (d + 0.5);
					t = (LONG) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i] == b || d < l || d > h)? b: s;
				}
			}
			break;
		case '*':
			{
				const LONG b = (LONG) blank;
				const float v = value.x;
				const float h = (float) LONG_Max;
				const float l = (float) LONG_Min;
				LONG *p = (LONG *) buffer;
				double d;
				LONG s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d *= v;
					s = (LONG) (d + 0.5);
					t = (LONG) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i] == b || d > h || d < l)? b: s;
				}
			}
			break;
		case '/':
			{
				const LONG b = (LONG) blank;
				const float v = value.x;
				const float h = (float) LONG_Max;
				const float l = (float) LONG_Min;
				LONG *p = (LONG *) buffer;
				double d;
				LONG s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (double) p[i];
					d /= v;
					s = (LONG) (d + 0.5);
					t = (LONG) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i] == b || d > h || d < l)? b: s;
				}
			}
			break;
		case '@':
			{
				const LONG b = (LONG) blank;
				const double v = (double) value.x;
				const double h = (double) LONG_Max;
				const double l = (double) LONG_Min;
				LONG *p = (LONG *) buffer;
				double d;
				LONG s, t;
				int i;
				for (i = 0; i < size; ++i) {
					d = (p[i] < 0)? 0.0: (double) p[i];
					d = sqrt(d);
					d *= v;
					s = (LONG) (d + 0.5);
					t = (LONG) (d - 0.5);
					s = (t < 0)? t: s;
					p[i] = (p[i]==b || p[i]<0 || d>h || d<l)? b: s;
				}
			}
			break;
		case '=':
			{
				const LONG b = (LONG) blank;
				const LONG v = NINT(value.x);
				LONG *p = (LONG *) buffer;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = (p[i] == b)? b: v;
			}
			break;
		case '>':
			{
				const int v = (int) NINT(value.x);
				LONG *p = (LONG *) buffer;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = (ULONG)p[i] >> v;
			}
			break;
		case '<':
			{
				const int v = (int) NINT(value.x);
				LONG *p = (LONG *) buffer;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = (ULONG)p[i] << v;
			}
			break;
		default:
			break;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void const_float(void *buffer,
const int size,
const int op,
const COMPLEX value,
const double blank)
#else	/* old style */
void const_float(buffer, size, op, value, blank)
void *buffer;
const int size;
const int op;
const COMPLEX value;
const double blank;
#endif
{
	switch (op) {
		case '+':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const float v = value.x;
				float *p = (float *) buffer;
				float d;
				int i;
				for (i = 0; i < size; ++i) {
					d = p[i] + v;
					p[i] = (p[i] > h || d > h || d < -h)? b: d;
				}
			}
			break;
		case '-':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const float v = value.x;
				float *p = (float *) buffer;
				float d;
				int i;
				for (i = 0; i < size; ++i) {
					d = p[i] - v;
					p[i] = (p[i] > h || d > h || d < -h)? b: d;
				}
			}
			break;
		case '*':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const double v = value.x;
				float *p = (float *) buffer;
				double d;
				int i;
				for (i = 0; i < size; ++i) {
					d = p[i]*v;
					p[i] = (p[i] > h || d > h || d < -h)? b: (float)d;
				}
			}
			break;
		case '/':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const double v = value.x;
				float *p = (float *) buffer;
				double d;
				int i;
				for (i = 0; i < size; ++i) {
					d = p[i]/v;
					p[i] = (p[i] > h || d > h || d < -h)? b: (float)d;
				}
			}
			break;
		case '@':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const double v = value.x;
				float *p = (float *) buffer;
				double d;
				int i;
				for (i = 0; i < size; ++i) {
					d = (p[i] < 0)? 0.0: (double) p[i];
					d = sqrt(d)*v;
					p[i] = (p[i]>h || p[i]<0 || d>h || d<-h)? b: (float)d;
				}
			}
			break;
		case '=':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const float v = value.x;
				float *p = (float *) buffer;
				int i;
				for (i = 0; i < size; ++i)
					p[i] = (p[i] > h)? b: v;
			}
			break;
		default:
			break;
	}
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void const_complex(void *buffer,
const int size,
const int op,
const COMPLEX value,
const double blank)
#else	/* old style */
void const_complex(buffer, size, op, value, blank)
void *buffer;
const int size;
const int op;
const COMPLEX value;
const double blank;
#endif
{
	switch (op) {
		case '+':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const float vx = value.x;
				const float vy = value.y;
				COMPLEX *p = (COMPLEX *) buffer;
				float dx, dy;
				int i, ip;
				for (i = 0; i < size; ++i) {
					dx = p[i].x + vx;
					dy = p[i].y + vy;
					ip = p[i].x > h || p[i].y > h;
					p[i].x = (ip || dx > h || dx < -h)? b: dx;
					p[i].y = (ip || dy > h || dy < -h)? b: dy;
				}
			}
			break;
		case '-':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const float vx = value.x;
				const float vy = value.y;
				COMPLEX *p = (COMPLEX *) buffer;
				float dx, dy;
				int i, ip;
				for (i = 0; i < size; ++i) {
					dx = p[i].x - vx;
					dy = p[i].y - vy;
					ip = p[i].x > h || p[i].y > h;
					p[i].x = (ip || dx > h || dx < -h)? b: dx;
					p[i].y = (ip || dy > h || dy < -h)? b: dy;
				}
			}
			break;
		case '*':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const double vx = value.x;
				const double vy = value.y;
				COMPLEX *p = (COMPLEX *) buffer;
				double dx, dy;
				int i, ip;
				for (i = 0; i < size; ++i) {
					dx = vx*p[i].x - vy*p[i].y;
					dy = vx*p[i].y + vy*p[i].x;
					ip = p[i].x > h || p[i].y > h;
					p[i].x = (ip || dx > h || dx < -h)? b: (float)dx;
					p[i].y = (ip || dy > h || dy < -h)? b: (float)dy;
				}
			}
			break;
		case '/':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const double vx = value.x;
				const double vy = value.y;
				const double vsq = vx*vx + vy*vy;
				COMPLEX *p = (COMPLEX *) buffer;
				double dx, dy;
				int i, ip;
				for (i = 0; i < size; ++i) {
					dx = vx*p[i].x + vy*p[i].y;
					dy = vx*p[i].y - vy*p[i].x;
					dx /= vsq;
					dy /= vsq;
					ip = p[i].x > h || p[i].y > h;
					p[i].x = (ip || dx > h || dx < -h)? b: (float)dx;
					p[i].y = (ip || dy > h || dy < -h)? b: (float)dy;
				}
			}
			break;
		case '#':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				COMPLEX *p = (COMPLEX *) buffer;
				int i, ip;
				for (i = 0; i < size; ++i) {
					ip = p[i].x > h || p[i].y > h;
					p[i].y = (ip)? b: -p[i].y;
				}
			}
			break;
		case '=':
			{
				const float b = (float) blank;
				const float h = (float) (0.9*blank);
				const float vx = value.x;
				const float vy = value.y;
				COMPLEX *p = (COMPLEX *) buffer;
				int i, ip;
				for (i = 0; i < size; ++i) {
					ip = p[i].x > h || p[i].y > h;
					p[i].x = (ip)? b: vx;
					p[i].y = (ip)? b: vy;
				}
			}
			break;
		default:
			break;
	}
}
/*----------------------------------------------------------------------*/
