#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD and SYSV */
#if	defined(HC1)	/* High-C V1.6 */
#include <msdos.cf>
#endif
#if	defined(MSC) || defined(HC2) || defined(WC)
/* MS-C, High-C V2.3, or WATCOM-C 9.0/386 */
#include <dos.h>
#include <conio.h>
#endif
#if	defined(MSC) || defined(WC)	/* MS-C or WATCOM-C 9.0/386 */
#include <graph.h>
#endif
#include "vga256.h"	/* constants and prototypes for vga256.c */
/*--------------------------------------------------------------------*/
#define	VIDEO_IO	0x10	/* Video Interrupt */
/*--------------------------------------------------------------------*/
/* Tseng Labs ET4000 and Paradise Graphics Controllers */
#define	TS_Ind	0x3c4	/* Timing Sequencer Index Register */
#define	TS_Dat	0x3c5	/* TS Data Register */
#define	GDC_Seg	0x3cd	/* Graphics Data Controller Segment Select */
#define	GDC_Ind	0x3ce	/* GDC Index Register*/
#define	GDC_Dat	0x3cf	/* GDC Data Register */
#define	DAC_Sta	0x3c7	/* DAC State Register (read only) */
#define	PEL_ARM	0x3c7	/* PEL Address Read Mode (write only) */
#define	PEL_AWM	0x3c8	/* PEL Address Write Mode */
#define	PEL_Dat	0x3c9	/* PEL Data Register */
/*--------------------------------------------------------------------*/
#if	defined(MSC)	/* MS-C or Zortech-C */
#if	defined(FP_SEG)
#undef	FP_SEG
#endif
#if	defined(FP_OFF)
#undef	FP_OFF
#endif
#define FP_SEG(fp) (*((unsigned *)&(fp) + 1))
#define FP_OFF(fp) (*((unsigned *)&(fp)))
#endif
/*--------------------------------------------------------------------*/
#if	defined(HC1)	/* High-C V1.6 */
typedef struct {	/* _far pointer */
	UINT off;
	USHORT seg;
} OVERLAY;
#define FP_SEG(fp) ((OVERLAY *)&(fp))->seg	/* segment selector */
#define FP_OFF(fp) ((OVERLAY *)&(fp))->off	/* 32-bit offset */
#endif
/*--------------------------------------------------------------------*/
#if	defined(WC)	/* WATCOM-C 9.0/386 */
#if	defined(FP_SEG)
#undef	FP_SEG
#endif
#if	defined(FP_OFF)
#undef	FP_OFF
#endif
#define FP_SEG(farptr) (*(unsigned short *)((unsigned *)&(farptr) + 1))
#define FP_OFF(farptr) (*((unsigned *)&(farptr)))
#endif
/*--------------------------------------------------------------------*/
typedef union {
	ULONG w32;
	struct {
		unsigned short lo;
		unsigned short hi;
	} w16;
} LONGWORD;
/*--------------------------------------------------------------------*/
#if	defined(MSC)	/* MS-C */
#define	VGASEG	0xa000	/* base paragraph of video RAM */
#define	VGAOFF	0x0000	/* base offset of video RAM */
#endif
/*--------------------------------------------------------------------*/
#if	defined(HC1) || defined(HC2) || defined(WC)
/* High-C or WATCOM-C 9.0/386 */
#define	VGASEG	0x34	/* segment that maps the first Mb of RAM */
#define	VGAOFF	0xa0000	/* base offset of video RAM */
#endif
/*--------------------------------------------------------------------*/
#if	defined(MSC)	/* MS-C */
static UCHAR far *vram_seg = (UCHAR far *)0;
#endif
/*--------------------------------------------------------------------*/
#if	defined(HC1) || defined(HC2)	/* High-C */
static _far volatile UCHAR *vram_seg = (_far UCHAR *)0;
#endif
/*--------------------------------------------------------------------*/
#if	defined(WC)	/* WATCOM-C 9.0/386 */
#if	defined(DOS4G)	/* Rational Systems DOS/4GW */
static volatile UCHAR *vram_seg = (UCHAR *)0;
#else	/* Phar Lap 386|DOS-Extender */
static volatile UCHAR __far *vram_seg = (UCHAR __far *)0;
#endif
#endif
/*--------------------------------------------------------------------*/
static int row_siz = 0;	/* row size in pixels */
static int col_siz = 0;	/* screen size in rows */
static unsigned int scr_seg = 0;	/* current screen segment */
/*--------------------------------------------------------------------*/
int rowsize(void) { return row_siz; }
/*--------------------------------------------------------------------*/
int colsize(void) { return col_siz; }
/*--------------------------------------------------------------------*/
void set_mode(int mode)	/* Set Video Mode */
{
#if	defined(WC)	/* WATCOM-C 9.0/386 */
	union REGS r;
	r.h.ah = 0x0;
	r.h.al = (UCHAR) mode;
	int386(VIDEO_IO, &r, &r);
#endif
#if	defined(MSC) || defined(HC2)	/* MS-C or High-C V2.3 */
	union REGS r;
	r.h.ah = 0x0;
	r.h.al = (UCHAR) mode;
	int86(VIDEO_IO, &r, &r);
#endif
#if	defined(HC1)	/* High-C V1.6 */
	Registers.AX.LH.H = 0x0;
	Registers.AX.LH.L = (UCHAR) mode;
	callint(VIDEO_IO);
#endif
#if	defined(DOS4G)	/* Rational Systems DOS/4GW */
	vram_seg = (UCHAR *) VGAOFF;
#else	/* otherwise a far pointer */
	FP_OFF(vram_seg) = VGAOFF;
	FP_SEG(vram_seg) = VGASEG;
#endif
	switch (mode) {
		case _MRES256COLOR:
			row_siz = 320;
			col_siz = 200;
			break;
		case _ET640x350:
			row_siz = 640;
			col_siz = 350;
			break;
		case _ET640x480:
			row_siz = 640;
			col_siz = 480;
			break;
		case _ET640x400:
			row_siz = 640;
			col_siz = 400;
			break;
		case _ET800x600:
			row_siz = 800;
			col_siz = 600;
			break;
		case _ET1024x768:
			row_siz = 1024;
			col_siz = 768;
			break;
		case _PA640x400:
			row_siz = 640;
			col_siz = 400;
			break;
		case _PA640x480:
			row_siz = 640;
			col_siz = 480;
			break;
		default:
			break;
	}
	scr_seg = 0;
}
/*--------------------------------------------------------------------*/
void set_cursor(int row, int col)	/* Set Cursor Position */
{
#if	defined(WC)	/* WATCOM-C 9.0/386 */
	union REGS r;
	r.h.ah = 0x2;
	r.h.dh = (UCHAR) row;
	r.h.dl = (UCHAR) col;
	r.h.bh = 0x0;	/* page number */
	int386(VIDEO_IO, &r, &r);
#endif
#if	defined(MSC) || defined(HC2)	/* MS-C or High-C V2.3 */
	union REGS r;
	r.h.ah = 0x2;
	r.h.dh = (UCHAR) row;
	r.h.dl = (UCHAR) col;
	r.h.bh = 0x0;	/* page number */
	int86(VIDEO_IO, &r, &r);
#endif
#if	defined(HC1)	/* High-C V1.6 */
	Registers.AX.LH.H = 0x2;
	Registers.DX.LH.H = (UCHAR) row;
	Registers.DX.LH.L = (UCHAR) col;
	Registers.BX.LH.H = 0x0;	/* page number */
	callint(VIDEO_IO);
#endif
}
/*--------------------------------------------------------------------*/
void draw_pixel(int row, int col, UCHAR color)	/* Draw Graphics Pixel */
{
#if	defined(WC)	/* WATCOM-C 9.0/386 */
	union REGS r;
	r.h.ah = 0xc;
	r.h.al = color;
	r.h.bh = 0x0;	/* page number */
	r.w.dx = row;
	r.w.cx = col;
	int386(VIDEO_IO, &r, &r);
#endif
#if	defined(MSC) || defined(HC2)	/* MS-C or High-C V2.3 */
	union REGS r;
	r.h.ah = 0xc;
	r.h.al = color;
	r.h.bh = 0x0;	/* page number */
	r.x.dx = row;
	r.x.cx = col;
	int86(VIDEO_IO, &r, &r);
#endif
#if	defined(HC1)	/* High-C V1.6 */
	Registers.AX.LH.H = 0xc;
	Registers.AX.LH.L = color;
	Registers.BX.LH.H = 0x0;	/* page number */
	Registers.DX.W = row;
	Registers.CX.W = col;
	callint(VIDEO_IO);
#endif
}
/*--------------------------------------------------------------------*/
UCHAR read_pixel(int row, int col)	/* Read Graphics Pixel */
{
#if	defined(WC)	/* WATCOM-C 9.0/386 */
	union REGS r;
	r.h.ah = 0xd;
	r.h.bh = 0x0;	/* page number */
	r.w.dx = row;
	r.w.cx = col;
	int386(VIDEO_IO, &r, &r);
	return r.h.al;
#endif
#if	defined(MSC) || defined(HC2)	/* MS-C or High-C V2.3 */
	union REGS r;
	r.h.ah = 0xd;
	r.h.bh = 0x0;	/* page number */
	r.x.dx = row;
	r.x.cx = col;
	int86(VIDEO_IO, &r, &r);
	return r.h.al;
#endif
#if	defined(HC1)	/* High-C V1.6 */
	Registers.AX.LH.H = 0xd;
	Registers.BX.LH.H = 0x0;	/* page number */
	Registers.DX.W = row;
	Registers.CX.W = col;
	callint(VIDEO_IO);
	return Registers.AX.LH.L;
#endif
}
/*--------------------------------------------------------------------*/
void write_tty(	/* Write Character in TTY Mode */
UCHAR ascii,	/* ASCII code */
UCHAR color)	/* foreground color, if graphics mode */
{
#if	defined(WC)	/* WATCOM-C 9.0/386 */
	union REGS r;
	r.h.ah = 0xe;
	r.h.al = ascii;
	r.h.bl = color;
	int386(VIDEO_IO, &r, &r);
#endif
#if	defined(MSC) || defined(HC2)	/* MS-C or High-C V2.3 */
	union REGS r;
	r.h.ah = 0xe;
	r.h.al = ascii;
	r.h.bl = color;
	int86(VIDEO_IO, &r, &r);
#endif
#if	defined(HC1)	/* High-C V1.6 */
	Registers.AX.LH.H = 0xe;
	Registers.AX.LH.L = ascii;
	Registers.BX.LH.L = color;
	callint(VIDEO_IO);
#endif
}
/*--------------------------------------------------------------------*/
void write_str(char *str, int row, int col, UCHAR color)
{
	const char *p;
	set_cursor(row, col);
	for (p = str; *p != '\0'; p++)
		write_tty(*p, color);
}
/*--------------------------------------------------------------------*/
unsigned get_mode(void)	/* Read Video State */
{
#if	defined(WC)	/* WATCOM-C 9.0/386 */
	union REGS r;
	r.h.ah = 0xf;
	int386(VIDEO_IO, &r, &r);
	return r.h.al;
#endif
#if	defined(MSC) || defined(HC2)	/* MS-C or High-C V2.3 */
	union REGS r;
	r.h.ah = 0xf;
	int86(VIDEO_IO, &r, &r);
	return r.h.al;
#endif
#if	defined(HC1)	/* High-C V1.6 */
	Registers.AX.LH.H = 0xf;
	callint(VIDEO_IO);
	return Registers.AX.LH.L;
#endif
}
/*--------------------------------------------------------------------*/
void set_color(	/* Sets the 18-bit color values */
UCHAR *rgb)	/* (R, G, B) for each pixel value */
{
	int i;
#if	defined(MSC) || defined(WC)	/* MS-C or WATCOM-C 9.0/386 */
	outp(PEL_AWM, 0);
	for (i = 0; i < 768; i++)
		outp(PEL_Dat, rgb[i]);
#endif
#if	defined(HC1) || defined(HC2)	/* High-C */
	_outb(PEL_AWM, 0);
	for (i = 0; i < 768; i++)
		_outb(PEL_Dat, rgb[i]);
#endif
}
/*--------------------------------------------------------------------*/
void get_color(	/* Returns the 18-bit color values */
UCHAR *rgb)	/* (R, G, B) for each pixel value */
{
	int i;
#if	defined(MSC) || defined(WC)	/* MS-C or WATCOM-C 9.0/386 */
	outp(PEL_ARM, 0);
	for (i = 0; i < 768; i++)
		rgb[i] = (UCHAR)inp(PEL_Dat);
#endif
#if	defined(HC1) || defined(HC2)	/* High-C */
	_outb(PEL_ARM, 0);
	for (i = 0; i < 768; i++)
		rgb[i] = _inb(PEL_Dat);
#endif
}
/*--------------------------------------------------------------------*/
void set_seg(	/* select video segment */
int n)	/* segment number */
{
	int m;
	n &= 0xf;
	m = n << 4;
#if	defined(ET4000)	/* ET4000 Graphics Controller */
	m |= n;
#if	defined(MSC) || defined(WC)	/* MS-C or WATCOM-C 9.0/386 */
	outp(TS_Ind, 2);	/* select TS-2 register */
	outp(TS_Dat, 0xf);	/* enable all planes */
	outp(TS_Ind, 4);	/* select TS-4 register */
	outp(TS_Dat, 0xc);	/* chain 4, not odd/even */
	outp(GDC_Ind, 1);	/* select GDC-1 register */
	outp(GDC_Dat, 0x0);	/* no set/reset */
	outp(GDC_Ind, 3);	/* select GDC-3 register */
	outp(GDC_Dat, 0x0);	/* no rotate/func */
	outp(GDC_Ind, 5);	/* select GDC-5 register */
	outp(GDC_Dat, 0x40);	/* write/read mode 0 */
	outp(GDC_Ind, 6);	/* select DGC-6 register */
	outp(GDC_Dat, 0x5);	/* map = A0000h for 64k, graphics mode */
	outp(GDC_Ind, 8);	/* select DGC-8 register */
	outp(GDC_Dat, 0xff);	/* bitmask = all bits */
	outp(GDC_Seg, m);	/* select video segment */
#endif
#if	defined(HC1) || defined(HC2)	/* High-C */
	_outb(TS_Ind, 2);	/* select TS-2 register */
	_outb(TS_Dat, 0xf);	/* enable all planes */
	_outb(TS_Ind, 4);	/* select TS-4 register */
	_outb(TS_Dat, 0xc);	/* chain 4, not odd/even */
	_outb(GDC_Ind, 1);	/* select GDC-1 register */
	_outb(GDC_Dat, 0x0);	/* no set/reset */
	_outb(GDC_Ind, 3);	/* select GDC-3 register */
	_outb(GDC_Dat, 0x0);	/* no rotate/func */
	_outb(GDC_Ind, 5);	/* select GDC-5 register */
	_outb(GDC_Dat, 0x40);	/* write/read mode 0 */
	_outb(GDC_Ind, 6);	/* select DGC-6 register */
	_outb(GDC_Dat, 0x5);	/* map = A0000h for 64k, graphics mode */
	_outb(GDC_Ind, 8);	/* select DGC-8 register */
	_outb(GDC_Dat, 0xff);	/* bitmask = all bits */
	_outb(GDC_Seg, m);	/* select video segment */
#endif
#else	/* The Paradise Graphics Controller */
#if	defined(MSC)	/* MS-C or WATCOM-C 9.0/386 */
	 outp(GDC_Ind, 15);	/* select PR5 register */
	 outp(GDC_Dat, 5);	/* unlock PR0A to PR4 */
	 outp(GDC_Ind, 9);	/* select PR0A register */
	 outp(GDC_Dat, m);	/* select first page in segment */
	 outp(GDC_Ind, 15);	/* select PR5 register */
	 outp(GDC_Dat, 0);	/* lock PR0A to PR4 */
#endif
#if	defined(HC1) || defined(HC2)	/* High-C */
	_outb(GDC_Ind, 15);	/* select PR5 register */
	_outb(GDC_Dat, 5);	/* unlock PR0A to PR4 */
	_outb(GDC_Ind, 9);	/* select PR0A register */
	_outb(GDC_Dat, m);	/* select first page in segment */
	_outb(GDC_Ind, 15);	/* select PR5 register */
	_outb(GDC_Dat, 0);	/* lock PR0A to PR4 */
#endif
#endif
}
/*--------------------------------------------------------------------*/
void draw_line(int row, int col, UCHAR *color, int len)
#if	defined(USE_BIOS)	/* use BIOS calls to draw line */
{
	int maxlen = row_siz - col;
	int i;
	if (len > maxlen)
		len = maxlen;
	if (row < col_siz)
		for (i = col; i < len; ++i)
			draw_pixel(row, i, color[i]);
}
#else	/*	write directly to video RAM */
{
	int maxlen = row_siz - col;
	LONGWORD r;
	int i;
	if (len > maxlen)
		len = maxlen;
	if (row < col_siz) {
		r.w32 = row;
		r.w32 *= row_siz;
		r.w32 += col;
		for (i = len+1; --i > 0; ++color, ++r.w32) {
			if (r.w16.hi != scr_seg) {
				scr_seg = r.w16.hi;
				set_seg(scr_seg);
			}
			vram_seg[r.w16.lo] = *color;
		}
	}
}
#endif
