#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#if	defined(BSD)	/* BSD-UNIX */
#include <strings.h>	/* string header */
#else	/* not BSD-UNIX */
#include <string.h>	/* string header */
#endif
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
#include <stddef.h>	/* defines useful macros, errno declared */
#include <stdlib.h>	/* errno declared */
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*----------------------------------------------------------------------*/
#if	defined(WC)	/* WATCOM-C */
#include <malloc.h>
#endif
/*----------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "vmalloc.h"	/* function declarations */
/*----------------------------------------------------------------------*/
#if	defined(WC)	/* WATCOM-C */
void heap_dump(void)
{
	struct _heapinfo h_info;
	int heap_status;

	h_info._pentry = NULL;
	for ( ; ; ) {
		heap_status = _heapwalk( &h_info );
		if (heap_status != _HEAPOK) break;
		printf(" %s block at %Fp of size %4.4X\n",
			(h_info._useflag == _USEDENTRY ? "USED" : "FREE"),
			h_info._pentry, h_info._size);
	}
	switch (heap_status) {
		case _HEAPEND:
			printf("OK - end of heap\n");
			break;
		case _HEAPEMPTY:
			printf("OK - heap is empty\n");
			break;
		case _HEAPBADBEGIN:
			printf("ERROR - heap is damaged\n");
			break;
		case _HEAPBADPTR:
			printf("ERROR - bad pointer to heap\n");
			break;
		case _HEAPBADNODE:
			printf("ERROR - bad node in heap\n");
			break;
	}
}
int bad_heap(void)
{
	int heap_status;

	heap_status = _heapset( 0xff );
	switch (heap_status) {
		case _HEAPOK:
			return 0;
			break;
		case _HEAPEMPTY:
			return 0;
			break;
		case _HEAPBADBEGIN:
			return 1;
			break;
		case _HEAPBADNODE:
			return 1;
			break;
	}
	return 0;
}
#else	/* dummy functions */
void heap_dump() { }
int bad_heap() { return 0; }
#endif
/*----------------------------------------------------------------------*/
/*
	Maximum number of separate buffers which have been
	allocated but not freed
*/
#define MAXMALLOCS	10000
#define KP	0xaa	/* definition of known pattern */
#define KPW	8	/* definition of known pattern width */

static int nummallocs = 0;

struct mtype {
	unsigned char *addr;
	unsigned size;
};
static struct mtype m[MAXMALLOCS];

#if	defined(ANSI)	/* ANSI-C */
static void trace(char *s)
#else	/* old style */
static void trace(s)
char *s;
#endif
{
	static FILE *out = (FILE *)0;
	if (out == (FILE *)0) {
#if	defined(ANSI)	/* ANSI-C */
		remove("vmalloc");
#else	/* old style */
		unlink("vmalloc");
#endif
		out = fopen("vmalloc", "w");
		setbuf(out, (char *)0);
	}
	fputs(s, out);
}

#if	defined(ANSI)	/* ANSI-C */
static void dumpbuf(int x)
#else	/* old style */
static void dumpbuf(x)
int x;
#endif
{
	char s[80];
	unsigned char *c = m[x].addr - 2;

	/* dump malloc buffer to the vmalloc file */
	while (c <= m[x].addr + m[x].size + KPW + KPW - 1) {
		sprintf(s, "%08x : %02x ", (unsigned)c, *c);
		if (c == m[x].addr)
			strcat(s, "<= leading known pattern");
		if (c == m[x].addr + KPW)
			strcat(s, "<= address of malloc buffer");
		if (c == m[x].addr + m[x].size + KPW)
			strcat(s, "<= trailing known pattern");
		strcat(s, "\n");
		trace(s);
		++c;
	}
}

#if	defined(ANSI)	/* ANSI-C */
void vverify(char *id)
#else	/* old style */
void vverify(id)
char *id;
#endif
{
	char s[80];
	int x, c;

	/* verify the entire malloc heap */
	for (x = 0; x < nummallocs; ++x)
		if (m[x].addr != (unsigned char *)0)
			for (c = 0; c < KPW; ++c) {
				if (*(m[x].addr + c) != KP || *(m[x].addr +
						m[x].size + KPW + c) != KP) {
					sprintf(s, "ERROR: Malloc Area Corrupted (%s)\n", id);
					trace(s);
					fputs(s, stderr);
					dumpbuf(x);
					exit(EXIT_FAILURE);
				}
			}
}

#if	defined(ANSI)	/* ANSI-C */
void *vmalloc(unsigned size)
#else	/* old style */
void *vmalloc(size)
unsigned size;
#endif
{
	char s[80];
	unsigned char *buffer;
	int c, x;
	vverify("vmalloc");
	if ( bad_heap() ) heap_dump();
	if ((buffer = (unsigned char *)
			calloc(size + KPW + KPW, 1)) == (unsigned char *)0) {
		fprintf(stderr, "ERROR: calloc returned NULL\n");
		trace("ERROR: calloc returned NULL\n");
		exit(EXIT_FAILURE);
	}
	sprintf(s, "%08x : vmalloc size  = %d\n",(unsigned)buffer, size);
	trace(s);

	/* find a place for an entry in m */
	for (x = 0; x < MAXMALLOCS && m[x].addr != (unsigned char *)0; ++x)
		;
	if (x == MAXMALLOCS) {
		sprintf(s, "ERROR: too many mallocs\n");
		fprintf(stderr, s);
		trace(s);
		exit(EXIT_FAILURE);
	}
	m[x].addr = buffer;
	m[x].size = size;
	if (x == nummallocs)
		++nummallocs;
	for (c = 0; c < KPW; ++c) {
		*(m[x].addr + c) = KP;
		*(m[x].addr + m[x].size + KPW + c) = KP;
	}
	return (void *)(buffer + KPW);
}

#if	defined(ANSI)	/* ANSI-C */
void *vcalloc(unsigned n, unsigned size)
#else	/* old style */
void *vcalloc(n, size)
unsigned n;
unsigned size;
#endif
{
	return vmalloc(n * size);
}

#if	defined(ANSI)	/* ANSI-C */
void vfree(void *buffer)
#else	/* old style */
void vfree(buffer)
void *buffer;
#endif
{
	char s[80];
	unsigned char *b = (unsigned char *)buffer - KPW;
	int x;
	vverify("vfree");
	for (x = 0; x < nummallocs && m[x].addr != b; ++x)
		;
	if (x == nummallocs) {
		strcpy(s, "ERROR: location to free is not in list\n");
		fprintf(stderr, s);
		trace(s);
		exit(EXIT_FAILURE);
	}
	sprintf(s, "%08x : vfree\n", (unsigned)b);
	trace(s);
	free(b);
	m[x].addr = (unsigned char *)0;
	if (x == nummallocs - 1)
		--nummallocs;
}

#if	defined(ANSI)	/* ANSI-C */
void *vrealloc(void *buffer, unsigned size)
#else	/* old style */
void *vrealloc(buffer, size)
void *buffer;
unsigned size;
#endif
{
	char s[80];
	unsigned char *b2, *b = (unsigned char *)buffer - KPW;
	int c, x;
	vverify("vrealloc");
	for (x = 0; x < nummallocs && m[x].addr != b; ++x)
		;
	if (x == nummallocs) {
		sprintf(s, "ERROR: location to realloc not in list\n");
		fprintf(stderr, s);
		trace(s);
		exit(EXIT_FAILURE);
	}
	sprintf(s, "%08x : vrealloc request\n", (unsigned)b);
	trace(s);
	for (c = 0; c < KPW; ++c)
		*(m[x].addr + m[x].size + KPW + c) = 0;
	if ( bad_heap() ) heap_dump();
	b2 = (unsigned char *)realloc(b, size + KPW + KPW);
	sprintf(s, "%08x : vrealloc size = %d\n", (unsigned)b2, size);
	trace(s);
	m[x].addr = b2;
	m[x].size = size;
	for (c = 0; c < KPW; ++c)
		*(m[x].addr + m[x].size + KPW + c) = KP;
	return (void *)(b2 + KPW);
}

#if	defined(ANSI)	/* ANSI-C */
void vdump(char *id)
#else	/* old style */
void vdump(id)
char *id;
#endif
{
	char s[80];
	int x;
	/* dump the entire malloc heap to the vmalloc file */
	sprintf(s, "========== Dump of malloc heap (%s) ==========\n", id);
	trace(s);
	for (x = 0; x < nummallocs; ++x)
		if (m[x].addr != (unsigned char *)0) {
			sprintf(s, "===== Malloc buffer addr : %08x\n",
				(unsigned)m[x].addr);
			trace(s);
			sprintf(s, "===== Malloc buffer size : %08x\n",
				m[x].size + KPW + KPW);
			trace(s);
			dumpbuf(x);
		}
}
