#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <ctype.h>
/*----------------------------------------------------------------------*/
#if	defined(BSD)	/* BSD-UNIX */
#include <sys/types.h>	/* size_t defined */
#include <strings.h>	/* string header */
#else	/* not BSD-UNIX */
#include <string.h>	/* string header */
#endif
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>	/* errno declared */
#endif
/*----------------------------------------------------------------------*/
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "wild.h"	/* function prototypes for wild.c */
/*----------------------------------------------------------------------*/
#if	!(defined(BSD) || defined(SYSV))	/* not UNIX-C */
#if	defined(strcmp)
#undef	strcmp
#endif
#if	defined(strncmp)
#undef	strncmp
#endif
#define	strcmp	StrCmp	/* MS-DOS is case insensitive */
#define	strncmp	StrNCmp
#endif
/*----------------------------------------------------------------------*/
#define STRSIZE	32	/* size of string containing wild card */
/*----------------------------------------------------------------------*/
#if	!(defined(BSD) || defined(SYSV))	/* MS-DOS is case insensitive */
#if	defined(ANSI)	/* ANSI-C */
static int StrCmp(const char *s, const char *t)
#else	/* old style */
static int StrCmp(s, t)	/* case insensitive compare */
const char *s;
const char *t;
#endif
{
	for (; toupper(*s) == toupper(*t); s++, t++)
		if (*s == '\0')
			return 0;
	if (toupper((int)*s) > toupper((int)*t))
		return 1;
	else
		return -1;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static int StrNCmp(const char *s, const char *t, size_t n)
#else	/* old style */
static int StrNCmp(s, t, n)	/* case insensitive compare */
const char *s;
const char *t;
size_t n;
#endif
{
	size_t i;
	for (i = 1; toupper(*s) == toupper(*t); s++, t++, i++)
		if (*s == '\0' || i == n)
			return 0;
	if (toupper((int)*s) > toupper((int)*t))
		return 1;
	else
		return -1;
}
#endif
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static int WildCard(char *w, const char *s)
#else	/* old style */
static int WildCard(w, s)	/* find type of wild card, and delete '*' */
char *w;
const char *s;
#endif
{
	int n = 0;
	if (*s == '*') {
		s++;
		n = 2;
	}
	while (*s != '*' && *s != '\0')
		*w++ = *s++;
	*w = *s;
	if (*s == '*') {
		*w = '\0';
		n++;
	}
	return n;
}
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int wildcmp(const char *s, const char *t)
#else	/* old style */
int wildcmp(s, t)	/* compare string s with wild card w */
const char *s;
const char *t;
#endif
{
	int i, m, n, cmp = 1;
	char w[STRSIZE];
	switch (WildCard(w, t)) {
		case 0:
			cmp = strcmp(s, w); break;
		case 1:
			cmp = strncmp(s, w, strlen(w)); break;
		case 2:
			cmp = strcmp(s+strlen(s)-strlen(w), w); break;
		case 3:
			if ((m = strlen(s)) < (n = strlen(w)))
				cmp = strcmp(s, w);
			else
				for (i = 0; i <= m-n; i++)
					if ((cmp = strncmp(s+i, w, n)) == 0)
						break;
			break;
	}
	return cmp;
}
/*----------------------------------------------------------------------*/
/*
levdist():
					Determines to what extend two character strings
					are (un)equal using the 'Levenstein distance'
					algorithm.

Limitations:
					For memory space conversation and execution speed
					this implementation compares the first 10
					characters of both strings only if longer strings
					are supplied to levdist().
					'#define COMP_LEN' may be changed to decrease or
					increase the number of characters compared.

*/
/*----------------------------------------------------------------------*/
#define ADDITION	1	/* change constants in this block */
#define CHANGE		3	/* of four lines if needed */
#define DELETION	5
#define COMP_LEN	10
/*----------------------------------------------------------------------*/
#define ARR_SIZE	(COMP_LEN + 1)
#define TU(x)		toupper(x)
#define ZERO_IF_EQUAL(x,y) (TU(requested[x-1])==TU(found[y-1])? 0: CHANGE)
/*----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int levdist(const char *requested, const char *found)
#else	/* old style */
int levdist(requested, found)
const char *requested;
const char *found;
#endif
{
	int j, itmp, idist, iprev;
	int i, r_len, f_len, distance[ARR_SIZE];

	if ((r_len = strlen(requested)) > COMP_LEN)
		r_len = COMP_LEN;
	if ((f_len = strlen(found)) > COMP_LEN)
		f_len = COMP_LEN;

	distance[0] = itmp = 0;

	for (j = 1; j < ARR_SIZE; j++)
		distance[j] = (itmp += ADDITION);

	for (i = 1; i <= r_len; i++) {
		iprev = distance[0] + DELETION;
		for (j = 1; j <= f_len; j++, iprev = idist) {
			idist = iprev + ADDITION;
			if ((itmp = distance[j-1] + ZERO_IF_EQUAL(i,j)) < idist)
				idist = itmp;
			distance[j-1] = iprev;
			if ((itmp = distance[j] + DELETION) < idist)
				idist = itmp;
		}
		distance[f_len] = idist;
	}
	return idist;
}
/*----------------------------------------------------------------------*/
