#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#if	!defined(PI)
#define	PI	3.141592653589793
#endif
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int transfer = 0;	/* set Transfer Function? */
	int subtract = 0;	/* subtract PSF? */
	double xc, yc, amp1, amp2, rad1, rad2;
	int type, npix, nlin, xcen, ycen, rmax;
	void *buffer;
	char *image;
	int flag, id, rw = 2;	/* read and write */

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "addpsf: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "ts")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 't':	transfer++; break;
			case 's':	subtract++; break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 7) {
		printf("Usage: addpsf -t -s image Xc Yc A1 A2 R1 R2\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "addpsf: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	sscanf(arg[1], "%lf", &xc);
	sscanf(arg[2], "%lf", &yc);
	sscanf(arg[3], "%lf", &amp1);
	sscanf(arg[4], "%lf", &amp2);
	sscanf(arg[5], "%lf", &rad1);
	sscanf(arg[6], "%lf", &rad2);

	if (rad1 < 1E-9 || rad2 < 1E-9) {
		fprintf(stderr, "addpsf: rad1 = %g  rad2 = %g\n", rad1, rad2);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((flag = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "addpsf: prolog: Error = %d\n", flag);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	setbufsiz(4096);
	if ((id = iopen(image, rw, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "addpsf: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "addpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 5) {
		fprintf(stderr, "addpsf: Type = %d is not implemented\n", type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "addpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((buffer = (void *)calloc(npix, bytepix(type))) == (void *)0) {
		fprintf(stderr, "addpsf: allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"addpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	xcen = (int) (xc + 0.5);
	ycen = (int) (yc + 0.5);

	if (transfer) {	/* add Optical Transfer Function */
		rad1 = nlin/rad1;
		rad2 = nlin/rad2;
	}

	rmax = (int) (4.0*rad2 + 1.0);
	rad1 *= rad1;
	rad2 *= rad2;

	if (!transfer) {	/* add Point Spread Function */
		amp1 /= rad1;
		amp2 /= rad2;
	}

	switch(type) {
		case 1:
			{
				long rmax2 = (long)rmax * rmax;
				UCHAR *pix = (UCHAR *)buffer;
				int y;
				for (y = 0; y < nlin; ++y) {
					long y2 = y - ycen;
					int x;
					y2 *= y2;
					getline(id, y, buffer);
					for (x = 0; x < npix; ++x) {
						long r2 = x - xcen;
						r2 *= r2;
						r2 += y2;
						if (r2 <= rmax2) {
							double dx = x - xc;
							double dy = y - yc;
							double r, d;
							r = -PI*(dx*dx + dy*dy);
							d = amp1*exp(r/rad1) + amp2*exp(r/rad2);
							pix[x] += (d < 0.0)? 0: (UCHAR)(d + 0.5);
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 2:
			{
				long rmax2 = (long)rmax * rmax;
				short *pix = (short *)buffer;
				int y;
				for (y = 0; y < nlin; ++y) {
					long y2 = y - ycen;
					int x;
					y2 *= y2;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						long r2 = x - xcen;
						r2 *= r2;
						r2 += y2;
						if (r2 <= rmax2) {
							double dx = x - xc;
							double dy = y - yc;
							double r, d;
							r = -PI*(dx*dx + dy*dy);
							d = amp1*exp(r/rad1) + amp2*exp(r/rad2);
							if (subtract) d = -d;
							pix[x] += (short)((d < 0.0)? d - 0.5: d + 0.5);
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 3:
			{
				long rmax2 = (long)rmax * rmax;
				float *pix = (float *)buffer;
				int y;
				for (y = 0; y < nlin; ++y) {
					long y2 = y - ycen;
					int x;
					y2 *= y2;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						long r2 = x - xcen;
						r2 *= r2;
						r2 += y2;
						if (r2 <= rmax2) {
							double dx = x - xc;
							double dy = y - yc;
							double r, d;
							r = -PI*(dx*dx + dy*dy);
							d = amp1*exp(r/rad1) + amp2*exp(r/rad2);
							if (subtract) d = -d;
							pix[x] += (float) d;
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 4:
			{
				long rmax2 = (long)rmax * rmax;
				LONG *pix = (LONG *)buffer;
				int y;
				for (y = 0; y < nlin; ++y) {
					long y2 = y - ycen;
					int x;
					y2 *= y2;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						long r2 = x - xcen;
						r2 *= r2;
						r2 += y2;
						if (r2 <= rmax2) {
							double dx = x - xc;
							double dy = y - yc;
							double r, d;
							r = -PI*(dx*dx + dy*dy);
							d = amp1*exp(r/rad1) + amp2*exp(r/rad2);
							if (subtract) d = -d;
							pix[x] += (LONG)((d < 0.0)? d - 0.5: d + 0.5);
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 5:
			{
				long rmax2 = (long)rmax * rmax;
				COMPLEX *pix = (COMPLEX *)buffer;
				int y;
				for (y = 0; y < nlin; ++y) {
					long y2 = y - ycen;
					int x;
					y2 *= y2;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						long r2 = x - xcen;
						r2 *= r2;
						r2 += y2;
						if (r2 <= rmax2) {
							double dx = x - xc;
							double dy = y - yc;
							double r, d;
							r = -PI*(dx*dx + dy*dy);
							d = amp1*exp(r/rad1) + amp2*exp(r/rad2);
							if (subtract) d = -d;
							pix[x].x += (float) d;
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
	}
	free(buffer);
	iclose(id);
	sethistory(image, "addpsf");
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "addpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
