#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "daa.h"	/* function declarations for daa.c */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define M1	259200L
#define IA1	7141L
#define IC1	54773L
#define RM1	(1.0/M1)
#define M2	134456L
#define IA2	8121L
#define IC2	28411L
#define RM2	(1.0/M2)
#define M3	243000L
#define IA3	4561L
#define IC3	51349L
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double ran1(int *idum)
#else	/* old style */
double ran1(idum)
int *idum;
#endif
{
	static long ix1, ix2, ix3;
	static double r[98];
	double temp;
	static int iff = 0;
	int j;
	if ((*idum) < 0 || iff == 0) {
		iff = 1;
		ix1 = (IC1 - (*idum)) % M1;
		ix1 = (IA1*ix1 + IC1) % M1;
		ix2 = ix1 % M2;
		ix1 = (IA1*ix1 + IC1) % M1;
		ix3 = ix1 % M3;
		for (j = 1; j <= 97; j++) {
			ix1 = (IA1*ix1 + IC1) % M1;
			ix2 = (IA2*ix2 + IC2) % M2;
			r[j] = (ix1 + ix2*RM2)*RM1;
		}
		(*idum) = 1;
	}
	ix1 = (IA1*ix1 + IC1) % M1;
	ix2 = (IA2*ix2 + IC2) % M2;
	ix3 = (IA3*ix3 + IC3) % M3;
	j = 1 + (int) ((97*ix3)/M3);
	if (j > 97 || j < 1) {
		fprintf(stderr, "RAN1: This cannot happen.");
		exit(1);
	}
	temp = r[j];
	r[j] = (ix1 + ix2*RM2)*RM1;
	return temp;
}
#undef M1
#undef IA1
#undef IC1
#undef RM1
#undef M2
#undef IA2
#undef IC2
#undef RM2
#undef M3
#undef IA3
#undef IC3
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double gasdev(int *idum)
#else	/* old style */
double gasdev(idum)
int *idum;
#endif
{
	static int iset = 0;
	static double gset;
	double fac, r, v1, v2;
	if  (iset == 0) {
		do {
			v1 = 2.0*ran1(idum) - 1.0;
			v2 = 2.0*ran1(idum) - 1.0;
			r = v1*v1 + v2*v2;
		} while (r >= 1.0);
		fac = sqrt(-2.0*log(r)/r);
		gset = v1*fac;
		iset = 1;
		return v2*fac;
	} else {
		iset = 0;
		return gset;
	}
}
/*--------------------------------------------------------------------*/
#define	EPDN	2.6	/* 1 ADU = 2.6e */
#define	ERON	50.0	/* readout noise as equivalent photoelectrons */
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	double gpow2, gpow = 2.0;	/* Pseudo-Gaussian exponent */
	int i, y, idum = (-13);
	int xcen, ycen, radmax;
	int stx, sty, extx, exty;
	int id, type, npix, nlin;
	double Blank, star, dy2, scale, fac, top, xc, yc, fwhm;
	void *buffer;
	char *image, *ptr;
	float **a;
	unsigned int err;
	int start[2], dim[2];
	const float FltBlank = (float)FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);
	const float Half = (float)0.5;
	const float Zero = (float)0;
	float Init_Value = FltBlank;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "addstar: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "g:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'g':	sscanf(optarg, "%lf", &gpow); break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 5) {
		printf("Usage: addstar -g pow image top Xc Yc fwhm\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);	/* image name */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "addstar: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}
	sscanf(arg[1], "%lf", &top);
	if (top < 0.0) {
		fprintf(stderr, "addstar: top < 0\n");
		exit(EXIT_FAILURE);
	}
	sscanf(arg[2], "%lf", &xc);
	if (xc < 0.0) {
		fprintf(stderr, "addstar: xc < 0\n");
		exit(EXIT_FAILURE);
	}
	sscanf(arg[3], "%lf", &yc);
	if (yc < 0.0) {
		fprintf(stderr, "addstar: yc < 0\n");
		exit(EXIT_FAILURE);
	}
	sscanf(arg[4], "%lf", &fwhm);	/* FWHM of star */
	if (fwhm < 0.5) {
		fprintf(stderr, "addstar: fwhm < 0.5\n");
		exit(EXIT_FAILURE);
	}
	gpow2 = gpow/2.0;
	scale = pow(2.0, gpow)*log(2.0);
	fac = pow(log(1E20)/scale, 1.0/gpow);
	xcen = (int) (xc + 0.5);
	ycen = (int) (yc + 0.5);
	radmax = (int) (fac*fwhm + 0.5);
	extx = 2*radmax + 1;
	exty = extx;
	stx = xcen - radmax;
	if (stx < 0) {
		extx += stx;
		stx = 0;
	}
	sty = ycen - radmax;
	if (sty < 0) {
		exty += sty;
		sty = 0;
	}
	if (extx <= 0 || exty <= 0) {
		fprintf(stderr,
			"addstar: window is placed above/left of image\n");
		exit(EXIT_FAILURE);
	}
	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "addstar: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	if ((id = iopen(image, 0, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "addstar: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"addstar: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 4) {	/* COMPLEX not allowed */
		fprintf(stderr,
			"addstar: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"addstar: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank = blank(image);
	if (stx >= npix || sty >= nlin) {
		fprintf(stderr,
			"addstar: window is placed below/right of image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"addstar: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (stx + extx > npix)
		extx = npix - stx;
	if (sty + exty > nlin)
		exty = nlin - sty;
	if (!window(id, stx, sty, extx, exty)) {
		fprintf(stderr, "addstar: Object window is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"addstar: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((buffer = (void *)calloc(extx, bytepix(type))) == (void *)0) {
		fprintf(stderr, "addstar: Allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"addstar: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	dim[0] = exty;	/* y dimension of object window */
	dim[1] = extx;	/* x dimension of object window */
	start[0] = sty;	/* first index of y dimension */
	start[1] = stx;	/* first index of x dimension */
	if ((a = (float **)daa(sizeof(float), 2, dim, start,
			&err, &ptr, &Init_Value)) == (float **)0) {
		fprintf(stderr, "addstar: daa: Error code = %d\n", err);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"addstar: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (y = sty; y < (sty+exty); y++) {	/* read window into a[][] */
		getline(id, y-sty, buffer);	/* fetch new line */
		switch (type) {
			case 1:	/* convert unsigned char to float */
				{
					const UCHAR *q = (UCHAR *)buffer;
					int x;
					for (x = stx, i = 0; i < extx; x++, i++)
						a[y][x] = (float)q[i];
				}
				break;
			case 2:	/* convert short int to float */
				{
					const short *q = (short *)buffer;
					const short b = (short)Blank;
					int x;
					for (x = stx, i = 0; i < extx; x++, i++)
						a[y][x] = (q[i] == b)? FltBlank: (float)q[i];
				}
				break;
			case 3:	/* float */
				{
					const float *q = (float *)buffer;
					const float h = (float) (0.9*Blank);
					int x;
					for (x = stx, i = 0; i < extx; x++, i++)
						a[y][x] = (q[i] > h)? FltBlank: q[i];
				}
				break;
			case 4:	/* convert LONG int to float */
				{
					const LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank;
					int x;
					for (x = stx, i = 0; i < extx; x++, i++)
						a[y][x] = (q[i] == b)? FltBlank: (float)q[i];
				}
				break;
		}
	}
	for (y = sty; y < (sty+exty); y++) {
		double dr2;
		int x;
		dy2 = (y - yc)/fwhm;
		dy2 *= dy2;
		for (x = stx; x < (stx+extx); x++)
			if (a[y][x] <= FltMax) {
				dr2 = (x - xc)/fwhm;
				dr2 *= dr2;
				dr2 += dy2;
				star = top*exp(-scale*pow(dr2, gpow2));
				a[y][x] += (float)star;	/* add star */
				a[y][x] += (float)(sqrt(EPDN*star+ERON)*gasdev(&idum)/EPDN);
			}
	}
	setbufsiz(4096);	/* change buffer size */
	if (!chiomod(id, 2)) {	/* change IO mode to update */
		fprintf(stderr, "addstar: Can't change IO mode for %s\n", image);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "addstar: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (y = sty; y < (sty+exty); y++) {	/* write window back */
		switch (type) {
			case 1:	/* convert float to unsigned char */
				{
					UCHAR *q = (UCHAR *)buffer;
					int x;
					for (x = stx, i = 0; i < extx; x++, i++)
						q[i] = (a[y][x] < Zero)? 0: (UCHAR)(a[y][x] + Half);
				}
				break;
			case 2:	/* convert float to short int */
				{
					short *q = (short *)buffer;
					const short b = (short)Blank;
					int x;
					for (x = stx, i = 0; i < extx; x++, i++)
						q[i] = (a[y][x] > FltMax)? b:
								(short)((a[y][x]<Zero)? a[y][x]-Half: a[y][x]+Half);
				}
				break;
			case 3:	/* float */
				{
					float *q = (float *)buffer;
					const float b = (float)Blank;
					int x;
					for (x = stx, i = 0; i < extx; x++, i++)
						q[i] = (a[y][x] > FltMax)? b: a[y][x];
				}
				break;
			case 4:	/* convert float to LONG int */
				{
					LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank;
					int x;
					for (x = stx, i = 0; i < extx; x++, i++)
						q[i] = (a[y][x] > FltMax)? b:
								(LONG)((a[y][x]<Zero)? a[y][x]-Half: a[y][x]+Half);
				}
				break;
		}
		putline(id, y-sty, buffer);	/* write line buffer back */
	}
	free(ptr);
	free(buffer);
	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "addstar: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
