#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "daa.h"	/* function declarations for daa.c */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define M1	259200L
#define IA1	7141L
#define IC1	54773L
#define RM1	(1.0/M1)
#define M2	134456L
#define IA2	8121L
#define IC2	28411L
#define RM2	(1.0/M2)
#define M3	243000L
#define IA3	4561L
#define IC3	51349L
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double ran1(int *idum)
#else	/* old style */
double ran1(idum)
int *idum;
#endif
{
	static long ix1, ix2, ix3;
	static double r[98];
	double temp;
	static int iff = 0;
	int j;
	if ((*idum) < 0 || iff == 0) {
		iff = 1;
		ix1 = (IC1 - (*idum)) % M1;
		ix1 = (IA1*ix1 + IC1) % M1;
		ix2 = ix1 % M2;
		ix1 = (IA1*ix1 + IC1) % M1;
		ix3 = ix1 % M3;
		for (j = 1; j <= 97; j++) {
			ix1 = (IA1*ix1 + IC1) % M1;
			ix2 = (IA2*ix2 + IC2) % M2;
			r[j] = (ix1 + ix2*RM2)*RM1;
		}
		(*idum) = 1;
	}
	ix1 = (IA1*ix1 + IC1) % M1;
	ix2 = (IA2*ix2 + IC2) % M2;
	ix3 = (IA3*ix3 + IC3) % M3;
	j = 1 + (int) ((97*ix3)/M3);
	if (j > 97 || j < 1) {
		fprintf(stderr, "RAN1: This cannot happen.");
		exit(1);
	}
	temp = r[j];
	r[j] = (ix1 + ix2*RM2)*RM1;
	return temp;
}
#undef M1
#undef IA1
#undef IC1
#undef RM1
#undef M2
#undef IA2
#undef IC2
#undef RM2
#undef M3
#undef IA3
#undef IC3
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double gasdev(int *idum)
#else	/* old style */
double gasdev(idum)
int *idum;
#endif
{
	static int iset = 0;
	static double gset;
	double fac, r, v1, v2;
	if  (iset == 0) {
		do {
			v1 = 2.0*ran1(idum) - 1.0;
			v2 = 2.0*ran1(idum) - 1.0;
			r = v1*v1 + v2*v2;
		} while (r >= 1.0);
		fac = sqrt(-2.0*log(r)/r);
		gset = v1*fac;
		iset = 1;
		return v2*fac;
	} else {
		iset = 0;
		return gset;
	}
}
/*--------------------------------------------------------------------*/
#define	EPDN	2.6	/* 1 ADU = 2.6e */
#define	ERON	50.0	/* readout noise as equivalent photoelectrons */
#define	SCALE	7.67
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int i, y, idum = (-13);
	int id, type, npix, nlin;
	double Blank, co, si, ae, be, xc, yc, ie, re, q;
	double pa = 0.0;	/* position angle of major axis */
	double eps = 0.0;	/* default ellipticity */
	void *buffer;
	char *image, *ptr;
	float **a;
	unsigned int err;
	int start[2], dim[2];
	const float FltBlank = (float)FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);
	const float Half = (float)0.5;
	const float Zero = (float)0;
	float Init_Value = FltBlank;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if (argc < 6) {
		printf("Usage: adgalaxy image Xc Yc Ie Re [PA] [eps]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}
	image = truncname(argv[1]);	/* image name */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "adgalaxy: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}
	sscanf(argv[2], "%lf", &xc);
	if (xc < 0.0) {
		fprintf(stderr, "adgalaxy: xc < 0\n");
		exit(EXIT_FAILURE);
	}
	sscanf(argv[3], "%lf", &yc);
	if (yc < 0.0) {
		fprintf(stderr, "adgalaxy: yc < 0\n");
		exit(EXIT_FAILURE);
	}
	sscanf(argv[4], "%lf", &ie);
	if (ie < 0.0) {
		fprintf(stderr, "adgalaxy: Ie < 0\n");
		exit(EXIT_FAILURE);
	}
	sscanf(argv[5], "%lf", &re);
	if (re < 2.0) {
		fprintf(stderr, "adgalaxy: Re < 2\n");
		exit(EXIT_FAILURE);
	}
	if (argc > 6) {
		sscanf(argv[6], "%lf", &pa);
		pa *= atan(1.0)/45.0;
	}
	if (argc > 7) {
		sscanf(argv[7], "%lf", &eps);
		if (eps < 0.0 || eps >= 1.0) {
			fprintf(stderr, "adgalaxy: eps < 0 or eps >= 1\n");
			exit(EXIT_FAILURE);
		}
	}
	q = 1.0 - eps;	/* ratio of the minor to the major axis */

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "adgalaxy: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	if ((id = iopen(image, 0, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "adgalaxy: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"adgalaxy: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 4) {	/* COMPLEX not allowed */
		fprintf(stderr,
			"adgalaxy: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"adgalaxy: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank = blank(image);
	if ((buffer = (void *)calloc(npix, bytepix(type))) == (void *)0) {
		fprintf(stderr, "adgalaxy: Allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"adgalaxy: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	dim[0] = nlin;	/* y dimension */
	dim[1] = npix;	/* x dimension */
	start[0] = 0;	/* first index of y-dimension */
	start[1] = 0;	/* first index of x-dimension */
	if ((a = (float **)daa(sizeof(float), 2, dim, start,
			&err, &ptr, &Init_Value)) == (float **)0) {
		fprintf(stderr, "adgalaxy: daa: Error code = %d\n", err);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"adgalaxy: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (y = 0; y < nlin; ++y) {	/* read image into a[][] */
		getline(id, y, buffer);	/* fetch new line */
		switch (type) {
			case 1:	/* convert unsigned char to float */
				{
					const UCHAR *q = (UCHAR *)buffer;
					int x;
					for (x = 0; x < npix; ++x)
						a[y][x] = (float)q[x];
				}
				break;
			case 2:	/* convert short int to float */
				{
					const short *q = (short *)buffer;
					const short b = (short)Blank;
					int x;
					for (x = 0; x < npix; ++x)
						a[y][x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 3:	/* float */
				{
					const float *q = (float *)buffer;
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < npix; ++x)
						a[y][x] = (q[x] > h)? FltBlank: q[x];
				}
				break;
			case 4:	/* convert LONG int to float */
				{
					const LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank;
					int x;
					for (x = 0; x < npix; ++x)
						a[y][x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
		}
	}
	ae = re/sqrt(q);
	be = re*sqrt(q);
	co = cos(pa);
	si = sin(pa);
	for (y = 0; y < nlin; ++y) {
		double dy_si = (y - yc)*si;
		double dy_co = (y - yc)*co;
		double dx, dy, g;
		int x;
		for (x = 0; x < npix; ++x)
			if (a[y][x] <= FltMax) {
				dx = ((x - xc)*co + dy_si)/ae;
				dy = (dy_co - (x - xc)*si)/be;
				g = ie*exp(-SCALE*(sqrt(sqrt(sqrt(dx*dx+dy*dy)))-1.0));
				a[y][x] += (float)(g+sqrt(EPDN*g+ERON)*gasdev(&idum)/EPDN);
			}
	}
	setbufsiz(8192);	/* change buffer size */
	if (!chiomod(id, 2)) {	/* change IO mode to update */
		fprintf(stderr, "adgalaxy: Can't change IO mode for %s\n", image);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "adgalaxy: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (y = 0; y < nlin; ++y) {	/* write image back from a[][] */
		switch (type) {
			case 1:	/* convert float to unsigned char */
				{
					UCHAR *q = (UCHAR *)buffer;
					int x;
					for (x = 0; x < npix; ++x)
						q[x] = (a[y][x] < Zero)? 0: (UCHAR)(a[y][x] + Half);
				}
				break;
			case 2:	/* convert float to short int */
				{
					short *q = (short *)buffer;
					const short b = (short)Blank;
					int x;
					for (x = 0; x < npix; ++x)
						q[x] = (a[y][x] > FltMax)? b:
								(short)((a[y][x]<Zero)? a[y][x]-Half: a[y][x]+Half);
				}
				break;
			case 3:	/* float */
				{
					float *q = (float *)buffer;
					const float b = (float)Blank;
					int x;
					for (x = 0; x < npix; ++x)
						q[x] = (a[y][x] > FltMax)? b: a[y][x];
				}
				break;
			case 4:	/* convert float to LONG int */
				{
					LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank;
					int x;
					for (x = 0; x < npix; ++x)
						q[x] = (a[y][x] > FltMax)? b:
								(LONG)((a[y][x]<Zero)? a[y][x]-Half: a[y][x]+Half);
				}
				break;
		}
		putline(id, y, buffer);	/* write line buffer */
	}
	free(ptr);
	free(buffer);
	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "adgalaxy: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
