#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define MAX(a,b)  (((a) > (b)) ? (a) : (b))
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv, **arg;
	int option, opt_err = 0, n_arg = 0;
	int m = 0;	/* running mean bias over rows [y-m,y+m] */

	int i, n, x, y;
	int x_min, x_max, y_min, y_max, b_min, b_max, nbias, nrow;
	int id1, type1, npix1, nlin1, rw1 = 0;	/* read only */
	int id2, type2, npix2, nlin2, rw2 = 1;	/* write only */
	double sum, mean, Blank1, Blank2;
	char *im1, *im2;
	double *col;	/* column */
	double *bias;	/* bias column */
	short *row;	/* input row */
	void  *buf;	/* output row */
	const double FltMax = 0.9*FLT_BLANK;
	const float Zero = (float)0;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "bias: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	/* process command line options */
	while (((option = getopt(argc, argv, "m:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'm':	sscanf(optarg, "%d", &m); break;
			case '?':	opt_err++; break;	/* unknown option */
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	/* illegal options */
	if (opt_err || n_arg < 2) {
		printf(
		"bias -m<pts> image1 image2 [Bmin] [Bmax] [Xmin] [Xmax] [type] [Ymin] [Ymax]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	im1 = truncname(arg[0]);	/* input image */
	if (strpbrk(im1, punct) != (char *)0) {
		fprintf(stderr, "bias: Illegal image name: %s\n", im1);
		exit(EXIT_FAILURE);
	}

	im2 = truncname(arg[1]);	/* output image */
	if (strpbrk(im2, punct) != (char *)0) {
		fprintf(stderr, "bias: Illegal image name: %s\n", im2);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "bias: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */
	if ((id1 = iopen(im1, rw1, &type1, &npix1, &nlin1)) < 0) {
		fprintf(stderr, "bias: can't open %s\n", im1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type1 != 2) {
		fprintf(stderr, "bias: %s has illegal type = %d\n", im1, type1);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank1 = blank(im1);

	if (n_arg > 2) {
		sscanf(arg[2], "%d", &b_min);
		if (b_min < 0) b_min = 0;
	} else b_min = 0;

	if (n_arg > 3) {
		sscanf(arg[3], "%d", &b_max);
		if (b_max >= npix1) b_max = npix1 - 1;
	} else b_max = npix1 - 1;

	if (n_arg > 4) {
		sscanf(arg[4], "%d", &x_min);
		if (x_min < 0) x_min = 0;
	} else x_min = 0;

	if (n_arg > 5) {
		sscanf(arg[5], "%d", &x_max);
		if (x_max >= npix1) x_max = npix1 - 1;
	} else x_max = npix1 - 1;

	if (n_arg > 6)
		sscanf(arg[6], "%d", &type2);
	else type2 = type1;

	if (n_arg > 7) {
		sscanf(arg[7], "%d", &y_min);
		if (y_min < 0) y_min = 0;
	} else y_min = 0;

	if (n_arg > 8) {
		sscanf(arg[8], "%d", &y_max);
		if (y_max >= nlin1) y_max = nlin1 - 1;
	} else y_max = nlin1 - 1;

	if (b_max < b_min) {
		fprintf(stderr,
			"bias: Illegal range [%d:%d] of bias columns\n", b_min, b_max);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (x_max < x_min) {
		fprintf(stderr,
			"bias: Illegal range [%d:%d] of trim columns\n", x_min, x_max);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type2 < 2 || type2 > 5) {
		fprintf(stderr, "bias: %s has illegal type = %d\n", im2, type2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (y_max < y_min) {
		fprintf(stderr,
			"bias: Illegal range [%d:%d] of trim rows\n", y_min, y_max);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	nbias = b_max - b_min + 1;	/* No. of bias pixels per row */
	npix2 = x_max - x_min + 1;	/* No. of trim pixels per row */
	nlin2 = y_max - y_min + 1;	/* No. of trim rows */
	nrow = MAX(nbias, npix2);

	if (!window(id1, b_min, 0, nbias, nlin1)) {
		fprintf(stderr, "bias: Bias strip is not inside image\n");
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate input row buffer */
	if ((row = (short *)calloc(nrow, sizeof(short))) == (short *)0) {
		fprintf(stderr, "bias: Can't allocate row[%d]\n", nrow);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate output row buffer */
	if ((buf = (void *)calloc(npix2, bytepix(type2))) == (void *)0) {
		fprintf(stderr, "bias: Can't allocate buf[%d]\n", npix2);
		free(row);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate bias column */
	if ((bias = (double *)calloc(nlin1, sizeof(double))) == (double *)0) {
		fprintf(stderr, "bias: Can't allocate bias[%d]\n", nlin1);
		free(buf);
		free(row);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate column buffer */
	if ((col = (double *)calloc(nlin1, sizeof(double))) == (double *)0) {
		fprintf(stderr, "bias: Can't allocate col[%d]\n", nlin1);
		free(bias);
		free(buf);
		free(row);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* calculate a column of bias values from the bias strip */
	for (y = 0; y < nlin1; ++y) {
		const short b1 = (short) Blank1;
		getline(id1, y, row);	/* read next bias row */
		sum = 0.0;
		for (n = 0, x = 0; x < nbias; ++x)
			if (row[x] != b1) {
				sum += (double) row[x];
				++n;
			}
		col[y] = (n == 0)? FLT_BLANK: sum/n;
	}

	sum = 0.0;
	/* calculate the mean bias, excluding the first row */
	for (n = 0, i = 1; i < nlin1; ++i)
		if (col[i] <= FltMax) {
			sum += col[i];
			++n;
		}
	mean = (n == 0)? 0.0: sum/n;

	/* calculate an average bias value for each row */
	if (m < 0)	/* use the mean bias for all rows */
		for (y = 0; y < nlin1; ++y)
			bias[y] = mean;
	else	/* use a running mean centred on the given row */
		for (y = 0; y < nlin1; ++y) {
			/* do not include the first row in running mean */
			int k = MIN(y-1,nlin1-y-1);
			k = MAX(k,0);	/* take care of first row */
			k = MIN(m,k);
			sum = 0.0;
			for (n = 0, i = (y-k); i <= (y+k); ++i)
				if (col[i] <= FltMax) {
					sum += col[i];
					++n;
				}
			bias[y] = (n == 0)? mean: sum/n;
		}
	free(col);

	if (!window(id1, x_min, 0, npix2, nlin1)) {
		fprintf(stderr, "bias: Trim section is not inside image\n");
		free(bias);
		free(buf);
		free(row);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (xist(im2)) {	/* does output image exist? */
		fprintf(stderr, "bias: %s: Existing image\n", im2);
		free(bias);
		free(buf);
		free(row);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "runmed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (cpi(im1, im2) < 1) {	/* dublicate image header */
		fprintf(stderr, "bias: %s.hdr -> %s.hdr copy failed\n", im1, im2);
		free(bias);
		free(buf);
		free(row);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (setimagefile(im2, "") == 0) {	/* erase FILE key */
		fprintf(stderr, "bias: Can't erase old FILE value for %s\n", im2);
		rmi(im2);	/* remove image header */
		free(bias);
		free(buf);
		free(row);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((id2 = icreat(im2, rw2, type2, npix2, nlin2)) < 0) {
		fprintf(stderr, "bias: Can't create %s\n", im2);
		rmi(im2);	/* remove image header */
		free(bias);
		free(buf);
		free(row);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "bias: update: Disk header not created\n");

	Blank2 = blank(im2);

	for (y = y_min; y <= y_max; ++y) { /* read only trim rows */
		const short b1 = (short) Blank1;
		getline(id1, y, row);	/* read an image row from disk */
		switch (type2) {
			case 2:
			{
				short *q = (short *)buf;
				const short b2 = (short) Blank2;
				const short c = (short)((bias[y]<0.0)? bias[y]-0.5: bias[y]+0.5);
				int x;
				for (x = 0; x < npix2; ++x)
					q[x] = (row[x] == b1)? b2: row[x] - c;
			}
			break;
			case 3:
			{
				float *q = (float *)buf;
				const float b2 = (float) Blank2;
				const float c = (float) bias[y];
				int x;
				for (x = 0; x < npix2; ++x)
					q[x] = (row[x] == b1)? b2: (float)row[x] - c;
			}
			break;
			case 4:
			{
				LONG *q = (LONG *)buf;
				const LONG b2 = (LONG) Blank2;
				const LONG c = (LONG)((bias[y]<0.0)? bias[y]-0.5: bias[y]+0.5);
				int x;
				for (x = 0; x < npix2; ++x)
					q[x] = (row[x] == b1)? b2: (LONG)row[x] - c;
			}
			break;
			case 5:
			{
				COMPLEX *q = (COMPLEX *)buf;
				const float b2 = (float) Blank2;
				const float c = (float) bias[y];
				int x, ip;
				for (x = 0; x < npix2; ++x) {
					ip = (row[x] == b1);
					q[x].x = (ip)? b2: (float)row[x] - c;
					q[x].y = (ip)? b2: Zero;
				}
			}
			break;
		}
		putline(id2, y - y_min, buf);	/* write row to disk */
	}

	iclose(id2);
	free(bias);
	free(buf);
	free(row);
	iclose(id1);

	setiname(im2);
	sethistory(im2, "bias");
	sethistory(im2, im2);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "bias: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
