#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <stdlib.h>	/* function declarations */
#if	defined(MSC) || defined(HC2) || defined(WC)
/* MS-C, High-C V2.3, or WATCOM-C 9.0/386 */
#include <dos.h>
#include <conio.h>	/* Console I/O declarations */
#endif
#if	defined(SYSV) || defined(BSD)	/* Linux */
#include <unistd.h>
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "fbuf.h"	/* constants and prototypes for fbuf.c */
#if	defined(HC1)	/* High-C V1.6 */
#include <msdos.cf>
#include "hcconio.h"	/* function prototypes for hcconio.c */
#endif
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#define SHUTTEROPEN  2
#define SHUTTERCLOSE 3
#define BOARD        0x1B0
#define P2_CTRL      0x07
#define P2_A         0x04
#define P2_B         0x05
#define PORTA        (BOARD + P2_A)
#define PORTB        (BOARD + P2_B)
#define CTRL         (BOARD + P2_CTRL)
/*--------------------------------------------------------------------*/
/*
	Initialize parallel board
	Use Port 2 channel A and B
	A for output (bit 0-3)
	B for input (bit 0-1)
*/
void Init_B(void)
{
#if	defined(MSC) || defined(WC)	/* MS-C or WATCOM-C 9.0/386 */
	outp(CTRL, 0x82);
	outp(PORTA, 0xFF);
#endif
}
/*--------------------------------------------------------------------*/
void Perform(int op)	/* open or close shutter */
{
#if	defined(MSC) || defined(WC)	/* MS-C or WATCOM-C 9.0/386 */
	switch (op) {
		case SHUTTEROPEN:  outp(PORTA, 0xFE); break;
		case SHUTTERCLOSE: outp(PORTA, 0xFD); break;
		default: break;
	}
	delay(20);
	outp(PORTA, 0xFF);
#endif
}
/*--------------------------------------------------------------------*/
void Expose(long msec)	/* expose in millisec */
{
#if	defined(MSC) || defined(WC)	/* MS-C or WATCOM-C 9.0/386 */
	long i;
	Perform(SHUTTEROPEN);
	for (i = 0; i < msec; ++i) delay(1);
	Perform(SHUTTERCLOSE);
#endif
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	long msec = 0;	/* default exposure time in msec */
	int frame = 1;	/* default frame buffer number */
	int page = 0;	/* default window page number */

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

#if   defined(SYSV) || defined(BSD)	/* Linux */
   if (iopl(3) != 0) {
		fprintf(stderr, "ccdatn: Can't get I/O permission\n");
		exit(EXIT_FAILURE);
	}
#endif

	if (argc > 1) {
		sscanf(argv[1], "%ld", &msec);	/* exposure time in msec */
		if (msec < 0) {
			fprintf(stderr, "ccdatn: Exposure time = %ld\n", msec);
			exit(EXIT_FAILURE);
		}
	}

	if (argc > 2) {
		sscanf(argv[1], "%d", &frame);	/* frame number */
		if (frame < 1 || frame > 2) {
			fprintf(stderr, "ccdatn: Frame number = %d\n", frame);
			exit(EXIT_FAILURE);
		}
	}

	if (argc > 3) {
		sscanf(argv[2], "%d", &page);	/* page number */
		if (page < 0 || page >= RAM_PAG) {
			fprintf(stderr, "ccdatn: Page number = %d\n", page);
			exit(EXIT_FAILURE);
		}
	}

	if (msec > 0) {
		Init_B();	/* Initialize parallel board */
		Expose(msec);	/* expose CCD in millisec */
	}

	put_cnt(frame, page * PAG_SIZ);	/* load pixel counter */
	ccd_on(frame);	/* enable serial port */

/*
	while (kbhit() == 0)
		printf("%6ld\r", get_cnt(frame));
	ccd_off(frame);
*/

	fflush(stdout);
	exit(EXIT_SUCCESS);
}
