#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
#include <math.h>
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "matalloc.h"	/* function declaration for matalloc.c */
#include "nr.h"	/* function declarations for nr.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define ORDER	10	/* maximum polynomial order */
#define MD	(ORDER + 1)	/* maximum number of parameters */
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void fleg(double x, double *pl, int nl)
#else	/* old style */
void fleg(x, pl, nl)
double x, pl[];
int nl;
#endif
{
	int j;
	double twox, f2, f1, d;

	if (nl > 0)
		pl[0] = 1.0;
	if (nl > 1)
		pl[1] = x;
	if (nl > 2) {
		twox = 2.0*x;
		f2 = x;
		d = 1.0;
		for (j = 2; j < nl; ++j) {
			f1 = d;
			d += 1.0;
			f2 += twox;
			pl[j] = (f2*pl[j-1] - f1*pl[j-2])/d;
		}
	}
}
/*----------------------------------------------------------------------
	'solve' calculates the solution of linear system, A*x = b
	Do not use if 'decomp' has detected singularity
	Input:
      n      Order of matrix.
      *A[]   Triangular matrix A obtained from 'decomp'.
      b[]    Right hand side vector, b.
      ipvt[] Pivot vector obtained from 'decomp'.
	Output:
      b[]    Solution vector, x.
------------------------------------------------------------------------
	G. E. Forsythe, M. A. Malcolm, and C. B. Moler
	Computer Methods for Mathematical Computations, P. 55.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 24-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void solve(int n, double * const A[], double b[], const int ipvt[])
#else	/* old style */
void solve(n, A, b, ipvt)
int n;
double * const A[];
double b[];
const int ipvt[];
#endif
{
	int i, k, m;
	double t;

	if (n > 1) {

		/* forward elimination */
		for (k = 0; k < (n-1); ++k) {
			m = ipvt[k];
			t = b[m];
			b[m] = b[k];
			b[k] = t;
			for (i = (k+1); i < n; ++i)
				b[i] += A[k][i]*t;
		}

		/* back substitution */
		for (k = (n-1); k > 0; --k) {
			b[k] /= A[k][k];
			t = -b[k];
			for (i = 0; i < k; ++i)
				b[i] += A[k][i]*t;
		}
	}
	b[0] /= A[0][0];
}
/*----------------------------------------------------------------------
	'decomp' decomposes a square matrix A by Gaussian elimination and
	estimates the condition of the matrix.
	Use 'solve' to compute solutions to linear systems.

	Input:
      n      Order of square matrix.
      *A[]   The matrix A to be triangularized.

	Output:
		A contains an upper triangular matrix U and a permuted
		version of a lower triangular matrix I-L so that
		(Permutation Matrix)*A = L*U.

	Return:
		An estimate of the 'condition' of A.

		For the linear system A*x = b, changes in A and b MAY
		cause changes 'condition' times as large in x.
		If (condition+1.0 == condition), A is singular to working
		precision. condition is set to 1.0e+32 if exact singularity
		is detected.

      ipvt[]    The pivot vector.
      ipvt[k]   The index of the k-th pivot row.
      ipvt[n-1] (-1)**(Number of Interchanges)

	Work Space:
		The wector work[] must be declared and included in the
		call. Its input contents are ignored. Its output
		contents are usually unimportant.

	The determinant of A can be obtained on output by
		Det(A) = ipvt[n-1]*A[0][0]*A[1][1]*......*A[n-1][n-1]
------------------------------------------------------------------------
	G. E. Forsythe, M. A. Malcolm, and C. B. Moler
	Computer Methods for Mathematical Computations, P. 52.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 24-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double decomp(int n, double *A[], int ipvt[], double work[])
#else	/* old style */
double decomp(n, A, ipvt, work)
int n, ipvt[];
double *A[], work[];
#endif
{
	int i, j, k, m;
	double ek, t, anorm, ynorm, znorm, cond;

	ipvt[n-1] = 1;

	/* 1-by-1 */
	if (n == 1) {
		if (A[0][0] != 0.0)
			cond = 1.0;
		else
			cond = 1.0e+32;
		return cond;
	}

	/* compute 1-norm of A */
	anorm = 0.0;
	for (j = 0; j < n; ++j) {
		t = 0.0;
		for (i = 0; i < n; ++i)
			t += fabs(A[j][i]);
		if (t > anorm)
			anorm = t;
	}

	/* Gaussian elimination with partial pivoting */
	for (k = 0; k < (n-1); ++k) {
		/* find pivot */
		m = k;
		for (i = (k+1); i < n; ++i)
			if (fabs(A[k][i]) > fabs(A[k][m]))
				m = i;
		ipvt[k] = m;
		if(m != k)
			ipvt[n-1] = -ipvt[n-1];
		t = A[k][m];
		A[k][m] = A[k][k];
		A[k][k] = t;

		/* skip step if pivot is zero */
		if (t != 0.0) {

			/* compute multipliers */
			for (i = (k+1); i < n; ++i)
				A[k][i] /= -t;

			/* interchange and eliminate by columns */
			for (j = (k+1); j < n; ++j) {
				t = A[j][m];
				A[j][m] = A[j][k];
				A[j][k] = t;
				if (t != 0.0)
					for (i = (k+1); i < n; ++i)
						A[j][i] += A[k][i]*t;
			}
		}
	}
	/*
	cond = (1-norm of A)*(an estimate of 1-norm of A-inverse)
	Estimate obtained by one step of inverse iteration for the
	small singular vector. This involves solving two systems
	of equations, (A-transpose)*Y = E and A*Z = Y where E is
	a vector of +1 or -1 chosen to cause growth in Y.
	Estimate = (1-norm of Z)/(1-norm of Y)
	*/

	/* solve (A-transpose)*Y = E */
	for (k = 0; k < n; ++k) {
		t = 0.0;
		if (k > 0)
			for (i = 0; i < k; ++i)
				t += A[k][i]*work[i];
		ek = 1.0;
		if (t < 0.0)
			ek = -1.0;
		if (A[k][k] == 0.0) {
			cond = 1.0e+32;
			return cond;
		}
		work[k] = -(ek + t)/A[k][k];
	}
	for (k = (n-2); k >= 0; --k) {
		t = 0.0;
		for (i = (k+1); i < n; ++i)
			t += A[k][i]*work[k];
		work[k] = t;
		m = ipvt[k];
		if (m != k) {
			t = work[m];
			work[m] = work[k];
			work[k] = t;
		}
	}
	ynorm = 0.0;
	for (i = 0; i < n; ++i)
		ynorm += fabs(work[i]);

	/* solve A*Z = Y */
	solve(n, A, work, ipvt);

	znorm = 0.0;
	for (i = 0; i < n; ++i)
		znorm += fabs(work[i]);

	/* estimate condition */
	cond = anorm*znorm/ynorm;
	if (cond < 1.0)
		cond = 1.0;
	return cond;
}
/*----------------------------------------------------------------------
	Finds the Median, or any other Quantile.
	The median of n float items a[i] is defined as the item which is
	less than (or equal to) half of the n items and which is larger
	than (or equal to) the other half of the n items. The method to
	be displayed easily generalizes to the problem of finding the
	kth smallest of n items. Finding the median represents the
	special case k = n/2.
	The algorithm invented by C. A. R. Hoare functions as
	follows. First, the partitioning operation of Quicksort (p. 77)
	is applied with l = 0 and r = (n-1), and with a[k] selected as
	splitting value x. The resulting index values i and j are such that
		1. a[h] <= x for all h < i
		2. a[h] >= x for all h > j
		3. i > j
	There are three possible cases that may arise:
		1. The splitting value x was too small; as a result,
		   the limit between the two partitions is below the
		   desired value k. The partitioning process has to be
		   repeated upon the elements a[i]...a[r].
		2. The chosen bound x was too large. The splitting
		   operation has to be repeated on the partition
		   a[l]...a[j].
		3. j < k < i: the element a[k] splits the array into two
		   partitions in the specified proportions and
		   therefore is the desired quantile.
	The splitting process has to be repeated until case 3. arises.
------------------------------------------------------------------------
	Niklaus Wirth:	Algorithms + Data Structure = Programs, p. 82.
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void find(int n, float a[], int k)
#else	/* old style */
void find(n, a, k)
int n, k;
float a[];
#endif
{
	int l, r, i, j;
	float w, x;

	l = 0;
	r = n - 1;

	while (l < r) {
		x = a[k];
		i = l;
		j = r;
		/* begin split */
		while (i <= j) {
			while (a[i] < x)
				++i;
			while (x < a[j])
				--j;
			if (i <= j) {
				w = a[i];
				a[i] = a[j];
				a[j] = w;
				++i;
				--j;
			}
		}
		/* end split */
		if (j < k)
			l = i;
		if (k < i)
			r = j;
	}
}
/*----------------------------------------------------------------------
	Define the functions rho(t), wgt(t), and phi(t) for various cases.
		wgt(t) = rho'(t)/t; phi(t) = rho''(t).
------------------------------------------------------------------------
	Gaussian Distribution of r[i]/(scale*sig[i]): Ordinary Least Squares
		rho0(t) = 0.5*t^2
		wgt0(t) = 1
		phi0(t) = 1
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 22-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double rho0(double t)
#else	/* old style */
double rho0(t)
double t;
#endif
{
	return (0.5*t*t);
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double one(double t)
#else	/* old style */
double one(t)
double t;
#endif
{
	return (1.0);
}
/*----------------------------------------------------------------------
	Hyperbolic Distribution of r[i]/(scale*sig[i]):
		rho1(t) = sqrt(1 + t^2) - 1 ~ 0.5*t^2
		wgt1(t) = 1/sqrt(1 + t^2)
		phi1(t) = 1/sqrt(1 + t^2)^3
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 22-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double rho1(double t)
#else	/* old style */
double rho1(t)
double t;
#endif
{
	return (sqrt(t*t + 1.0) - 1.0);
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double wgt1(double t)
#else	/* old style */
double wgt1(t)
double t;
#endif
{
	return (1.0/sqrt(t*t + 1.0));
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double phi1(double t)
#else	/* old style */
double phi1(t)
double t;
#endif
{	double z = t*t + 1.0;
	return (1.0/(sqrt(z)*z));
}
/*----------------------------------------------------------------------
	Cauchy or Lorentzian Distribution of r[i]/(scale*sig[i]):
		rho2(t) = log(1 + 0.5*t^2) ~ 0.5*t^2
		wgt2(t) = 1/(1 + 0.5*t^2)
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 22-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double rho2(double t)
#else	/* old style */
double rho2(t)
double t;
#endif
{
	return (log(0.5*t*t + 1.0));
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double wgt2(double t)
#else	/* old style */
double wgt2(t)
double t;
#endif
{
	return (1.0/(0.5*t*t + 1.0));
}
/*----------------------------------------------------------------------
	Hoaglin, Mosteller and Tukey: Table 8-2, page 293.
	Tukey's biweight estimator (with B = 3.6):
		rho3(t) = B^2*(1 - (1 - (t/B)^2)^3)/6 ~ 0.5*t^2,  |t| < B
		wgt3(t) = (1 - (t/B)^2)^2,  |t| < B
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 15-Jan-92
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double rho3(double t)
#else	/* old style */
double rho3(t)
double t;
#endif
{
	double B = 3.6;
	double x = B*B/6.0;
	if (fabs(t) < B) {
		t /= B;
		t = (1.0 - t)*(1.0 + t);
		x *= (1.0 - t*t*t);
	}
	return x;
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double wgt3(double t)
#else	/* old style */
double wgt3(t)
double t;
#endif
{
	double B = 3.6;
	double x = 0.0;
	if (fabs(t) < B) {
		t /= B;
		x = (1.0 - t)*(1.0 + t);
		x *= x;
	}
	return x;
}
/*----------------------------------------------------------------------
	The functions rho0(t), rho1(t), rho2(t), and rho3(t) are assigned
	to an array rho[] of pointers to function.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 01-Jan-92
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double (*rho[])(double) = { rho0, rho1, rho2, rho3 };
#else	/* old style */
double (*rho[])()       = { rho0, rho1, rho2, rho3 };
#endif
/*----------------------------------------------------------------------
	The functions one(t), wgt1(t), wgt2(t), and wgt3(t) are assigned
	to an array wgt[] of pointers to function.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 01-Jan-92
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double (*wgt[])(double) = { one, wgt1, wgt2, wgt3 };
#else	/* old style */
double (*wgt[])()       = { one, wgt1, wgt2, wgt3 };
#endif
/*----------------------------------------------------------------------
	The functions one(t), phi1(t), wgt2(t), and wgt3(t) are assigned
	to an array phi[] of pointers to function.
	The phi1(t) is an M-estimator, wgt2(t) and wgt3(t) are W-estimators.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 01-Jan-92
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double (*phi[])(double) = { one, phi1, wgt2, wgt3 };
#else	/* old style */
double (*phi[])()       = { one, phi1, wgt2, wgt3 };
#endif
/*----------------------------------------------------------------------
	'normal' calculates a set of general 'Normal Equations', Ap = b,
	which are used to solve an over-determined system of linear equations,
	y = Xp + r, by stepwise minimizing the general merit function
	chi(scale; p) = Sum of rho(r[i]/(scale*sig[i])).
	Exploring Data Tables, Trends, and Shapes, Chp 8.
	Input:
      n     Number of observations.
      m     Number of regression coefficients.
      *X[]  Two dimensional float array containing the
            design matrix in the first m rows and n columns.
      r[]   One dimensional float array containing the n
            residuals calculated from the current regression
            coefficients.
		sig[] One dimensional float array containing the n
            standard deviations.
      scale Scale estimate, which is kept fixed during the
            optimization procedure.
      wgt() wgt(z) = psi(z)/z = rho'(z)/z.
      phi() phi(z) = psi'(z) = rho''(z).
	Output:
      *A[]  Two dimensional double array containing the A
            matrix in the first m rows and m columns.
      b[]   One dimensional double array of size m containing
            the right hand side of the 'Normal Equations'.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 22-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
void normal(int n, int m, float * const X[], const float r[],
	const float sig[], double scale, double (*wgt)(double),
	double (*phi)(double), double *A[], double b[])
#else	/* old style */
void normal(n, m, X, r, sig, scale, wgt, phi, A, b)
int n, m;
float * const X[];
const float r[];
const float sig[];
double scale, (*wgt)(), (*phi)(), *A[], b[];
#endif
{
	int i, j, k;
	double s, t, u, v;

	/* initialize b and lower triangular part of A */
	for (j = 0; j < m; ++j) {
		b[j] = 0.0;
		for (k = 0; k <= j; ++k)
			A[j][k] = 0.0;
	}

	/* calculate b and lower triangular part of A */
	for (i = 0; i < n; ++i) {
		s = 1.0/(sig[i]*sig[i]);
		t = r[i]/(scale*sig[i]);
		u = s*phi(t);
		v = s*wgt(t)*r[i];
		for (j = 0; j < m; ++j) {
			b[j] += v*X[j][i];
			s = u*X[j][i];
			for (k = 0; k <= j; ++k)
				A[j][k] += s*X[k][i];
		}
	}

	/* assign upper triangular part of A */
	for (j = 1; j < m; ++j)
		for (k = 0; k < j; ++k)
				A[k][j] = A[j][k];

}
/*----------------------------------------------------------------------
	'chi' calculates the residuals r[i], and the merit function:
	chi(scale; p) = Sum of rho(r[i]/(scale*sig[i])).
	Exploring Data Tables, Trends, and Shapes, Chp 8.
	Input:
      n     Number of observations.
      m     Number of regression coefficients.
      *X[]  Two dimensional float array containing the
            design matrix in the first m rows and n columns.
      y[]   One dimensional float array containing the n
            observations.
      sig[] One dimensional float array containing the n
            standard deviations.
      scale Scale estimate, which is kept fixed during the
            optimization procedure.
      rho() chi(scale; p) = Sum of rho(r[i]/(scale*sig[i])).
      p[]   One dimensional double array containing the m
            regression coefficients.
	Output:
      r[]   One dimensional float array containing the n
            residuals.
	Return:
            The merit function.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 24-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double chi(int n, int m, float * const X[], const float y[],
	const float sig[], double scale, double (*rho)(double),
	const double p[], float r[])
#else	/* old style */
double chi(n, m, X, y, sig, scale, rho, p, r)
int n, m;
float * const X[];
const float y[];
const float sig[];
double scale, (*rho)();
const double p[];
float r[];
#endif
{
	int i, j;
	double dif, sum = 0.0;

	for (i = 0; i < n; ++i) {
		dif = y[i];
		for (j = 0; j < m; ++j)
			dif -= X[j][i]*p[j];
		r[i] = (float) dif;
		sum += rho(dif/(scale*sig[i]));
	}
	return sum;
}
/*----------------------------------------------------------------------
	'ols', solves the Normal Equations, Ap = b, of the linear
	regression model, y = Xp + r.
	Exploring Data Tables, Trends, and Shapes, Chp 8, p283.
	Input:
      n     Number of observations.
      m     Number of regression coefficients.
      *X[]  Two dimensional float array containing the
            design matrix in the first m rows and n columns.
      y[]   One dimensional float array containing the n
            observations.
      sig[] One dimensional float array containing the n
            standard deviations.
	Output:
      p[]   One dimensional double array of size m containing
            the OLS estimate of the regression coefficients.
      r[]   One dimensional float array containing the n
            residuals.
	Returns:
            The current value of the merit function.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 22-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double ols(int n, int m, float * const X[], const float y[],
	const float sig[], double p[], float r[])
#else	/* old style */
double ols(n, m, X, y, sig, p, r)
int n, m;
float * const X[];
const float y[];
const float sig[];
double p[];
float r[];
#endif
{
	int ipvt[MD], k;
	double a[MD][MD], *A[MD], work[MD], cond;

	if (m > MD) {
		fprintf(stderr, "ols: Too many parameters m = %d\n", m);
		exit(EXIT_FAILURE);
	}

	/* assign pointer array */
	for (k = 0; k < MD; ++k)
		A[k] = a[k];

	/* calculate the normal equations */
	normal(n, m, X, y, sig, 1.0, wgt[0], phi[0], A, p);

	/* decompose the A matrix */
	cond = decomp(m, A, ipvt, work);

	/* Check if A is singular to working precision */
	if ((cond + 1.0) == cond) {
		fprintf(stderr, "ols: A is singular to working precision\n");
		exit(EXIT_FAILURE);
	}

	/* solve the normal equations */
	solve(m, A, p, ipvt);

	/* calculate the residuals and return the merit function */
	return chi(n, m, X, y, sig, 1.0, rho[0], p, r);
}
/*----------------------------------------------------------------------
	'mad' calculates the Median Absolute Deviation (MAD) of the
	normalized residuals r[i]/sig[i].
	Exploring Data Tables, Trends, and Shapes, Chp 8, p302.
	Input:
      n     Number of observations.
      r[]   One dimensional float array containing the n
            current residuals.
      sig[] One dimensional float array containing the n
            standard deviations.
	Return:
	          The Median Absolute Deviation of r[i]/sig[i].
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 29-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double mad(int n, const float r[], const float sig[])
#else	/* old style */
double mad(n, r, sig)
int n;
const float r[];
const float sig[];
#endif
{
	int i, k1, k2;
	float *work;
	double rmad, rmedi;

	/* allocate the work[] vector */
	if ((work = (float *)calloc(n, sizeof(float))) == (float *)0) {
		fprintf(stderr, "mad: Can't allocate work[%d]\n", n);
		exit(EXIT_FAILURE);
	}

	/* calculate the normalized residuals */
	for (i = 0; i < n; ++i)
		work[i] = (float) (r[i]/sig[i]);

	/* find the median */
	k1 = (n - 1)/2;
	k2 = n/2;
	find(n, work, k1);
	rmedi = work[k1];
	if (k2 != k1) {
		find(n, work, k2);
		rmedi = (rmedi + work[k2])/2.0;
	}

	/* calculate the absolute deviations */
	for (i = 0; i < n; ++i)
		work[i] = (float) fabs(work[i] - rmedi);

	/* find the Median of the Absolute Deviations */
	find(n, work, k1);
	rmad = work[k1];
	if(k2 != k1) {
		find(n, work, k2);
		rmad = (rmad + work[k2])/2.0;
	}

	/* free the work[] array */
	free(work);

	/* The MAD of a Normal Distribution is 0.6745 */
	/* times the usual Standard Deviation */

	return rmad;
}
/*----------------------------------------------------------------------
	'chistep' uses the Levenberg-Marquardt method to reduce the merit
	function 'chi' of the linear regression model, y = Xp + r, by
	using a robust estimator, possibly derived by maximum-likelihood
	arguments.
	Exploring Data Tables, Trends, and Shapes, Chp 8, p305.
	Numerical Recipes in C, Chp. 14.4, p540.
	Input:
      n      Number of observations.
      m      Number of regression coefficients.
      *X[]   Two dimensional float array containing the
             design matrix in the first m rows and n columns.
      y[]    One dimensional float array containing the n
             observations.
      sig[]  One dimensional float array containing the n
             standard deviations.
      scale  Scale estimate, which is kept fixed during the
             optimization procedure.
      rho()  chi(scale; p) = Sum of rho(r[i]/(scale*sig[i])).
      wgt()  wgt(z) = psi(z)/z = rho'(z)/z.
      phi()  phi(z) = psi'(z) = rho''(z).
      p[]    One dimensional double array containing the m
             regression coefficients obtained from the
             previous iteration step.
      *lamda Marquardt's parameter.
	Output:
      p[]    One dimensional double array containing the m
             updated regression coefficients.
      r[]    One dimensional float array containing the n
             updated residuals.
      *lamda Marquardt's lamda parameter.
	Returns:
             The updated value of the merit function.
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 29-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double chistep(int n, int m, float * const X[], const float y[],
	const float sig[], double scale, double (*rho)(double),
	double (*wgt)(double), double (*phi)(double),
	double p[], float r[], double *lamda)
#else	/* old style */
double chistep(n, m, X, y, sig, scale, rho, wgt, phi, p, r, lamda)
int n, m;
float * const X[];
const float y[];
const float sig[];
float r[];
double scale, (*rho)(), (*wgt)(), (*phi)(), p[], *lamda;
#endif
{
	int ipvt[MD], j, k;
	double a[MD][MD], q[MD][MD];
	double *A[MD], *Q[MD], b[MD], dp[MD], work[MD];
	double oldchi, newchi, cond;

	if (m > MD) {
		fprintf(stderr, "chistep: Too many parameters m = %d\n", m);
		exit(EXIT_FAILURE);
	}

	/* assign pointer arrays */
	for (k = 0; k < MD; ++k) {
		A[k] = a[k];
		Q[k] = q[k];
	}

	/* calculate the residuals and the merit function */
	oldchi = chi(n, m, X, y, sig, scale, rho, p, r);

	/* calculate the general 'Normal Equations' */
	normal(n, m, X, r, sig, scale, wgt, phi, A, b);

	/* increase lamda until the merit function declines */
	do {

		/* make new working copies of A and b */
		for (j = 0; j < m; ++j) {
			for (k = 0; k < m; ++k)
				Q[j][k] = A[j][k];
			dp[j] = b[j];
		}

		/* increase the diagonal part of Q by Marquardt's prescription */
		if (lamda[0] > 0.0)
			for (k = 0; k < m; ++k)
				Q[k][k] *= (1.0 + lamda[0]);

		/* decompose the Q matrix */
		cond = decomp(m, Q, ipvt, work);

		/* Check if Q is singular to working precision */
		if ((cond + 1.0) == cond) {
			fprintf(stderr, "chistep: Q is singular to working precision\n");
			exit(EXIT_FAILURE);
		}

		/* solve for the increments to the regression coefficients */
		solve(m, Q, dp, ipvt);

		/* make a working copy of the updated coefficients */
		for (k = 0; k < m; ++k)
			work[k] = p[k] + dp[k];

		/* calculate the new residuals, and the new merit function */
		newchi = chi(n, m, X, y, sig, scale, rho, work, r);

		/* update Marquardt's parameter */
		if (lamda[0] > 0.0) {
			if (newchi < oldchi)
				lamda[0] *= 0.1;
			else
				lamda[0] *= 10.0;
		}
	} while (lamda[0] > 0.0 && newchi >= oldchi);

	/* update the regression coefficient */
	for (k = 0; k < m; ++k)
		p[k] += dp[k];

	return newchi;
}
/*----------------------------------------------------------------------
	'robust' does a robust regression to a linear model, y = Xp + r.
	The methods used are based on:
		Chapter 8, Robust Regression, by Guoying Li in
		Exploring Data Tables, Trends, and Shapes, edited by
		D. C. Hoaglin, F. Mosteller, and J. W. Tukey
		John Wiley & Sons
------------------------------------------------------------------------
	The following steps are done					
		0. An ordinary least squares solution (OLS)
		1. An iterative M1-estimate with
			psi(t) = t*w(t), where w(t) = 1/sqrt(1 + t*t)
		2. An iterative W2-estimate with
			psi(t) = t*w(t), where w(t) = 1/(1 + 0.5*t*t)
		3. An iterative W3-estimate with
			psi(t) = t*w(t), where w(t) = exp(-0.5*t*t)
------------------------------------------------------------------------
	Input:
      method 0: Least Squares, 1: M1-estimate,
             2: W2-estimate, and 3: W3-estimate.
      n      Number of observations.
      m      Number of regression coefficients.
      *X[]   Two dimensional float array containing the
             design matrix in the first m rows and n columns.
      y[]    One dimensional float array containing the n
             observations.
      sig[]  One dimensional float array containing the n
             standard deviations.
	Output:
      par[]  One dimensional double array containing the m
             regression coefficients.
      r[]    One dimensional float array containing the n
             final residuals.
	Return:
             The Median Absolute Deviation (MAD) divided by
             the MAD of a Standard Normal Distribution (StMAD).
------------------------------------------------------------------------
	Bjarne Thomsen, Aarhus, 29-Dec-91
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double robust(int method, int n, int m, float * const X[],
	const float y[], const float sig[], double par[], float r[])
#else	/* old style */
double robust(method, n, m, X, y, sig, par, r)
int method, n, m;
float * const X[];
const float y[];
const float sig[];
double par[];
float r[];
#endif
{
	int i, k, nit[3];
	double oldchi, newchi, lamda, scale;
	const double StMAD = 0.6745;

	/* iteration limits */
	nit[0] = 20;
	nit[1] = 20;
	nit[2] = 20;

	if (m > MD) {
		fprintf(stderr, "robust: Too many parameters m = %d\n", m);
		exit(EXIT_FAILURE);
	}

	/* begin with an Ordinary Least Squares solution */
	newchi = ols(n, m, X, y, sig, par, r);

	/* find scale factor from MAD of the OLS-residuals */
	scale = mad(n, r, sig)/StMAD;

	/* first obtain M1-estimate, then obtain W2-estimate */
	/* finally obtain W3-estimate */
	/* by using Marquardt's method, and a constant scale factor */
	for (k = 1; k <= method; ++k) {	/* step through methods */

		/* initialize Marquardt's parameter */
		lamda = 0.001;

		/* do the first step */
		newchi = chistep(n, m, X, y, sig, scale,
					rho[k], wgt[k], phi[k], par, r, &lamda);

		/* do at most 'nit' steps */
		for (i = 0; i < nit[k-1]; ++i) {
			oldchi = newchi;
			newchi = chistep(n, m, X, y, sig, scale,
					rho[k], wgt[k], phi[k], par, r, &lamda);
			if (fabs(oldchi - newchi) < 0.1)
				break;
		}

		/* no solution found after 'nit' steps */
		/* but continue anyway */
		if (i == nit[k-1]) {
			fprintf(stderr,
				"robust: No solution after %d %d-steps\n", nit[k-1], k);
		}

		/* do the final step */
		lamda = 0.0;
		newchi = chistep(n, m, X, y, sig, scale,
					rho[k], wgt[k], phi[k], par, r, &lamda);

		/* find scale factor from MAD of the residuals */
		scale = mad(n, r, sig)/StMAD;
	}

	return scale;
}
/*----------------------------------------------------------------------
   'FitRows' fits a polynomial series to each row of an image.
	The following expression is fitted:
		pixel(j, i) = a0*Pl0(x) + a1*Pl1(x) + a2*Pl2(x) + ...
	where
		x = 2*(i-xmin)/(xmax-xmin) - 1, and i = xmin,..., xmax.
	Pixels which are flagged as bad, or which deviate more than
	reject*StandardDeviation are replaced by the fitted value.
------------------------------------------------------------------------
	Input:
      method  0: Least Squares, 1: M1-estimate,
              2: W2-estimate, and 3: W3-estimate.
      nrows   Number of rows.
      xmin    First column in fit.
      xmax    Last column in fit.
      **pixel Two dimensional pixel array (float).
      order   Maximum order of the polynomial series.
      rnoise  Read noise in ADUs.
      epadu   Electrons per ADU.
		reject  Pixel value is rejected,
              if abs(res) > reject*StandardDeviation.
	Output:
      **pixel All bad are pixels replaced by fitted values.
------------------------------------------------------------------------
		Bjarne Thomsen, Aarhus, 25-Dec-1991
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int FitRows(
int method,
int nrows,
int xmin,
int xmax,
float **pixel,
int order,
double rnoise,
double epadu,
double reject)
#else	/* old style */
int FitRows(method, nrows, xmin, xmax, pixel, order, rnoise, epadu, reject)
int method, nrows, xmin, xmax, order;
float **pixel;
double rnoise, epadu, reject;
#endif
{
	int X_rows, i, j, k, n, m = order + 1;
	double par[MD], scale, noise2 = rnoise*rnoise;
	double poly[MD], x, fitval, resmax, limit;
	float **X, *y, *sig, *r;
	const float FltMax = (float) (0.9*FLT_BLANK);

	if (m > MD) {
		fprintf(stderr, "FitRows: Too many parameters: %d\n", m);
		return 0;
	}

	/* initialize parameter array */
	for (k = 0; k < MD; ++k)
		par[k] = 0.0;

	/* allocate r[] vector */
	if ((r = (float *)calloc(xmax-xmin+1, sizeof(float))) == (float *)0) {
		fprintf(stderr, "FitRows: Can't allocate r[%d]\n", xmax-xmin+1);
		return 0;
	}

	/* allocate sig[] vector */
	if ((sig = (float *)calloc(xmax-xmin+1, sizeof(float))) == (float *)0) {
		fprintf(stderr, "FitRows: Can't allocate sig[%d]\n", xmax-xmin+1);
		free(r);
		return 0;
	}

	/* allocate y[] vector */
	if ((y = (float *)calloc(xmax-xmin+1, sizeof(float))) == (float *)0) {
		fprintf(stderr, "FitRows: Can't allocate y[%d]\n", xmax-xmin+1);
		free(sig);
		free(r);
		return 0;
	}

	/* allocate X[][] matrix */
	X = (float**)matalloc(m, sizeof(float *),
			xmax-xmin+1, sizeof(float), &X_rows);
	/* check if rows are allocated */
	if (X_rows < m) {
		fprintf(stderr,
			"FitRows: Can't allocate X[%d][%d]\n", m, xmax-xmin+1);
		free(y);
		free(sig);
		free(r);
		return 0;
	}

	for (j = 0; j < nrows; ++j) {	/* fit next row */
		/* Calculate the design matrix */
		n = 0;
		for (i = xmin; i <= xmax; ++i)	/* process next pixel */
			if (pixel[j][i] <= FltMax) {
				y[n] = pixel[j][i];
				if (y[n] > 0)
					sig[n] = (float)sqrt(y[n]/epadu + noise2);
				else
					sig[n] = (float)rnoise;
				x = (double)(i - xmin);
				x *= 2.0;
				x /= (xmax - xmin);
				x -= 1.0;
				fleg(x, poly, m);
				for (k = 0; k < m; ++k)
					X[k][n] = (float)poly[k];
				n++;
			}

		if (n < m) {
			fprintf(stderr, "FitRows: Too few equations: %d\n", n);
			matfree(X, X_rows);
			free(y);
			free(sig);
			free(r);
			return 0;
		}

		/* Solve equations */
		scale = robust(method, n, m, X, y, sig, par, r);

		/* rejection limit */
		limit = reject*scale;

		/* replace bad pixels with fitted values */
		for (i = xmin; i <= xmax; ++i) {	/* process next pixel */
			x = (double)(i - xmin);
			x *= 2.0;
			x /= (xmax - xmin);
			x -= 1.0;
			fleg(x, poly, m);
			fitval = 0.0;
			for (k = 0; k < m; ++k)
				fitval += par[k]*poly[k];
			if (fitval > 0.0)
				resmax = limit*sqrt(fitval/epadu + noise2);
			else
				resmax = limit*rnoise;
			if (pixel[j][i] > FltMax)
				pixel[j][i] = (float) fitval;
			else if (fabs(pixel[j][i] - fitval) > resmax)
				pixel[j][i] = (float) fitval;
		}
	}

	/* free arrays */
	matfree(X, X_rows);
	free(y);
	free(sig);
	free(r);

	return 1;
}
/*----------------------------------------------------------------------
   'FitCols' fits a polynomial series to each column of an image.
	The following expression is fitted:
		pixel(j, i) = a0*Pl0(y) + a1*Pl1(y) + a2*Pl2(y) + ....
	where
		y = 2*(j-ymin)/(ymax-ymin) - 1, and j = ymin,..., ymax.
	Pixels which are flagged as bad, or which deviate more than
	reject*StandardDeviation are replaced by the fitted value.
------------------------------------------------------------------------
	Input:
      method  0: Least Squares, 1: M1-estimate,
              2: W2-estimate, and 3: W3-estimate.
      ymin    First row in fit.
      ymax    Last row in fit.
      ncols   Number of columns.
      **pixel Two dimensional pixel array (float).
      order   Maximum order of the polynomial series.
      rnoise  Read noise in ADUs.
      epadu   Electrons per ADU.
		reject  Pixel value is rejected,
              if abs(res) > reject*StandardDeviation.
	Output:
      **pixel All bad are pixels replaced by fitted values.
------------------------------------------------------------------------
		Bjarne Thomsen, Aarhus, 25-Dec-1991
----------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
int FitCols(
int method,
int ymin,
int ymax,
int ncols,
float **pixel,
int order,
double rnoise,
double epadu,
double reject)
#else	/* old style */
int FitCols(method, ymin, ymax, ncols, pixel, order, rnoise, epadu, reject)
int method, ymin, ymax, ncols, order;
float **pixel;
double rnoise, epadu, reject;
#endif
{
	int X_rows, i, j, k, n, m = order + 1;
	double par[MD], scale, noise2 = rnoise*rnoise;
	double poly[MD], x, fitval, resmax, limit;
	float **X, *y, *sig, *r;
	const float FltMax = (float) (0.9*FLT_BLANK);

	if (m > MD) {
		fprintf(stderr, "FitCols: Too many parameters: %d\n", m);
		return 0;
	}

	/* initialize parameter array */
	for (k = 0; k < MD; ++k) {
		par[k] = 0.0;
	}

	/* allocate r[] vector */
	if ((r = (float *)calloc(ymax-ymin+1, sizeof(float))) == (float *)0) {
		fprintf(stderr, "FitCols: Can't allocate r[%d]\n", ymax-ymin+1);
		return 0;
	}

	/* allocate sig[] vector */
	if ((sig = (float *)calloc(ymax-ymin+1, sizeof(float))) == (float *)0) {
		fprintf(stderr, "FitRows: Can't allocate sig[%d]\n", ymax-ymin+1);
		free(r);
		return 0;
	}

	/* allocate y[] vector */
	if ((y = (float *)calloc(ymax-ymin+1, sizeof(float))) == (float *)0) {
		fprintf(stderr, "FitCols: Can't allocate y[%d]\n", ymax-ymin+1);
		free(sig);
		free(r);
		return 0;
	}

	/* allocate X[][] matrix */
	X = (float**)matalloc(m, sizeof(float *),
			ymax-ymin+1, sizeof(float), &X_rows);
	/* check if rows are allocated */
	if (X_rows < m) {
		fprintf(stderr,
			"FitCols: Can't allocate X[%d][%d]\n", m, ymax-ymin+1);
		free(y);
		free(sig);
		free(r);
		return 0;
	}

	for (i = 0; i < ncols; ++i) {	/* fit next column */
		/* Calculate design matrix */
		n = 0;
		for (j = ymin; j <= ymax; ++j)	/* process next pixel */
			if (pixel[j][i] <= FltMax) {
				y[n] = pixel[j][i];
				if (y[n] > 0)
					sig[n] = (float)sqrt(y[n]/epadu + noise2);
				else
					sig[n] = (float)rnoise;
				x = (double)(j - ymin);
				x *= 2.0;
				x /= (ymax - ymin);
				x -= 1.0;
				fleg(x, poly, m);
				for (k = 0; k < m; ++k)
					X[k][n] = (float)poly[k];
				n++;
			}

		if (n < m) {
			fprintf(stderr, "FitCols: Too few equations: %d\n", n);
			matfree(X, X_rows);
			free(y);
			free(sig);
			free(r);
			return 0;
		}

		/* Solve equations */
		scale = robust(method, n, m, X, y, sig, par, r);

		/* rejection limit */
		limit = reject*scale;

		/* replace bad pixels with fitted values */
		for (j = ymin; j <= ymax; ++j) {	/* process next pixel */
			x = (double)(j - ymin);
			x *= 2.0;
			x /= (ymax - ymin);
			x -= 1.0;
			fleg(x, poly, m);
			fitval = 0.0;
			for (k = 0; k < m; ++k)
				fitval += par[k]*poly[k];
			if (fitval > 0.0)
				resmax = limit*sqrt(fitval/epadu + noise2);
			else
				resmax = limit*rnoise;
			if (pixel[j][i] > FltMax)
				pixel[j][i] = (float) fitval;
			else if (fabs(pixel[j][i] - fitval) > resmax)
				pixel[j][i] = (float) fitval;
		}
	}

	/* free arrays */
	matfree(X, X_rows);
	free(y);
	free(sig);
	free(r);

	return 1;
}
/*----------------------------------------------------------------------
      This program cleans an image for cosmic-ray events by fitting a
		polynomial series to either rows or columns by the methods of
		robust regression. Pixels having large residuals are replaced
		by values calculated from the polynomial fit.
      The methods used are based on:
		Chapter 8, Robust Regression, by Guoying Li in
		Exploring Data Tables, Trends, and Shapes, edited by
		D. C. Hoaglin, F. Mosteller, and J. W. Tukey
		John Wiley & Sons
------------------------------------------------------------------------
		Bjarne Thomsen, Aarhus, 14-Dec-91
----------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#define ROWCOL	1	/* 0: select rows, 1: select columns */
#define METHOD	3	/* 0: LS, 1: M1-est, 2: W2-est, and 3: W3-est */
#define RNOISE	3.0	/* read noise in ADUs */
#define EPDN	2.6	/* Electrons per Digital Number */
#define REJECT	4.0	/* reject limit */
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv, **arg;
	int option, opt_err = 0, n_arg = 0;
	int rowcol = ROWCOL;	/* select rows or columns */
	int order  = ORDER;	/* maximum polynomial order */
	int method = METHOD;	/* fitting method */
	double rnoise = RNOISE;	/* read noise */
	double epdn  = EPDN;	/* electrons per DN */
	double reject = REJECT;	/* reject limit */
	void *row;	/* row buffer */
	char *image, *image1;	/* image names */
	int i, x, y, xymin, xymax, crmax;
	int id, type, ncols, nrows, a_rows;
	double Blank;
	float **a;	/* pixel array a[][] */
	const float FltBlank = (float)FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);
	const float Half = (float)0.5;
	const float Zero = (float)0;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "cmray: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	/* process command line options */
	while (((option = getopt(argc, argv, "xyo:m:n:e:r:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'x':	rowcol = 0; break;	/* select image rows */
			case 'y':	rowcol = 1; break;	/* select image columns */
			case 'o':	sscanf(optarg, "%d", &order); break;
			case 'm':	sscanf(optarg, "%d", &method); break;
			case 'n':	sscanf(optarg, "%lf", &rnoise); break;	/* read noise */
			case 'e':	sscanf(optarg, "%lf", &epdn); break;	/* Ele pr DN */
			case 'r':	sscanf(optarg, "%lf", &reject); break; /* reject lim */
			case '?':	opt_err++; break;	/* unknown option */
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	/* illegal options */
	if (opt_err || n_arg < 2) {
		printf("Usage: cmray -x -y -o or -m me -n rn -e epd -r rej im im1 [min] [max]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	if (order > ORDER) {
		fprintf(stderr, "cmray: polynomial order = %d is too big\n", order);
		exit(EXIT_FAILURE);
	}

	image  = truncname(arg[0]);
	/* is input image name legal? */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "cmray: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	image1 = truncname(arg[1]);
	/* is output image name legal? */
	if (strpbrk(image1, punct) != (char *)0) {
		fprintf(stderr, "cmray: Illegal image name: %s\n", image1);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "cmray: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	/* open input image for read only */
	if ((id = iopen(image, 0, &type, &ncols, &nrows)) < 0) {
		fprintf(stderr, "cmray: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "cmray: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type < 1 || type > 5) {
		fprintf(stderr, "cmray: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "cmray: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank = blank(image);

	if (xist(image1)) {	/* does output image exist? */
		fprintf(stderr, "cmray: %s: Existing image\n", image1);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "cmray: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	crmax = (rowcol == 0)? (ncols - 1): (nrows - 1);

	if (n_arg > 2) {
		sscanf(arg[2], "%d", &xymin);
		if (xymin < 0) xymin = 0;
	} else xymin = 0;

	if (n_arg > 3) {
		sscanf(arg[3], "%d", &xymax);
		if (xymax > crmax) xymax = crmax;
	} else xymax = crmax;

	if (xymax <= xymin) {
		fprintf(stderr, "cmray: xymax <= xymin\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "cmray: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate float array a[][] */
	a = (float **)matalloc(
		nrows, sizeof(float *), ncols, sizeof(float), &a_rows);
	if (a_rows < nrows) {
		fprintf(stderr, "cmray: Can't allocate a[%d][%d]\n", nrows, ncols);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "cmray: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate row buffer */
	if ((row = (void *)calloc(ncols, bytepix(type))) == (void *)0) {
		fprintf(stderr, "cmray: Can't allocate row[%d]\n", ncols);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "cmray: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (y = 0; y < nrows; ++y) {	/* read image into array a[][] */
		getline(id, y, row);	/* read next row */
		switch (type) {
			case 1:
				{
					const UCHAR *q = (UCHAR *)row;
					float *p = a[y];
					for (x = 0; x < ncols; ++x)
						p[x] = (float)q[x];
				}
				break;
			case 2:
				{
					const short *q = (short *)row;
					float *p = a[y];
					const short b = (short)Blank;
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 3:
				{
					const float *q = (float *)row;
					float *p = a[y];
					const float h = (float) (0.9*Blank);
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x] > h)? FltBlank: q[x];
				}
				break;
			case 4:
				{
					const LONG *q = (LONG *)row;
					float *p = a[y];
					const LONG b = (LONG)Blank;
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 5:
				{
					const COMPLEX *q = (COMPLEX *)row;
					float *p = a[y];
					const float h = (float) (0.9*Blank);
					for (x = 0; x < ncols; ++x)
						p[x] = (q[x].x > h || q[x].y > h)? FltBlank: q[x].x;
				}
				break;
		}
	}

	iclose(id);	/* close image */

	if (rowcol == 0)	/* fit rows, and replace bad data */
		if (FitRows(method, nrows, xymin, xymax,
				a, order, rnoise, epdn, reject) == 0) {
			fprintf(stderr, "cmray: Error while fitting rows\n");
			goto Finish;
		}
	if (rowcol == 1)	/* fit columns, and replace bad data */
		if (FitCols(method, xymin, xymax, ncols,
				a, order, rnoise, epdn, reject) == 0) {
			fprintf(stderr, "cmray: Error while fitting columns\n");
			goto Finish;
		}

	if (cpi(image, image1) < 1) {	/* dublicate image header */
		fprintf(stderr,
			"cmray: cpi(%s, %s) has failed\n", image, image1);
		goto Finish;
	} else if (setimagefile(image1, "") == 0) {	/* reset FILE key */
		fprintf(stderr,
			"cmray: Can't erase old FILE value for %s\n", image1);
		rmi(image1);	/* remove image header */
		goto Finish;
	} else if ((id = icreat(image1, 1, type, ncols, nrows)) < 0) {
		fprintf(stderr, "cmray: Can't create %s\n", image1);
		rmi(image1);	/* remove image header */
		goto Finish;
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "cmray: update: Disk header not created\n");

	Blank = blank(image1);
	for (y = 0; y < nrows; ++y) {	/* write image to disk */
		switch (type) {
			case 1:
				{
					UCHAR *q = (UCHAR *)row;
					const float *p = a[y];
					for (x = 0; x < ncols; ++x)
						q[x] = (UCHAR)(p[x] + Half);
				}
				break;
			case 2:
				{
					short *q = (short *)row;
					const float *p = a[y];
					const short b = (short)Blank;
					for (x = 0; x < ncols; ++x)
						q[x] = (p[x] > FltMax)? b:
								(short)((p[x] < 0)? p[x]-Half: p[x]+Half);
				}
				break;
			case 3:
				{
					float *q = (float *)row;
					const float *p = a[y];
					const float b = (float)Blank;
					for (x = 0; x < ncols; ++x)
						q[x] = (p[x] > FltMax)? b: p[x];
				}
				break;
			case 4:
				{
					LONG *q = (LONG *)row;
					const float *p = a[y];
					const LONG b = (LONG)Blank;
					for (x = 0; x < ncols; ++x)
						q[x] = (p[x] > FltMax)? b:
								(LONG)((p[x] < 0)? p[x]-Half: p[x]+Half);
				}
				break;
			case 5:
				{
					COMPLEX *q = (COMPLEX *)row;
					const float *p = a[y];
					const float b = (float)Blank;
					int ip;
					for (x = 0; x < ncols; ++x) {
						ip = p[x] > FltMax;
						q[x].x = (ip)? b: p[x];
						q[x].y = (ip)? b: Zero;
					}
				}
				break;
		}
		putline(id, y, row);	/* write row to disk */
	}

	iclose(id);	/* close output image */

	setiname(image1);
	sethistory(image1, "cmray");
	sethistory(image1, image1);

Finish:

	/* free row[] */
	free(row);
	/* free pixel array */
	matfree(a, a_rows);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "cmray: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}

