#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "fft2.h"	/* function declarations for fft2.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#if	!defined(PI)
#define	PI	3.141592653589793
#endif
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	float FltSub = (float)0;	/* substitute for blank input values */
	int n_arg = 0;
	char **arg;
	void *buffer;
	char *image;
	float *a;
	int i, y, ncols;
	int id, npix, nlin, type;
	double amp1, amp2, rad1, rad2, Blank;
	const float Zero = (float) 0;
	const float Half = (float) 0.5;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "convpsf: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "b:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'b':	sscanf(optarg, "%f", &FltSub); break;
			case '?':	opt_err++; break;	/* unknown option */
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 5) {
		printf("Usage: convpsf -b FltSub image A1 A2 R1 R2\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);	/* image name */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "convpsf: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	sscanf(arg[1], "%lf", &amp1);
	sscanf(arg[2], "%lf", &amp2);
	sscanf(arg[3], "%lf", &rad1);
	sscanf(arg[4], "%lf", &rad2);

	if (rad1 < 1E-9 || rad2 < 1E-9) {
		fprintf(stderr, "addpsf: rad1 = %g  rad2 = %g\n", rad1, rad2);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "convpsf: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */
	if ((id = iopen(image, 0, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "convpsf: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"convpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 4) {
		fprintf(stderr,
			"convpsf: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"convpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (npix != nlin) {
		fprintf(stderr, "convpsf: %s has npix = %d and nlin = %d\n",
			image, npix, nlin);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"convpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (i = 1; i < nlin; i <<= 1)
		;
	if (i > nlin) {
		fprintf(stderr,
			"convpsf: %s has illegal nlin = %d\n", image, nlin);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"convpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (nlin > max_dim()) {
		fprintf(stderr,
			"convpsf: %s has too big nlin = %d\n", image, nlin);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"convpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank = blank(image);
	if ((buffer = (void *)calloc(npix, bytepix(type))) == (void *)0) {
		fprintf(stderr, "convpsf: Buffer allocation failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "convpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	ncols = npix + 2;
	if ((a = (float *)calloc(nlin, ncols*sizeof(float))) == (float *)0) {
		fprintf(stderr, "convpsf: Array allocation failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"convpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (i = 0, y = 0; y < nlin; i += ncols, ++y) {
		getline(id, y, buffer);	/* fetch new row */
		switch (type) {
			case 1:	/* unsigned char */
				{
					const UCHAR *q = (UCHAR *) buffer;
					int x;
					for (x = 0; x < npix; ++x)
						a[i+x] = (float) q[x];
				}
				break;
			case 2:	/* short int */
				{
					const short *q = (short *) buffer;
					const short b = (short) Blank;
					int x;
					for (x = 0; x < npix; ++x)
						a[i+x] = (q[x] == b)? FltSub: (float)q[x];
				}
				break;
			case 3:	/* float */
				{
					const float *q = (float *) buffer;
					const float h = (float) (0.9*Blank);
					int x;
					for (x = 0; x < npix; ++x)
						a[i+x] = (q[x] > h)? FltSub: q[x];
				}
				break;
			case 4:	/* LONG int */
				{
					const LONG *q = (LONG *) buffer;
					const LONG b = (LONG) Blank;
					int x;
					for (x = 0; x < npix; ++x)
						a[i+x] = (q[x] == b)? FltSub: (float)q[x];
				}
				break;
		}
	}
	fft2(ncols, a, nlin, -1, 1);
	rad1 = nlin/rad1;
	rad2 = nlin/rad2;
	rad1 *= rad1;
	rad2 *= rad2;
	for (i = 0, y = 0; y < nlin; i += ncols, ++y) {
		double y2 = (double) ((y <= nlin/2)? y: nlin - y);
		int k, x;
		y2 *= y2;
		for(k = 0, x = 0; x <= npix/2; k += 2, ++x) {
			double r = -PI*((double)x*(double)x + y2);
			float f = (float)(amp1*exp(r/rad1) + amp2*exp(r/rad2));
			a[i+k] *= f;
			a[i+k+1] *= f;
		}
	}
	fft2(ncols, a, nlin, -1, -1);
	if (!chiomod(id, 2)) {	/* change IO mode to update */
		fprintf(stderr, "convpsf: Can't change IO mode for %s\n", image);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "addstar: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (i = 0, y = 0; y < nlin; i += ncols, ++y) {
		switch (type) {
			case 1:	/* unsigned char */
				{
					UCHAR *q = (UCHAR *) buffer;
					int x;
					for (x = 0; x < npix; ++x)
						q[x] = (a[i+x]<Zero)? 0: (UCHAR) (a[i+x] + Half);
				}
				break;
			case 2:	/* short int */
				{
					short *q = (short *) buffer;
					int x;
					for (x = 0; x < npix; ++x)
						q[x] = (short)((a[i+x]<Zero)? a[i+x]-Half: a[i+x]+Half);
				}
				break;
			case 3:	/* float */
				{
					float *q = (float *) buffer;
					int x;
					for (x = 0; x < npix; ++x)
						q[x] = a[i+x];
				}
				break;
			case 4:	/* LONG int */
				{
					LONG *q = (LONG *) buffer;
					int x;
					for (x = 0; x < npix; ++x)
						q[x] = (LONG)((a[i+x]<Zero)? a[i+x]-Half: a[i+x]+Half);
				}
				break;
		}
		putline(id, y, buffer);
	}
	free(a);
	free(buffer);
	iclose(id);
	sethistory(image, "convpsf");
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "convpsf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
