#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "opxtract.h"	/* function declaration for opxtract.c */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define MIN(a, b)  (((a) < (b)) ? (a) : (b))
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int i, extx, exty;
	int id1, type1, npix1, nlin1;
	int id2, type2, npix2, nlin2;
	double Blank1, Blank2;
	char *image1, *image2;
	void *buf1, *buf2;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if (argc < 3) {
		printf("Usage: embed image1 image2\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image1 = truncname(argv[1]);
	if (strpbrk(image1, punct) != (char *)0) {
		fprintf(stderr, "embed: Illegal image name: %s\n", image1);
		exit(EXIT_FAILURE);
	}

	image2 = truncname(argv[2]);
	if (strpbrk(image2, punct) != (char *)0) {
		fprintf(stderr, "embed: Illegal image name: %s\n", image2);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "embed: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	if ((id1 = iopen(image1, 0, &type1, &npix1, &nlin1)) < 0) {
		fprintf(stderr, "embed: Can't open %s\n", image1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "embed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank1 = blank(image1);

	setbufsiz(8192);	/* change buffer size */

	if ((id2 = iopen(image2, 2, &type2, &npix2, &nlin2)) < 0) {
		fprintf(stderr, "embed: Can't open %s\n", image2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "embed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank2 = blank(image2);

	if ((buf1 = (void *)calloc(npix1, bytepix(type1))) == (void *)0) {
		fprintf(stderr, "embed: Allocation of buffer #1 failed\n");
		iclose(id2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "embed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((buf2 = (void *)calloc(npix2, bytepix(type2))) == (void *)0) {
		fprintf(stderr, "embed: Allocation of buffer #2 failed\n");
		iclose(id2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "embed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	extx = MIN(npix1, npix2);
	exty = MIN(nlin1, nlin2);
	for (i = 0; i < exty; i++) {
		getline(id1, i, buf1);
		getline(id2, i, buf2);
		bufcpy(buf2, buf1, extx, type2, type1, Blank2, Blank1);
		putline(id2, i, buf2);
	}

	free(buf2);
	free(buf1);
	iclose(id2);
	iclose(id1);

	sethistory(image2, "embed");
	sethistory(image2, image1);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "embed: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	fflush(stdout);
	exit(EXIT_SUCCESS);
}
