#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "opifpix.h"	/* function declaration for opifpix.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	char *image, *buf, str[5];
	char *array = (char *)0, *ptr = (char *)0;
	int k, l, n, id, type, npix, nlin, rop, aop;
	int bsize, narray, rw = 2;	/* read and write */
	unsigned u, ioff, asize, nb, nblok, mblok, maxblok;
	long isize;
	double Blank, cons, value;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if (argc < 6) {
		fprintf(stderr, "Usage: ifpix image rop const aop value\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(argv[1]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "ifpix: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	sscanf(argv[2], "%4s", str);
	rop=(strncmp(str,  "<", 1) == 0)? '<':
		 (strncmp(str, "lt", 2) == 0)? '<':
		 (strncmp(str, "LT", 2) == 0)? '<':
		 (strncmp(str,  ">", 1) == 0)? '>':
		 (strncmp(str, "gt", 2) == 0)? '>':
		 (strncmp(str, "GT", 2) == 0)? '>':
		 (strncmp(str,  "=", 1) == 0)? '=':
		 (strncmp(str, "eq", 2) == 0)? '=':
		 (strncmp(str, "EQ", 2) == 0)? '=': 0;

	sscanf(argv[3], "%lf", &cons);

	sscanf(argv[4], "%4s", str);
	aop=(strncmp(str,   "+", 1) == 0)? '+':
		 (strncmp(str, "add", 3) == 0)? '+':
		 (strncmp(str, "ADD", 3) == 0)? '+':
		 (strncmp(str,   "-", 1) == 0)? '-':
		 (strncmp(str, "sub", 3) == 0)? '-':
		 (strncmp(str, "SUB", 3) == 0)? '-':
		 (strncmp(str,   "*", 1) == 0)? '*':
		 (strncmp(str, "mul", 3) == 0)? '*':
		 (strncmp(str, "MUL", 3) == 0)? '*':
		 (strncmp(str,   "/", 1) == 0)? '/':
		 (strncmp(str, "div", 3) == 0)? '/':
		 (strncmp(str, "DIV", 3) == 0)? '/':
		 (strncmp(str,   "=", 1) == 0)? '=':
		 (strncmp(str,  "eq", 2) == 0)? '=':
		 (strncmp(str,  "EQ", 2) == 0)? '=': 0;

	sscanf(argv[5], "%lf", &value);

	if (aop == '/' && value == 0.0) {
		fprintf(stderr, "ifpix: Divide by zero\n");
		exit(EXIT_FAILURE);
	}
	/* read image names from header directory */
	if ((k = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "ifpix: prolog: Error = %d\n", k);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	if ((id = iopen(image, rw, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "ifpix: can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "ifpix: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 5) {
		fprintf(stderr, "ifpix: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "ifpix: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank = blank(image);
	isize = imagesize(image);	/* image size in bytes */
	bsize = buffsize();	/* block size in bytes */
	maxblok = UINT_Max/bsize;	/* maximum number of blocks available */
	mblok = (unsigned)((isize - 1L)/bsize + 1L);	/* blocks in image */
	if (mblok > maxblok)	/* find maximum number of blocks in array */
		mblok = maxblok;
	asize = 0;	/* array size in bytes */
	for (nblok = mblok; nblok > 0; nblok--) {	/* find big array */
		asize = nblok*bsize;	/* try this array size */
		if ((array = (char *)malloc(asize)) != (char *)0)
			break;	/* big array allocated */
	}
	if (nblok <= 0) {	/* array has not been allocated */
		fprintf(stderr, "ifpix: malloc: Array allocation failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "ifpix: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (nblok < mblok) {	/* no space for maximum size */
		nb = nblok;
		nb++;
		nb >>= 1;
		asize = nb*bsize;	/* array size reduced by 50% */
		if ((ptr = (char *)realloc(array, asize)) != (char *)0) {
			array = ptr;	/* reallocation was a success */
			nblok = nb;
		} else
			asize = nblok*bsize;	/* reallocation failed */
	}
	narray = (int)((isize - 1L)/asize + 1L);	/* number of buf arrays */
	for (ioff = 0, k = 0; k < narray; k++) {	/* step array index */
		n = 0;	/* initialize pixel counter for array */
		buf = array;	/* initialize block pointer */
		for (u = 0; u < nblok; u++)	/* step block index */
			if ((l = bread(id, ioff + u, buf)) == 0)	/* read 1 block */
				break;	/* no more blocks, exit loop */
			else {
				n += l;	/* update number of pixels read into array */
				buf += bsize;	/* increment block pointer */
			}
		switch (type) {	/* operate on full array */
			case 1:
				ifpix_uchar(array, n, rop, cons, aop, value);
				break;
			case 2:
				ifpix_short(array, n, rop, cons, aop, value, Blank);
				break;
			case 3:
				ifpix_float(array, n, rop, cons, aop, value, Blank);
				break;
			case 4:
				ifpix_long(array, n, rop, cons, aop, value, Blank);
				break;
		}
		buf = array;	/* initialize block pointer */
		for (u = 0; u < nblok; u++)	/* step blocks index */
			if (bwrite(id, ioff + u, buf) == 0)	/* write 1 block */
				break;	/* no more blocks, exit loop */
			else
				buf += bsize;	/* increment block pointer */
		ioff += nblok;	/* increment block offset of array start */
	}
	free(array);	/* remember to deallocate the big array */
	iclose(id);
	sethistory(image, "ifpix");
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "ifpix: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
