#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#include <time.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define NUCHAR(x) ((UCHAR) (((x) < 0.0)? 0.0: (x) + 0.5))
#define NSHORT(x) ((short) (((x) < 0.0)? (x) - 0.5: (x) + 0.5))
#define NLONG(x) ((LONG) (((x) < 0.0)? (x) - 0.5: (x) + 0.5))
/*--------------------------------------------------------------------*/
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double ran1(long *idum)
#else	/* old style */
double ran1(idum)
long *idum;
#endif
{
	int j;
	long k;
	static long idum2 = 123456789;
	static long iy = 0;
	static long iv[NTAB];
	double temp;

	if (*idum <= 0) {
		if (-(*idum) < 1) *idum = 1;
		else *idum = -(*idum);
		idum2 = (*idum);
		for (j = NTAB+7; j >= 0; j--) {
			k = (*idum)/IQ1;
			*idum = IA1*(*idum - k*IQ1) - k*IR1;
			if (*idum < 0) *idum += IM1;
			if (j < NTAB) iv[j] = *idum;
		}
		iy = iv[0];
	}
	k = (*idum)/IQ1;
	*idum = IA1*(*idum - k*IQ1) - k*IR1;
	if (*idum < 0) *idum += IM1;
	k = idum2/IQ2;
	idum2 = IA2*(idum2 - k*IQ2) - k*IR2;
	if (idum2 < 0) idum2 += IM2;
	j = iy/NDIV;
	iy = iv[j] - idum2;
	iv[j] = *idum;
	if (iy < 1) iy += IMM1;
	if ((temp = AM*iy) > RNMX) return RNMX;
	else return temp;
}
/*--------------------------------------------------------------------*/
#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double gasdev(long *idum)
#else	/* old style */
double gasdev(idum)
long *idum;
#endif
{
	static int iset = 0;
	static double gset;
	double fac, rsq, v1, v2;

	if  (iset == 0) {
		do {
			v1 = 2.0*ran1(idum) - 1.0;
			v2 = 2.0*ran1(idum) - 1.0;
			rsq = v1*v1 + v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac = sqrt(-2.0*log(rsq)/rsq);
		gset = v1*fac;
		iset = 1;
		return v2*fac;
	} else {
		iset = 0;
		return gset;
	}
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double gammln(double xx)
#else	/* old style */
double gammln(xx)
double xx;
#endif
{
	double x, y, tmp, ser;
	static double cof[6]={76.18009172947146,-86.50532032941677,
		24.01409824083091,-1.231739572450155,
		0.1208650973866179e-2,-0.5395239384953e-5};
	int j;

	y = x = xx;
	tmp = x + 5.5;
	tmp -= (x + 0.5)*log(tmp);
	ser = 1.000000000190015;
	for (j = 0; j <= 5; j++) ser += cof[j]/++y;
	return ( -tmp + log(2.5066282746310005*ser/x));
}
/*--------------------------------------------------------------------*/
#define PI 3.141592653589793
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
double poidev(double xm, long *idum)
#else	/* old style */
double poidev(xm, idum)
double xm;
long *idum;
#endif
{
	static double sq, alxm, g, oldm = (-1.0);
	double em, t, y;

	if (xm < 12.0) {
		if (xm != oldm) {
			oldm = xm;
			g = exp(-xm);
		}
		em = -1;
		t = 1.0;
		do {
			++em;
			t *= ran1(idum);
		} while (t > g);
	} else {
		if (xm != oldm) {
			oldm = xm;
			sq = sqrt(2.0*xm);
			alxm = log(xm);
			g = xm*alxm - gammln(xm + 1.0);
		}
		do {
			do {
				y = tan(PI*ran1(idum));
				em = sq*y + xm;
			} while (em < 0.0);
			em = floor(em);
			t = 0.9*(1.0 + y*y)*exp(em*alxm - gammln(em + 1.0) - g);
		} while (ran1(idum) > t);
	}
	return em;
}
#undef PI
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int noise = 0;	/* add read noise? */
	int photon = 0;	/* photon noise? */
	double sdev = 5.0;	/* standard deviation in DN */
	double gain = 1.0;	/* gain in e/DN */
	long idum = 0L;	/* take (long)time(NULL) as seed */
	char *image;
	int id, i, npix, nlin, bloknr, blokpix;
	int type = 2;	/* short */
	int rw = 1;	/* write only */
	double value;
	void *buf;
	const float Zero = (float)0;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "imake: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "p:n:s:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'p':	photon++; sscanf(optarg, "%lf", &gain); break;
			case 'n':	noise++; sscanf(optarg, "%lf", &sdev); break;
			case 's':	sscanf(optarg, "%ld", &idum); break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 4) {
      printf(
      "Usage: imake -p gain -n sdev -s seed image value npix nlin [type]\n");
		fflush(stdout);
      exit(EXIT_FAILURE);
	}

	if (idum == 0L) idum = (long) time(NULL);

	image = truncname(arg[0]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "imake: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	sscanf(arg[1], "%lf", &value);	/* pixel value */
	sscanf(arg[2], "%d", &npix);	/* x-dimension */
	sscanf(arg[3], "%d", &nlin);	/* y-dimension */
	if (n_arg > 4) sscanf(arg[4], "%d", &type);	/* image type */

	if (type < 1 || type > 5) {
		fprintf(stderr, "imake: Type = %d is not implemented\n", type);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "imake: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	if (xist(image)) {	/* does output image exist? */
		fprintf(stderr, "imake: %s: Existing image\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "cosmic: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if((id = icreat(image, rw, type, npix, nlin)) < 0) {
		fprintf(stderr, "imake: Can't create %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "imake: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "imake: update: Disk header not created\n");

	if ((buf = (void *)malloc(buffsize())) == (void *)0) {
		fprintf(stderr, "imake: Can't allocate output buffer\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "imake: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	blokpix = bpixels(image);
	bloknr = -1;	/* block number */
	do {
		++bloknr;
		switch(type) {
		case 1:
			{
				UCHAR *p = (UCHAR *)buf;
				double d;
				int i;
				for(i = 0; i < blokpix; ++i) {
					d = value;
					if (photon) d = poidev(gain*d, &idum)/gain;
					if (noise) d += sdev*gasdev(&idum);
					p[i] = NUCHAR(d);
				}
			}
			break;
		case 2:
			{
				short *p = (short *)buf;
				double d;
				int i;
				for(i = 0; i < blokpix; ++i) {
					d = value;
					if (photon) d = poidev(gain*d, &idum)/gain;
					if (noise) d += sdev*gasdev(&idum);
					p[i] = NSHORT(d);
				}
			}
			break;
		case 3:
			{
				float *p = (float *)buf;
				double d;
				int i;
				for(i = 0; i < blokpix; ++i) {
					d = value;
					if (photon) d = poidev(gain*d, &idum)/gain;
					if (noise) d += sdev*gasdev(&idum);
					p[i] = (float)d;
				}
			}
			break;
		case 4:
			{
				LONG *p = (LONG *)buf;
				double d;
				int i;
				for(i = 0; i < blokpix; ++i) {
					d = value;
					if (photon) d = poidev(gain*d, &idum)/gain;
					if (noise) d += sdev*gasdev(&idum);
					p[i] = NLONG(d);
				}
			}
			break;
		case 5:
			{
				COMPLEX *p = (COMPLEX *)buf;
				double d;
				int i;
				for(i = 0; i < blokpix; ++i) {
					d = value;
					if (photon) d = poidev(gain*d, &idum)/gain;
					if (noise) d += sdev*gasdev(&idum);
					p[i].x = (float)d;
					p[i].y = Zero;
				}
			}
			break;
		}
	} while (bwrite(id, bloknr, buf) == blokpix);

	free(buf);
	iclose(id);
	sethistory(image, "imake");
	sethistory(image, image);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "imake: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
