#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
#include "medi.h" /* function declaration for medi.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define	ILIM	2	/* number of iterations */
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int bit_hist = 0;
	int short_median = 0;
	int i, x, y;
	int xmin, xmax, ymin, ymax;
	int stx, sty, extx, exty;
	int id, type, npix, nlin, rw = 0;
	double Blank;
	double sum, dat, mean, var, clip = 5.0;
	void *buffer;
	char *image;
	unsigned wsize, npt, m, n;
	float *p, *store, *work;
	unsigned long bit_count[16];

	for (i = 0; i < 16; ++i)
		bit_count[i] = 0;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "mean: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "hs")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'h':	bit_hist++; break;
			case 's':	short_median++; break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}
	if (opt_err || n_arg < 1) {
		printf("Usage: mean -h -s image [Xmin] [Xmax] [Ymin] [Ymax]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "mean: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}
	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "mean: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	if ((id = iopen(image, rw, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "mean: can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "mean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 5) {
		fprintf(stderr, "mean: %s has illegal type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "mean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank = blank(image);

	if (n_arg > 1) sscanf(arg[1], "%d", &xmin); else xmin = 0;
	if (n_arg > 2) sscanf(arg[2], "%d", &xmax); else xmax = npix - 1;
	if (n_arg > 3) sscanf(arg[3], "%d", &ymin); else ymin = 0;
	if (n_arg > 4) sscanf(arg[4], "%d", &ymax); else ymax = nlin - 1;

	stx = xmin;
	sty = ymin;
	extx = xmax - xmin + 1;
	exty = ymax - ymin + 1;
	if (stx < 0) {
		extx += stx;
		stx = 0;
	}
	if (sty < 0) {
		exty += sty;
		sty = 0;
	}
	if (extx <= 0 || exty <= 0) {
		fprintf(stderr, "mean: Window is placed above/left of image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "mean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (stx >= npix || sty >= nlin) {
		fprintf(stderr, "mean: Window is placed below/right of image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "mean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (stx + extx > npix)
		extx = npix - stx;
	if (sty + exty > nlin)
		exty = nlin - sty;
	if (!window(id, stx, sty, extx, exty)) {
		fprintf(stderr, "mean: Window is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "mean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((buffer = (void *)calloc(extx, bytepix(type))) == (void *)0) {
		fprintf(stderr, "mean: Allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "mean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	wsize = extx;
	wsize *= exty;	/* number of pixels in window */
	if ((store = (float *)calloc(wsize, sizeof(float))) == (float *)0) {
		fprintf(stderr, "mean: Allocation of window buffer failed\n");
		free(buffer);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "mean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	p = store;
	for (y = 0; y < exty; y++) {
		getline(id, y, buffer);
		switch (type) {
			case 1:
				{
					const UCHAR *q = (UCHAR *)buffer;
					for (x = 0; x < extx; ++x)
						*p++ = (float)q[x];
				}
				break;
			case 2:
				{
					const short *q = (short *)buffer;
					const short b = (short)Blank;
					for (x = 0; x < extx; ++x) {
						if (q[x] != b)
							*p++ = (float)q[x];
						if (bit_hist) {
							unsigned short word = q[x];
							int bit;
							for (bit = 0; bit < 16; ++bit) {
								if (word & 1)
									bit_count[bit]++;
								word >>= 1;
							}
						}
					}
				}
				break;
			case 3:
				{
					const float *q = (float *)buffer;
					const float h = (float) (0.9*Blank);
					for (x = 0; x < extx; ++x)
						if (q[x] <= h)
							*p++ = q[x];
				}
				break;
			case 4:
				{
					const LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank;
					for (x = 0; x < extx; ++x)
						if (q[x] != b)
							*p++ = (float)q[x];
				}
				break;
			case 5:
				{
					const COMPLEX *q = (COMPLEX *)buffer;
					const float h = (float) (0.9*Blank);
					for (x = 0; x < extx; ++x)
						if (q[x].x <= h && q[x].y <= h)
							*p++ = q[x].x;
				}
				break;
		}
	}
	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "mean: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	npt = p - store;
	if (npt < 2) {
		fprintf(stderr, "mean: Area has too few pixels\n");
		exit(EXIT_FAILURE);
	}
	sum = 0.0;
	for (n = 0; n < npt; ++n)
		sum += store[n];
	mean = sum/npt;
	sum = 0.0;
	for (n = 0; n < npt; ++n) {
		dat = store[n] - mean;
		dat *= dat;
		sum += dat;
	}
	var = sum/(npt - 1);
	if (short_median == 0)
		printf("n = %8d    mean = %8.2f    var = %8.2f\n", npt, mean, var);
	for (i = 0; i < ILIM; ++i) {
		double lim = clip * var;
		sum = 0.0;
		m = 0;
		for(n = 0; n < npt; ++n) {
			dat = store[n] - mean;
			dat *= dat;
			if (dat <= lim) {
				sum += store[n];
				store[m] = store[n];
				++m;
			}
		}
		npt = m;
		if (npt < 2) {
			fprintf(stderr, "mean: Area has too few pixels\n");
			exit(EXIT_FAILURE);
		}
		mean = sum/npt;
		sum = 0.0;
		for (n = 0; n < npt; ++n) {
			dat = store[n] - mean;
			dat *= dat;
			sum += dat;
		}
		var = sum/(npt - 1);
		if (short_median == 0)
			printf("n = %8d    mean = %8.2f    var = %8.2f\n", npt, mean, var);
	}
	if (short_median != 0) {
		if ((work=(float *)calloc(npt, sizeof(float))) == (float *)0) {
			fprintf(stderr, "mean: Allocation of work array failed\n");
			free(store);
			free(buffer);
			exit(EXIT_FAILURE);
		}
		mean = fmedian(npt, store, work);
		free(work);
		printf("%10.4g\n", mean);
	}
	if (bit_hist) {
		printf("\n");
		for (i = 0; i < 8; ++i)
			printf("%8ld", bit_count[i]);
		printf("\n");
		for (i = 8; i < 16; ++i)
			printf("%8ld", bit_count[i]);
		printf("\n\n");
	}
	free(store);
	free(buffer);
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
