#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <stdlib.h>	/* For ANSI compatibility, errno declared */
#include <string.h>	/* required only for function declarations */
#include <math.h>		/* sin(), cos(), atan2(), sqrt() */
#if	defined(MSC) || defined(WC)	/* MS-C or WATCOM-C 9.0/386 */
#include <graph.h>	/* Standard Video Modes defined in graph.h */
#endif
#include <conio.h>	/* Keyboard I/O */
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "imhdr.h"	/* function prototypes for imhdr.c */
#include "imkw.h"	/* function prototypes for imkw.c */
#include "imio.h"	/* function prototypes for imio.c */
#include "nr.h"	/* function declarations for nr.c */
#include "vga256.h"	/* function prototypes for vga256.c */
#include "hcconio.h"	/* function prototypes for hcconio.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
/*--------------------------------------------------------------------*/
#define LF	10
#define CR	13
/*--------------------------------------------------------------------*/
#define CURSIZ	5	/* cursor size */
#define NAMSIZ	40	/* full file name size */
/*--------------------------------------------------------------------*/
#define Pi	3.141592653589793
#define Rad(x)	((x)*Pi/180.0)
/*--------------------------------------------------------------------*/
#define NHUE	360	/* number of hues */
static UCHAR red[NHUE], green[NHUE], blue[NHUE];
/*--------------------------------------------------------------------*/
static float **_Array;	/* pointer to array of pointers */
static int _Xmin = 0;	/* first column displayed */
static int _Xmax = 0;	/* last column displayed */
static int _Ymin = 0;	/* first row displayed */
static int _Ymax = 0;	/* last row displayed */
static int _Irot = 0;	/* has image been rotated? */
static int _Mode = 0;	/* display mode */
/*--------------------------------------------------------------------*/
static double _Epdn = 2.6;	/* Electrons per Digital Number */
static double _Fwhm = 5.0;	/* Full Width at Half Maximum */
static int    _Rfit = 5;	/* Default radius of fitting area */
static int    _Rsky = 10;	/* Default inner radius of sky annulus */
/*--------------------------------------------------------------------*/
static double _Scale[] = {116.22, 64.69};	/* 1 pixel in guide units */
static double _Angle[] = {1.06, -179.14};	/* rotation of CCD in deg */
static double _RotPos  = 0.0;	/* rotator position in deg */
static int    _No      =   0;	/* 0: P8603, 1: TK512 */
/*--------------------------------------------------------------------*/
static void GetCCD_No(const char *name)
{
	const char *str = instrume(name);
	if (str != NULL) {	/* instrument specified in header */
		if (strstr(str, "P8603") != NULL)
			_No = 0;
		else if (strstr(str, "TK512") != NULL)
			_No = 1;
	}
}
/*--------------------------------------------------------------------*/
static void ReadRGB(void)	/* Read R, G, and B tables */
{
	char **path, fname[NAMSIZ];
	int count;
	FILE *stream;
	if ((path = getpath("IMSYS", &count)) == (char **)0)
		strcpy(fname, ".");	/* current directory */
	else
		strcpy(fname, path[0]);	/* pathname from environment */
	strcat(fname, "/rgbtab.dat");
	if ((stream = fopen(fname, "rb")) == (FILE *)0) {
		fprintf(stderr, "can't open %s\n", fname);
		exit(EXIT_FAILURE);
	}
	fread(red, sizeof (char), NHUE, stream);
	fread(green, sizeof (char), NHUE, stream);
	fread(blue, sizeof (char), NHUE, stream);
	fclose(stream);
}
/*--------------------------------------------------------------------*/
static UCHAR rgb[768];	/* RGB Lookup Table */
/*--------------------------------------------------------------------*/
static void Lookup(int hue)	/* Calculate Lookup Table */
{
	UCHAR *p = rgb;
	int i, j;
	while (hue < 0)
		hue += NHUE;
	while (hue >= NHUE)
		hue -= NHUE;
	for (i = 0; i < 256; i++) {
		j = (hue + i) % NHUE;
		*p++ = red[j];
		*p++ = green[j];
		*p++ = blue[j];
	}
}
/*--------------------------------------------------------------------*/
static void GrayLookup(void)	/* Calculate Gray Scale Lookup Table */
{
	UCHAR *p = rgb;
	int i, j;
	for (i = 0; i < 256; i++) {
		j = i >> 2;
		*p++ = (UCHAR)j;
		*p++ = (UCHAR)j;
		*p++ = (UCHAR)j;
	}
}
/*--------------------------------------------------------------------*/
static void DrawCursor(	/* Draw cursor at screen (x,y) */
int x, int y, UCHAR ucv)
{
	int i;
	int row_siz = rowsize();
	int col_siz = colsize();
	while (x < 0)
		x += row_siz;
	while (x >= row_siz)
		x -= row_siz;
	while (y < 0)
		y += col_siz;
	while (y >= col_siz)
		y -= col_siz;
	for (i = 1; i <= CURSIZ; ++i) {
		if ((x-i) >= 0)
			draw_pixel(y, x - i, (UCHAR) (read_pixel(y, x - i) + ucv));
		if ((x+i) < row_siz)
			draw_pixel(y, x + i, (UCHAR) (read_pixel(y, x + i) + ucv));
		if ((y-i) >= 0)
			draw_pixel(y - i, x, (UCHAR) (read_pixel(y - i, x) + ucv));
		if ((y+i) < col_siz)
			draw_pixel(y + i, x, (UCHAR) (read_pixel(y + i, x) + ucv));
	}
}
/*--------------------------------------------------------------------*/
static UCHAR DisplayValue(	/* Reads display value at screen (x,y) */
int x, int y)
{
	int row_siz = rowsize();
	int col_siz = colsize();
	while (x < 0)
		x += row_siz;
	while (x >= row_siz)
		x -= row_siz;
	while (y < 0)
		y += col_siz;
	while (y >= col_siz)
		y -= col_siz;
	return read_pixel(y, x);
}
/*--------------------------------------------------------------------*/
static double PixelValue(	/* Reads pixel value at screen (x,y) */
int x, int y)
{
	int row_siz = rowsize();
	int col_siz = colsize();
	while (x < 0)
		x += row_siz;
	while (x >= row_siz)
		x -= row_siz;
	while (y < 0)
		y += col_siz;
	while (y >= col_siz)
		y -= col_siz;
	if (_Irot) {
		if (x > (_Ymax-_Ymin) || y > (_Xmax-_Xmin)) return 0;
		return _Array[_Ymin+x][_Xmin+y];
	} else {
		if (x > (_Xmax-_Xmin) || y > (_Ymax-_Ymin)) return 0;
		return _Array[_Ymin+y][_Xmin+x];
	}
}
/*--------------------------------------------------------------------*/
typedef struct { double x; double y;} POINT;
/*--------------------------------------------------------------------*/
static POINT RotScale(POINT p)	/* scale and rotate point */
{
	double Theta, Radius;
	POINT q;
	Theta  = atan2(p.y, p.x) + Rad(_Angle[_No] + _RotPos);
	Radius =	sqrt(p.x * p.x + p.y * p.y) * _Scale[_No];
	q.x = Radius * cos(Theta);
	q.y = Radius * sin(Theta);
	return q;
}
/*--------------------------------------------------------------------*/
typedef struct { int x; int y;} PIXEL;
/*--------------------------------------------------------------------*/
static POINT Point(PIXEL pixel)
{
	POINT p;
	p.x = (double) pixel.x;
	p.y = (double) pixel.y;
	return p;
}
/*--------------------------------------------------------------------*/
static PIXEL PixelXY(	/* Pixel coordinates at screen (x,y) */
int x, int y)
{
	PIXEL p;
	int row_siz = rowsize();
	int col_siz = colsize();
	while (x < 0)
		x += row_siz;
	while (x >= row_siz)
		x -= row_siz;
	while (y < 0)
		y += col_siz;
	while (y >= col_siz)
		y -= col_siz;
	if (_Irot) {
		p.x = _Xmin + y;
		p.y = _Ymin + x;
	} else {
		p.x = _Xmin + x;
		p.y = _Ymin + y;
	}
	return p; 
}
/*--------------------------------------------------------------------*/
char _String1[50] = {'\0'};	/* global string containing output text */
char _String2[50] = {'\0'};	/* global string containing output text */
/*--------------------------------------------------------------------*/
static void CursorControl(int hue,	/* Cursor control */
const int scr_line)	/* string written on this line */
{
	char *str;
	int xcen, ycen;
	POINT point;	/* guide position */
	PIXEL pixel;	/* pixel position */
	double xc, yc, fwhm, dxc, dyc, dfwhm;
	double epdn = _Epdn;	/* Electrons per DN */
	int rfit = _Rfit;	/* Radius of fitting area */
	int smin = _Rsky;	/* Inner radius of sky annulus */
	int smax = 2*_Rsky;	/* Outer radius of sky annulus */
	int row_siz = rowsize();
	int col_siz = colsize();
	int c = 'm';	/* Move cursor */
	UCHAR astep = 1;
	UCHAR bstep = 10;
	UCHAR color = 128;
	UCHAR ucv = color;
	const float FltBlank = (float)FLT_BLANK;
	if (_Irot) {
		xcen = MIN(row_siz, _Ymax - _Ymin + 1)/2;
		ycen = MIN(col_siz, _Xmax - _Xmin + 1)/2;
	} else {
		xcen = MIN(row_siz, _Xmax - _Xmin + 1)/2;
		ycen = MIN(col_siz, _Ymax - _Ymin + 1)/2;
	}
	DrawCursor(xcen, ycen, ucv);
	while (c != 'e') {
		switch(c) {
			case 'c':	/* change cursor color */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					if (c == 'H' || c == 'M')
						color += 8;
					if (c == 'P' || c == 'K')
						color -= 8;
					DrawCursor(xcen, ycen, (UCHAR) (-ucv));
					ucv = color;
					sprintf(_String1,
						"                                        ");
					write_str(_String1, scr_line, 0, 255);
					sprintf(_String2,
						"cursor color: %6d                    ", color);
					write_str(_String2, scr_line+1, 0, 255);
					DrawCursor(xcen, ycen, ucv);
				}
			break;
			case 's':	/* change cursor step size */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					if (c == 'H' || c == 'M')
						astep++;
					if (c == 'P' || c == 'K')
						astep--;
					if (c == 'Q' || c == 'G')
						bstep++;
					if (c == 'S' || c == 'O')
						bstep--;
					sprintf(_String1,
						"                                        ");
					write_str(_String1, scr_line, 0, 255);
					sprintf(_String2,
						"steps: %6d %6d                    ", astep, bstep);
					write_str(_String2, scr_line+1, 0, 255);
				}
			break;
			case 'h':	/* change starting hue for color lookup table */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					if (c == 'H' || c == 'M') {
						hue += 4;
						while (hue >= NHUE)
							hue -= NHUE;
					}
					if (c == 'P' || c == 'K') {
						hue -= 4;
						while (hue < 0)
							hue += NHUE;
					}
					sprintf(_String1,
						"                                        ");
					write_str(_String1, scr_line, 0, 255);
					sprintf(_String2,
						"starting hue: %6d                    ", hue);
					write_str(_String2, scr_line+1, 0, 255);
					Lookup(hue);
					set_color(rgb);
				}
			break;
			case 'g':	/* use gray scale lookup table */
				GrayLookup();
				set_color(rgb);
				c = 'm';	/* move cursor next time */
			break;
			case 'f':   /* fit a Gaussian to stellar image */
				fwhm = _Fwhm;
				pixel = PixelXY(xcen, ycen);
				xc = (double) pixel.x;
				yc = (double) pixel.y;
				if (pixel.x-smax>=_Xmin && pixel.x+smax<=_Xmax
				 && pixel.y-smax>=_Ymin && pixel.y+smax<=_Ymax) {
					str = g2fit(_Array, FltBlank, epdn,
						rfit, smin, smax, &xc, &yc,
						&fwhm, &dxc, &dyc, &dfwhm);
					if (str != (char *)0) {
						set_mode(_TEXTC80);
						fprintf(stderr,"nshow: %s",str);
						exit(EXIT_FAILURE);
					}
				}
				point.x = xc; point.y = yc;
				point = RotScale(point);
				sprintf(_String1,
					"%7.2f+-%5.3f %7.2f+-%5.3f %4.2f+-%4.2f",
						xc, dxc, yc, dyc, fwhm, dfwhm);
				write_str(_String1, scr_line, 0, 255);
				sprintf(_String2,
					"Guide coord:  X= %9.1f  Y= %9.1f",
						point.x, point.y);
				write_str(_String2, scr_line+1, 0, 255);
				c = 'm';	/* move cursor next time */
			break;
			case 'x':  /* 1-dim Gaussian fit along the x-axis */
				fwhm = _Fwhm;
				pixel = PixelXY(xcen, ycen);
				xc = (double) pixel.x;
				yc = (double) pixel.y;
				if (pixel.x-smax>=_Xmin && pixel.x+smax<=_Xmax
				 && pixel.y-smax>=_Ymin && pixel.y+smax<=_Ymax) {
					str = g1xfit(_Array, FltBlank, epdn,
						rfit, smin, smax, &xc, &yc,
						&fwhm, &dxc, &dfwhm);
					if (str != (char *)0) {
						set_mode(_TEXTC80);
						fprintf(stderr,"nshow: %s",str);
						exit(EXIT_FAILURE);
					}
				}
				point.x = xc; point.y = yc;
				point = RotScale(point);
				sprintf(_String1,
					"%7.2f+-%5.3f %7.2f        %4.2f+-%4.2f",
						xc, dxc, yc, fwhm, dfwhm);
				write_str(_String1, scr_line, 0, 255);
				sprintf(_String2, "Guide coord:  X= %9.1f  Y= %9.1f",
					point.x, point.y);
				write_str(_String2, scr_line+1, 0, 255);
				c = 'm';	/* move cursor next time */
			break;
			case 'y':  /* 1-dim Gaussian fit along the y-axis */
				fwhm = _Fwhm;
				pixel = PixelXY(xcen, ycen);
				xc = (double) pixel.x;
				yc = (double) pixel.y;
				if (pixel.x-smax>=_Xmin && pixel.x+smax<=_Xmax
				 && pixel.y-smax>=_Ymin && pixel.y+smax<=_Ymax) {
					str = g1yfit(_Array, FltBlank, epdn,
						rfit, smin, smax, &xc, &yc,
						&fwhm, &dyc, &dfwhm);
					if (str != (char *)0) {
						set_mode(_TEXTC80);
						fprintf(stderr,"nshow: %s",str);
						exit(EXIT_FAILURE);
					}
				}
				point.x = xc; point.y = yc;
				point = RotScale(point);
				sprintf(_String1,
					"%7.2f        %7.2f+-%5.3f %4.2f+-%4.2f",
						xc, yc, dyc, fwhm, dfwhm);
				write_str(_String1, scr_line, 0, 255);
				sprintf(_String2,
					"Guide coord:  X= %9.1f  Y= %9.1f",
						point.x, point.y);
				write_str(_String2, scr_line+1, 0, 255);
				c = 'm';	/* move cursor next time */
			break;
			case 'm':	/* move cursor */
				while ((c = get_ch()) == 0) {	/* use arrows */
					c = get_ch();	/* get second byte */
					DrawCursor(xcen, ycen, (UCHAR) (-ucv));
					switch(c) {
						case 'H':
							ycen -= astep;
							while (ycen < 0)
								ycen += col_siz;
						break;
						case 'P':
							ycen += astep;
							while (ycen >= col_siz)
								ycen -= col_siz;
						break;
						case 'K':
							xcen -= astep;
							while (xcen < 0)
								xcen += row_siz;
						break;
						case 'M':
							xcen += astep;
							while (xcen >= row_siz)
								xcen -= row_siz;
						break;
						case 'G':
							ycen -= bstep;
							while (ycen < 0)
								ycen += col_siz;
						break;
						case 'O':
							ycen += bstep;
							while (ycen >= col_siz)
								ycen -= col_siz;
						break;
						case 'S':
							xcen -= bstep;
							while (xcen < 0)
								xcen += row_siz;
						break;
						case 'Q':
							xcen += bstep;
							while (xcen >= row_siz)
								xcen -= row_siz;
						break;
						default:
						break;
					}
					DrawCursor(xcen, ycen, ucv);
					pixel = PixelXY(xcen, ycen);
					point = RotScale(Point(pixel));
					sprintf(_String1,
						"x=%4d y=%4d dis=%4d pix= %12g",
							pixel.x, pixel.y, DisplayValue(xcen, ycen),
							PixelValue(xcen, ycen));
					write_str(_String1, scr_line, 0, 255);
					sprintf(_String2,
						"Guide coord:  X= %9.1f  Y= %9.1f",
							point.x, point.y);
					write_str(_String2, scr_line+1, 0, 255);
				}
			break;
			default:
				while ((c = get_ch()) == 0)	/* get next non-arrow */
					c = get_ch();	/* skip second byte */
			break;
		}
	}
}
/*--------------------------------------------------------------------*/
static void Display(
const float low_cut,
const float high_cut)
{
	UCHAR Row[2060];	/* <------ REMEMBER THIS BUFFER */
	int x, y, npts;
	int row_siz = rowsize();
	const float factor = 255/(high_cut - low_cut);
	const float FltMax = (float) (0.9*FLT_BLANK);
	if (_Irot) {
		npts = MIN(row_siz, _Ymax - _Ymin + 1);
		for (x = _Xmin; x <= _Xmax; ++x) {
			float w;
			for (y = _Ymin; y <= _Ymax; ++y) {
				w = (_Array[y][x] > FltMax)? low_cut: _Array[y][x];
				w -= low_cut;
				w *= factor;
				Row[y - _Ymin] = (w < 0)? 0: (w > 255)? 255 : (UCHAR)w;
			}
			draw_line(x - _Xmin, 0, Row, npts);
		}
	} else {
		npts = MIN(row_siz, _Xmax - _Xmin + 1);
		for (y = _Ymin; y <= _Ymax; ++y) {
			float w;
			for (x = _Xmin; x <= _Xmax; ++x) {
				w = (_Array[y][x] > FltMax)? low_cut: _Array[y][x];
				w -= low_cut;
				w *= factor;
				Row[x - _Xmin] = (w < 0)? 0: (w > 255)? 255 : (UCHAR)w;
			}
			draw_line(y - _Ymin, 0, Row, npts);
		}
	}
}
/*--------------------------------------------------------------------*/
#define	STRSIZ	80	/* String size */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int imag = 0;	/* display imaginary part? */
	int hue, y, i;
	int id, type, ncol, nrow, xdim, ydim;
	int scr_line, vga_mode;
	double Blank;
	char *image, text[STRSIZ];
	void *row;
	float low_cut;
	float high_cut;
	const float FltBlank = (float) FLT_BLANK;
	const float FltMax = (float) (0.9*FLT_BLANK);

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "nshow: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "ptir:e:w:f:s:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'p':	_No = 0; break;
			case 't':	_No = 1; break;
			case 'i':	imag++; break;
			case 'r':	sscanf(optarg, "%lf", &_RotPos);
			case 'e':	sscanf(optarg, "%lf", &_Epdn); break;
			case 'w':	sscanf(optarg, "%lf", &_Fwhm); break;
			case 'f':	sscanf(optarg, "%d", &_Rfit); break;
			case 's':	sscanf(optarg, "%d", &_Rsky); break;
			case '?':	opt_err++; break;	/* unknown option */
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 1) {
		printf(
"nshow -ipt -r rot -e epdn -w fwhm -f fr -s sr image [Xmin] [Xmax] [Ymin] [Ymax]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	if (_Epdn < 0.1) {
		fprintf(stderr,
			"nshow: Too small No. of lectrons per DN = %g\n", _Epdn);
		exit(EXIT_FAILURE);
	}

	if (_Fwhm < 1.0) {
		fprintf(stderr, "nshow: Too small guess of FWHM = %g\n", _Fwhm);
		exit(EXIT_FAILURE);
	}

	if (_Rfit < 3) {
		fprintf(stderr, "nshow: Too small fitting radius = %d\n", _Rfit);
		exit(EXIT_FAILURE);
	}

	if (_Rsky < _Rfit) {
		fprintf(stderr, "nshow: Too small inner sky radius = %d\n", _Rsky);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);	/* image name */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "nshow: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "nshow: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	/* open image for read only */
	if ((id = iopen(image, 0, &type, &ncol, &nrow)) < 0) {
		fprintf(stderr, "nshow: can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "nshow: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type < 1 || type > 5) {
		fprintf(stderr, "nshow: %s: Image type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "nshow: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank = blank(image);
	GetCCD_No(image);	/* which CCD is used? */

	if (n_arg > 1) {
		sscanf(arg[1], "%d", &_Xmin);
		if (_Xmin < 0) _Xmin = 0;
	} else _Xmin = 0;

	if (n_arg > 2) {
		sscanf(arg[2], "%d", &_Xmax);
		if (_Xmax >= ncol) _Xmax = ncol - 1;
	} else _Xmax = ncol - 1;

	if (n_arg > 3) {
		sscanf(arg[3], "%d", &_Ymin);
		if (_Ymin < 0) _Ymin = 0;
	} else _Ymin = 0;

	if (n_arg > 4) {
		sscanf(arg[4], "%d", &_Ymax);
		if (_Ymax >= nrow) _Ymax = nrow - 1;
	} else _Ymax = nrow - 1;

	if (_Xmax < _Xmin) {
		fprintf(stderr, "nshow: xmax < xmin\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "nshow: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (_Ymax < _Ymin) {
		fprintf(stderr, "nshow: ymax < ymin\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "nshow: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* define data window */
	xdim = _Xmax - _Xmin + 1;	/* number of columns */
	ydim = _Ymax - _Ymin + 1;	/* number of rows */
	if (!window(id, _Xmin, _Ymin, xdim, ydim)) {
		fprintf(stderr, "nshow: Window is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "nshow: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate row buffer */
	if ((row = (void *)calloc(xdim, bytepix(type))) == (void *)0) {
		fprintf(stderr, "nshow: Allocation of row buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"nshow: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* allocate float matrix for data window */
	if ((_Array = matrix(_Ymin, _Ymax, _Xmin, _Xmax)) == (float **)0) {
		fprintf(stderr, "nshow: Allocation of array matrix failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "nshow: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	/* read data window into float matrix */
	for (y = _Ymin; y <= _Ymax; ++y) {
		getline(id, y - _Ymin, row);	/* fetch new row */
		switch (type) {
			case 1:	/* unsigned char */
				{
					const UCHAR *q = (UCHAR *)row - _Xmin;
					int x;
					for (x = _Xmin; x <= _Xmax; ++x)
						_Array[y][x] = (float)q[x];
				}
				break;
			case 2:	/* short int */
				{
					const short *q = (short *)row - _Xmin;
					const short b = (short)Blank;
					int x;
					for (x = _Xmin; x <= _Xmax; ++x)
						_Array[y][x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 3:	/* float */
				{
					const float *q = (float *)row - _Xmin;
					const float h = (float) (0.9*Blank);
					int x;
					for (x = _Xmin; x <= _Xmax; ++x)
						_Array[y][x] = (q[x] > h)? FltBlank: q[x];
				}
				break;
			case 4:	/* LONG int */
				{
					const LONG *q = (LONG *)row - _Xmin;
					const LONG b = (LONG)Blank;
					int x;
					for (x = _Xmin; x <= _Xmax; ++x)
						_Array[y][x] = (q[x] == b)? FltBlank: (float)q[x];
				}
				break;
			case 5:	/* complex */
				{
					const COMPLEX *q = (COMPLEX *)row - _Xmin;
					const float h = (float) (0.9*Blank);
					int x;
					if (imag) for (x = _Xmin; x <= _Xmax; ++x)
						_Array[y][x] = (q[x].x>h || q[x].y>h)? FltBlank: q[x].y;
					else for (x = _Xmin; x <= _Xmax; ++x)
						_Array[y][x] = (q[x].x>h || q[x].y>h)? FltBlank: q[x].x;
				}
				break;
		}
	}
	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "nshow: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	low_cut = FltMax;
	high_cut = -FltMax;
	for (y = _Ymin+3; y <= _Ymax; y += 7) {
		int x;
		for (x = _Xmin+3; x <= _Xmax; x += 7)
			if (_Array[y][x] <= FltMax) {
				if (low_cut > _Array[y][x])	/* find minimum */
					low_cut = _Array[y][x];
				if (high_cut < _Array[y][x])	/* find maximum */
					high_cut = _Array[y][x];
			}
	}
	ReadRGB();	/* Read R, G, and B tables */
	hue = 200;
	Lookup(hue);	/* Calculate Lookup Table */
#if	defined(ET4000)	/* Tseng Labs Super VGA */
	vga_mode = _ET1024x768;
	scr_line = 45;
	_Mode = '6';
#else		/* Paradise Super VGA */
	vga_mode = _PA640x480;
	scr_line = 28;
	_Mode = '3';
#endif
	while (1) {
		printf("Low cut  (%5g): ", low_cut);
		if (gets(text) != NULL && strlen(text) > 0)
			sscanf(text, "%f", &low_cut);
		printf("High cut (%5g): ", high_cut);
		if (gets(text) != NULL && strlen(text) > 0)
			sscanf(text, "%f", &high_cut);
		printf("Rotate (0|1) (%d): ", _Irot);
		if (gets(text) != NULL && strlen(text) > 0)
			sscanf(text, "%d", &_Irot);
		set_mode(vga_mode);
		set_color(rgb);
		Display(low_cut, high_cut);
		CursorControl(hue, scr_line);
		set_mode(_TEXTC80);
		printf("%s\n", _String1);
		printf("%s\n", _String2);
		printf("Repeat with other VGA mode?\n");
		printf("Standard    320x200: 1\n");
#if	defined(ET4000)	/* Tseng Labs Super VGA */
		printf("Tseng Labs  640x350: 2\n");
		printf("Tseng Labs  640x400: 3\n");
		printf("Tseng Labs  640x480: 4\n");
		printf("Tseng Labs  800x600: 5\n");
		printf("Tseng Labs 1024x768: 6\n");
#else		/* Paradise Super VGA */
		printf("Paradise    640x400: 2\n");
		printf("Paradise    640x480: 3\n");
#endif
		printf("Exit...............: 0\n");
		printf("Enter mode (%c) :", _Mode);
		fflush(stdout);
		i = getche();
		printf("\n");
		if (i != CR && i != LF)
			_Mode = i;
		if (_Mode < '1' || _Mode > '6')
			break;
		switch (_Mode) {
			case '1':
				vga_mode = _MRES256COLOR;
				scr_line = 22;
				break;
#if	defined(ET4000)	/* Tseng Labs Super VGA */
			case '2':
				vga_mode = _ET640x350;
				scr_line = 23;
				break;
			case '3':
				vga_mode = _ET640x400;
				scr_line = 23;
				break;
			case '4':
				vga_mode = _ET640x480;
				scr_line = 28;
				break;
			case '5':
				vga_mode = _ET800x600;
				scr_line = 35;
				break;
			case '6':
				vga_mode = _ET1024x768;
				scr_line = 45;
				break;
#else		/* Paradise Super VGA */
			case '2':
				vga_mode = _PA640x400;
				scr_line = 23;
				break;
			case '3':
				vga_mode = _PA640x480;
				scr_line = 28;
				break;
#endif
		}
	}
	free_matrix(_Array, _Ymin, _Xmin);
	free(row);
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
