#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "nr.h"	/* function declarations for nr.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
/*--------------------------------------------------------------------*/
#define	GAM	0.08	/* guess the width, gam, of the OTF */
#define	POW	1.7	/* guess the power, pow, of the OTF */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	double gam = GAM;
	double pow = POW;
	double a2  = 0.0;
	double b2  = 0.0;
	int i, y;
	int xcen, ycen, radmax, noisemin, noisemax;
	int stx, sty, extx, exty;
	int id, type, npix, nlin;
	double Blank, dgam, dpow, da2, db2;
	char *image, *str;
	COMPLEX **a;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "otf: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "g:p:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'g':	sscanf(optarg, "%lf", &gam); break;
			case 'p':	sscanf(optarg, "%lf", &pow); break;
			case '?':	opt_err++; break;	/* unknown option */
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 6) {
		printf(
		"Usage: otf -g Gamma -p Power image Xc Yc RadMax NoiseMin NoiseMax\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);	/* image name */
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "otf: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	sscanf(arg[1], "%d", &xcen);
	sscanf(arg[2], "%d", &ycen);
	sscanf(arg[3], "%d", &radmax);	/* aperture radius */
	sscanf(arg[4], "%d", &noisemin);	/* inner rad of noise annulus */
	sscanf(arg[5], "%d", &noisemax);	/* outer rad of noise annulus */

	if (radmax <= 0) {
		fprintf(stderr, "otf: radmax <= 0\n");
		exit(EXIT_FAILURE);
	}
	if (noisemin < radmax) {
		fprintf(stderr, "otf: noisemin < radmax\n");
		exit(EXIT_FAILURE);
	}
	if (noisemax <= noisemin) {
		fprintf(stderr, "otf: noisemax <= noisemin\n");
		exit(EXIT_FAILURE);
	}
	stx = xcen;
	sty = ycen - noisemax;
	if (stx < 0 || sty < 0) {
		fprintf(stderr, "otf: Noise annulus extends outside image\n");
		exit(EXIT_FAILURE);
	}
	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "otf: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	if ((id = iopen(image, 0, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "otf: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"otf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type != 5) {
		fprintf(stderr,
			"otf: %s has incorrect type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "otf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank = blank(image);
	extx = noisemax + 1;
	exty = 2*noisemax + 1;
	if (stx + extx > npix || sty + exty > nlin) {
		fprintf(stderr, "otf: Noise annulus is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "otf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (!window(id, stx, sty, extx, exty)) {
		fprintf(stderr, "otf: Window is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "otf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((a = cmatrix(sty,sty+exty-1,stx,stx+extx-1)) == (COMPLEX **)0) {
		fprintf(stderr, "otf: Allocation failed in matrix()-2\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "otf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	for (y = sty; y < (sty+exty); ++y)
		getline(id, y-sty, &a[y][stx]);	/* fetch new line */
	iclose(id);
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "otf: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	gam *= nlin;	/* now in delta-v units */
	str = g4fit(a,	Blank, xcen, ycen, radmax, noisemin, noisemax,
			&gam, &pow, &a2, &b2, &dgam, &dpow, &da2, &db2);
	if (str != (char *)0) {
		fprintf(stderr, "otf: %s", str);
		exit(EXIT_FAILURE);
	}
	free_cmatrix(a, sty, stx);
	gam /= nlin;	/* now back in (delta-y)^-1 units */
	dgam /= nlin;	/* dgma should be in the same units */
	printf(
	"gam= %6.5f+-%6.5f pow= %5.3f+-%5.3f a2= %5.3f+-%5.3f b2= %5.3f+-%5.3f\n",
		gam, dgam, pow, dpow, a2, da2, b2, db2);
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
