#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "daa.h"	/* function declarations for nr.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	char *image1, *image2, *ptr;
	void *buffer;
	float **a;
	int i, x, y;
	int id1, id2, ncols, nrows, type;
	int start[2], dim[2];
	unsigned int err;
	double Blank1;
	const float FltBlank = (float)FLT_BLANK;
	const double FltMax = 0.9*FLT_BLANK;
	float Init_Value = FltBlank;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if (argc < 3) {
		printf("Usage: power image1 image2\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image1 = truncname(argv[1]);	/* image name */
	if (strpbrk(image1, punct) != (char *)0) {
		fprintf(stderr, "power: Illegal image name: %s\n", image1);
		exit(EXIT_FAILURE);
	}

	image2 = truncname(argv[2]);	/* image name */
	if (strpbrk(image2, punct) != (char *)0) {
		fprintf(stderr, "power: Illegal image name: %s\n", image2);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "power: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	if ((id1 = iopen(image1, 0, &type, &ncols, &nrows)) < 0) {
		fprintf(stderr, "power: Can't open %s\n", image1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"power: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type < 1 || type > 5) {
		fprintf(stderr,
			"power: %s has illegal type = %d\n", image1, type);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"power: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank1 = blank(image1);

	if ((buffer = (void *)calloc(ncols, bytepix(type))) == (void *)0) {
		fprintf(stderr, "power: Allocation of buffer failed\n");
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"power: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	dim[0] = nrows;	/* number of rows: y-dimension */
	dim[1] = ncols;	/* number of columns: x-dimension */
	start[0] = 0;	/* first index of y-dimension */
	start[1] = 0;	/* first index of x-dimension */

	if ((a = (float **)daa(sizeof(float), 2, dim, start,
			&err, &ptr, &Init_Value)) == (float **)0) {
		fprintf(stderr, "power: daa: Error code = %d\n", err);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"power: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (xist(image2)) {
		fprintf(stderr, "power: %s: Existing image\n", image2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "power: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (cpi(image1, image2) < 1) {	/* copy image header */
		fprintf(stderr,
			"power: %s.hdr -> %s.hdr copy failed\n", image1, image2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "power: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (setimagefile(image2, "") == 0) {	/* erase FILE value */
		fprintf(stderr,
			"power: Can't erase old FILE value for %s\n", image2);
		rmi(image2);	/* remove image header */
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "power: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((id2 = icreat(image2, 1, 3, ncols, nrows)) < 0) {
		fprintf(stderr, "power: Can't create %s\n", image2);
		rmi(image2);	/* remove image header */
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "power: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "power: update: Disk header not created\n");

	if (blank(image2) <= FltMax) {
		fprintf(stderr, "power: No BLANK in header for %s\n", image2);
		iclose(id2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "power: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (y = 0; y < nrows; ++y) {
		getline(id1, y, buffer);	/* fetch new row */
		switch (type) {
			case 1:	/* unsigned char */
				{
					const UCHAR *q = (UCHAR *)buffer;
					double d;
					for (x = 0; x < ncols; ++x) {
						d = (double)q[x];
						d *= d;
						a[y][x] = (d > FltMax)? FltBlank: (float)d;
					}
				}
				break;
			case 2:	/* short int */
				{
					const short *q = (short *)buffer;
					const short b = (short)Blank1;
					double d;
					for (x = 0; x < ncols; ++x) {
						d = (double)q[x];
						d *= d;
						a[y][x] = (q[x] == b || d > FltMax)? FltBlank: (float)d;
					}
				}
				break;
			case 3:	/* float */
				{
					const float *q = (float *)buffer;
					const float h = (float) (0.9*Blank1);
					double d;
					for (x = 0; x < ncols; ++x) {
						d = (double)q[x];
						d *= d;
						a[y][x] = (q[x] > h || d > FltMax)? FltBlank: (float)d;
					}
				}
				break;
			case 4:	/* LONG int */
				{
					const LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank1;
					double d;
					for (x = 0; x < ncols; ++x) {
						d = (double)q[x];
						d *= d;
						a[y][x] = (q[x] == b || d > FltMax)? FltBlank: (float)d;
					}
				}
				break;
			case 5:	/* complex */
				{
					const COMPLEX *q = (COMPLEX *)buffer;
					const float h = (float) (0.9*Blank1);
					double d, dy;
					for (x = 0; x < ncols; ++x) {
						d = (double)q[x].x;
						d *= q[x].x;
						dy = (double)q[x].y;
						dy *= q[x].y;
						d += dy;
						a[y][x] = (q[x].x > h || q[x].y > h || d > FltMax)?
									FltBlank: (float)d;
					}
				}
				break;
		}
	}

	for (y = 0; y < nrows; ++y)
		putline(id2, y, &a[y][0]);	/* save next row */

	iclose(id2);
	free(ptr);
	free(buffer);
	iclose(id1);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "power: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	fflush(stdout);
	exit(EXIT_SUCCESS);
}
