#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <ctype.h>	/* ctype macros */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stddef.h>	/* defines useful macros */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define	PSIZE	128	/* maximum path length */
#define	FNSIZE	51	/* maximum file name size */
#define	NFILES	512	/* maximum number of FITS files */
#define	RSIZE	2880	/* number of bytes in FITS record */
#define	MAXREC	128	/* maximum number of records in array */
/*--------------------------------------------------------------------*/
typedef union {	/* LONG or float word */
	LONG  l;
	float f;
} LONGWORD;
/*--------------------------------------------------------------------*/
typedef struct {	/* keyword structure */
	short type;
	char *key;
} KEY;
/*--------------------------------------------------------------------*/
#define	NKEYS	191	/* number of FITS keywords, plus "FILE" */
#define	LOGICAL	1	/* logical variable */
#define	INTEGER	2	/* integer variable */
#define	REAL	3	/* real variable */
#define	CHARACTER	4	/* character variable, normally 8 characters */
/*--------------------------------------------------------------------*/
static KEY  words[] = {	/* table containing FITS keywords */
	LOGICAL  , "SIMPLE",    /* +Standard FITS format */
	INTEGER  , "BITPIX",    /* +Number of bits storing pixel values */
	INTEGER  , "NAXIS",     /* +Number of axes in frame; 0 if no data follow */
	INTEGER  , "NAXIS1",    /* +Number of pixels/axis; 0 if no data follow */
	INTEGER  , "NAXIS2",    /* +Number of pixels/axis; 0 if no data follow */
	LOGICAL  , "EXTEND",    /* +Standard extensions are possible */
	INTEGER  , "BLANK",     /* +Specifies that the pixel value is undefined */
	REAL     , "BSCALE",    /* +PhysicalValue=ArrayValue*BSCALE+BZERO */
	REAL     , "BZERO",     /* +PhysicalValue=ArrayValue*BSCALE+BZERO */
	REAL     , "DATAMIN",   /* +Minimum valid physical value represented */
	REAL     , "DATAMAX",   /* +Maximum valid physical value represented */
	INTEGER  , "OPSIZE",    /* &PSIZE of original image */
	CHARACTER, "FITSDATE",  /* &Date FITS file was created */
	CHARACTER, "FILENAME",  /* &Original GEIS header file name with _cvt */
	CHARACTER, "ODATTYPE",  /* &Original datatype */
	INTEGER  , "SDASMGNU",  /* &GCOUNT of original image */
	CHARACTER, "DADSFILE",  /* & */
	CHARACTER, "DADSCLAS",  /* & */
	CHARACTER, "DADSDATE",  /* & */
	REAL     , "CD1_1",     /* & */
	REAL     , "CD1_2",     /* & */
	REAL     , "CD2_1",     /* & */
	REAL     , "CD2_2",     /* & */
	LOGICAL  , "MIR_REVR",  /* & */
	REAL     , "ORIENTAT",  /* & */
	INTEGER  , "FILLCNT",   /* & */
	INTEGER  , "ERRCNT",    /* & */
	REAL     , "FPKTTIME",  /* & */
	REAL     , "LPKTTIME",  /* & */
	INTEGER  , "DETECTOR",  /* & */
	REAL     , "DEZERO",    /* & */
	REAL     , "BIASEVEN",  /* & */
	REAL     , "BIASODD",   /* & */
	REAL     , "GOODMIN",   /* & */
	REAL     , "GOODMAX",   /* & */
	REAL     , "DATAMEAN",  /* & */
	INTEGER  , "GPIXELS",   /* & */
	INTEGER  , "SOFTERRS",  /* & */
	INTEGER  , "CALIBDEF",  /* & */
	INTEGER  , "STATICD",   /* & */
	INTEGER  , "ATODSAT",   /* & */
	INTEGER  , "DATALOST",  /* & */
	INTEGER  , "BADPIXEL",  /* & */
	INTEGER  , "OVERLAP",   /* & */
	CHARACTER, "PHOTMODE",  /* & */
	REAL     , "PHOTFLAM",  /* & */
	REAL     , "PHOTZPT",   /* & */
	REAL     , "PHOTPLAM",  /* & */
	REAL     , "PHOTBW",    /* & */
	REAL     , "MEDIAN",    /* & */
	REAL     , "MEDSHADO",  /* & */
	REAL     , "HISTWIDE",  /* & */
	REAL     , "SKEWNESS",  /* & */
	REAL     , "MEANC10",   /* & */
	REAL     , "MEANC25",   /* & */
	REAL     , "MEANC50",   /* & */
	REAL     , "MEANC100",  /* & */
	REAL     , "MEANC200",  /* & */
	REAL     , "MEANC300",  /* & */
	REAL     , "BACKGRND",  /* & */
	CHARACTER, "ROOTNAME",  /* &Rootname of the observation set */
	CHARACTER, "FILETYPE",  /* &shp, ext, edq, sdq, sci */
	CHARACTER, "MODE",      /* &instr. mode: FULL (full res.), AREA (area int.) */
	CHARACTER, "SERIALS",   /* &serial clocks: ON, OFF */
	CHARACTER, "IMAGETYP",  /* &DARK/BIAS/IFLAT/UFLAT/VFLAT/KSPOT/EXT/ECAL */
	CHARACTER, "CDBSFILE",  /* &GENERIC/BIAS/DARK/FLAT/MASK/NO */
	INTEGER  , "PKTFMT",    /* &packet format code */
	CHARACTER, "FILTNAM1",  /* &first filter name */
	CHARACTER, "FILTNAM2",  /* &second filter name */
	INTEGER  , "FILTER1",   /* &first filter number (0-48) */
	INTEGER  , "FILTER2",   /* &second filter number (0-48) */
	INTEGER  , "FILTROT",   /* &partial filter rotation angle (degrees) */
	INTEGER  , "LRFWAVE",   /* &linear ramp filter wavelength */
	REAL     , "UCH1CJTM",  /* &TEC cold junction #1 temperature (Celsius) */
	REAL     , "UCH2CJTM",  /* &TEC cold junction #2 temperature (Celsius) */
	REAL     , "UCH3CJTM",  /* &TEC cold junction #3 temperature (Celsius) */
	REAL     , "UCH4CJTM",  /* &TEC cold junction #4 temperature (Celsius) */
	REAL     , "UBAY3TMP",  /* &bay 3 A1 temperature (deg C) */
	CHARACTER, "KSPOTS",    /* &Status of Kelsall spot lamps: ON, OFF */
	CHARACTER, "SHUTTER",   /* &Shutter in place at the beginning of the exp */
	REAL     , "ATODGAIN",  /* &Analog to Digital Gain (Electrons/DN) */
	CHARACTER, "MASKCORR",  /* &Do mask correction: PERFORM, OMIT, COMPLETE */
	CHARACTER, "ATODCORR",  /* &Do A-to-D correction: PERFORM, OMIT, COMPLETE */
	CHARACTER, "BLEVCORR",  /* &Do bias level correction: PERFORM, OMIT, COMPLE */
	CHARACTER, "BIASCORR",  /* &Do bias correction: PERFORM, OMIT, COMPLETE */
	CHARACTER, "DARKCORR",  /* &Do dark correction: PERFORM, OMIT, COMPLETE */
	CHARACTER, "FLATCORR",  /* &Do flat field correction: PERFORM, OMIT, COMPLE */
	CHARACTER, "SHADCORR",  /* &Do shaded shutter correction: PERFORM, OMIT, CO */
	CHARACTER, "DOSATMAP",  /* &Output saturated pixel map: PERFORM, OMIT, COMP */
	CHARACTER, "DOPHOTOM",  /* &Fill photometry keywords: PERFORM, OMIT, COMPLE */
	CHARACTER, "DOHISTOS",  /* &Make histograms: PERFORM, OMIT, COMPLETE */
	CHARACTER, "OUTDTYPE",  /* &Output image datatype: REAL, LONG, SHORT */
	CHARACTER, "MASKFILE",  /* &name of the input DQF of known bad pixels */
	CHARACTER, "ATODFILE",  /* &name of the A-to-D conversion file */
	CHARACTER, "BLEVFILE",  /* &Engineering file with extended register data */
	CHARACTER, "BLEVDFIL",  /* &Engineering file DQF */
	CHARACTER, "BIASFILE",  /* &name of the bias frame reference file */
	CHARACTER, "BIASDFIL",  /* &name of the bias frame reference DQF */
	CHARACTER, "DARKFILE",  /* &name of the dark reference file */
	CHARACTER, "DARKDFIL",  /* &name of the dark reference DQF */
	CHARACTER, "FLATFILE",  /* &name of the flat field reference file */
	CHARACTER, "FLATDFIL",  /* &name of the flat field reference DQF */
	CHARACTER, "SHADFILE",  /* &name of the reference file for shutter shading */
	CHARACTER, "PHOTTAB",   /* &name of the photometry calibration table */
	CHARACTER, "GRAPHTAB",  /* &the HST graph table */
	CHARACTER, "COMPTAB",   /* &the HST components table */
 	INTEGER  , "SATURATE",  /* &Data value at which saturation occurs */
	REAL     , "USCALE",    /* &Scale factor for output image */
	REAL     , "UZERO",     /* &Zero point for output image */
	INTEGER  , "READTIME",  /* &Length of time for CCD readout in clock ticks */
	REAL     , "PA_V3",     /* &position angle of v3 axis of HST */
	REAL     , "RA_SUN",    /* &right ascension of the sun (deg) */
	REAL     , "DEC_SUN",   /* &declination of the sun (deg) */
	REAL     , "EQNX_SUN",  /* &equinox of the sun */
	LOGICAL  , "MTFLAG",    /* &moving target flag */
	REAL     , "EQRADTRG",  /* &equatorial radius of target */
	REAL     , "FLATNTRG",  /* &flattening of target */
	REAL     , "NPDECTRG",  /* &north pole declination of target */
	REAL     , "NPRATRG",   /* &north pole right ascension of target */
	REAL     , "ROTRTTRG",  /* &rotation rate of target */
	REAL     , "LONGPMER",  /* &longitude of prime meridian */
	REAL     , "EPLONGPM",  /* &epoch of longitude of prime meridian */
	REAL     , "SURFLATD",  /* &surface feature latitude */
	REAL     , "SURFLONG",  /* &surface feature longitude */
	REAL     , "SURFALTD",  /* &surface feature altitude */
	INTEGER  , "PODPSFF",   /* &0=(no podps fill), 1=(podps fill present) */
	INTEGER  , "STDCFFF",   /* &0=(no st dcf fill), 1=(st dcf fill present) */
	CHARACTER, "STDCFFP",   /* &st dcf fill pattern (hex) */
	INTEGER  , "RSDPFILL",  /* &bad data fill value for calibrated images */
	INTEGER  , "UEXPODUR",  /* &Commanded duration of exposure (seconds) */
	INTEGER  , "NSHUTA17",  /* &Number of AP17 shutter B closes */
	INTEGER  , "UEXPOTIM",  /* &Major frame pulse time preceding exposure start */
	CHARACTER, "PSTRTIME",  /* &Predicted obs. start time  (yyyy:ddd:hh:mm:ss) */
	CHARACTER, "PSTPTIME",  /* &Predicted obs. stop  time  (yyyy:ddd:hh:mm:ss) */
	REAL     , "SUNANGLE",  /* &angle between sun and V1 axis (deg) */
	REAL     , "MOONANGL",  /* &angle between moon and V1 axis (deg) */
	REAL     , "SUN_ALT",   /* &altitude of the sun above Earth's limb (deg) */
	CHARACTER, "FGSLOCK",   /* &commanded FGS lock (FINE,COARSE,GYROS,UNKNOWN) */
	CHARACTER, "TARGNAME",  /* &proposer's target name */
	REAL     , "RA_TARG",   /* &right ascension of the target (deg) (J2000) */
	REAL     , "DEC_TARG",  /* &declination of the target (deg) (J2000) */
	INTEGER  , "PROPOSID",  /* &PEP proposal identifier */
	CHARACTER, "PEP_EXPO",  /* &PEP exposure identifier including sequence */
	CHARACTER, "LINENUM",   /* &PEP proposal line number */
	CHARACTER, "SEQLINE",   /* &PEP line number of defined sequence */
	CHARACTER, "SEQNAME",   /* &PEP define/use sequence name */
	CHARACTER, "BUNIT",     /* +Unit after application of BSCALE and BZERO */
	REAL     , "CRVAL1",    /* +CTYPE1 coordinate at the reference point CRPIX1 */
	REAL     , "CRPIX1",    /* +Location of a reference point along axis-1 */
	REAL     , "CDELT1",    /* +Derivative of CTYPE1 wrt the pixel index at CRPIX1 */
	REAL     , "CROTA1",    /* +Rotation angle between axis-1 and direction of CTYPE1 */
	CHARACTER, "CTYPE1",    /* +Name of coordinate represented by axis-1 */
	REAL     , "CRVAL2",    /* +CTYPE2 coordinate at the reference point CRPIX2 */
	REAL     , "CRPIX2",    /* +Location of a reference point along axis-2 */
	REAL     , "CDELT2",    /* +Derivative of CTYPE2 wrt the pixel index at CRPIX2 */
	REAL     , "CROTA2",    /* +Rotation angle between axis-2 and direction of CTYPE2 */
	CHARACTER, "CTYPE2",    /* +Name of coordinate represented by axis-2 */
	REAL     , "EQUINOX",   /* +Equinox in years for the celestial coordinate system */
	REAL     , "EPOCH",     /* +Deprecated: Use EQUINOX instead */
	REAL     , "JD-START",  /* - */
	REAL     , "JD-STOP",   /* - */
	REAL     , "MJD-OBS",   /* ?MJD start */
	REAL     , "TM-START",  /* ?Exposure Time start */
	REAL     , "EXPSTART",  /* &Exposure start time (Modified Julian Date) */
	REAL     , "EXPEND",    /* &Exposure end time (Modified Julian Date) */
	REAL     , "EXPTIME",   /* &Exposure duration (seconds)--calculated */
	CHARACTER, "EXPFLAG",   /* &Exposure interruption indicator */
	REAL     , "DARKTIME",  /* &Dark time (seconds) */
	REAL     , "GAIN",      /* - */
	REAL     , "RDNOISE",   /* - */
	REAL     , "RA",        /* ? */
	REAL     , "DEC",       /* ? */
	REAL     , "AIRMASS",   /* ? */
	CHARACTER, "OBJECT",    /* +Name of the object observed */
	CHARACTER, "FILTER",    /* ? */
	CHARACTER, "GRISM",     /* - */
	CHARACTER, "TIMESYS",   /* +Principal time system; default is UTC */
	CHARACTER, "DATE",      /* +YYYY-MM-DDThh:mm:ss UTC creation date of HDU */
	CHARACTER, "DATE-OBS",  /* +YYYY-MM-DDThh:mm:ss UTC at start of observation */
	CHARACTER, "TIME-OBS",  /* &UT time of start of observation (hh:mm:ss) */
	CHARACTER, "RADECSYS",  /* &Coordinate reference frame */
	CHARACTER, "UTC",       /* ?UTC at start (sec) */
	CHARACTER, "LST",       /* ?LST at start (sec) */
	CHARACTER, "ORIGIN",    /* +Organization creating the FITS file */
	CHARACTER, "INSTRUME",  /* +Instrument used to acquire the data */
	CHARACTER, "DISPAXIS",  /* ? */
	CHARACTER, "TELESCOP",  /* +Telescope used to acquire the data */
	CHARACTER, "OBSERVER",  /* +Who acquired the data? */
	CHARACTER, "COMMENT",   /* + */
	CHARACTER, "IMAGE",     /* - */
	CHARACTER, "FILE"
};
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static const char *FileName(const char *path)
#else	/* old style */
static const char *FileName(path)	/* file name from full path name */
const char *path;
#endif
{
	const char *p = path;
	while (*p != '\0')
		p++;
	while (p != path && *p != '/')
		p--;
	if (*p == '/')
		p++;
	return p;
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static int KeyType(const char *keyw)
#else	/* old style */
static int KeyType(keyw)	/* value type as defined in FITS table */
const char *keyw;	/* keyword */
#endif
{
	int i;
	for (i = 0; i < NKEYS; i++)	/* search keyword table */
		if (strcmp(keyw, words[i].key) == 0)
			return(words[i].type);
	return(0);	/* keyword is not in FITS table */
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static char *KeyWord(char *keyw, const char *card)
#else	/* old style */
static char *KeyWord(keyw, card)	/* extract keyword from card image */
char *keyw;	/* keyword */
const char *card;	/* card image */
#endif
{
	int i;
	for (i = 0; i < 8; i++)	/* copy keyword field */
		keyw[i] = card[i];
	for (--i; (i >= 0) && (keyw[i] == ' '); i--)
		;	/* trailing spaces */
	keyw[++i] = '\0';	/* terminate keyword */
	return(keyw);
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static char *Logical(char *field, const char *card)
#else	/* old style */
static char *Logical(field, card)	/* copy logical field from card */
char *field;	/* logical field */
const char *card;	/* card image */
#endif
{
	field[0] = card[29];	/* should be 'T' or 'F' */
	field[1] = '\0';	/* terminate field */
	return(field);
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static char *Number(char *field, const char *card)
#else	/* old style */
static char *Number(field, card)	/* copy number field from card */
char *field;	/* integer or real number field */
const char *card;	/* card image */
#endif
{
	int i, j;
	for (i = 0; i < 20; i++)	/* copy number field */
		field[i] = card[10+i];
	field[20] = '\0';	/* terminate field */
	for (j = 0; (field[j] == ' '); j++)
		;	/* skip leading spaces */
	for (i = 0; j <= 20; i++, j++)
		field[i] = field[j];	/* shift number to start of field */
	return(field);
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static char *Character(char *field, const char *card)
#else	/* old style */
static char *Character(field, card)	/* copy character field from card */
char *field;	/* character field */
const char *card;	/* card image */
#endif
{
	int i;
	if (card[10] == '\'') {	/* check if character variable */
		for (i = 0; i < 69 && card[11+i] != '\''; i++)
			field[i] = card[11+i];	/* copy until next ' */
		for (--i; (i > 0) && (field[i] == ' '); i--)
			;	/* back-skip trailing spaces, but not the first one */
		field[++i] = '\0';	/* terminate field */
	} else	/* not a character variable */
		field[0] = '\0';
	return(field);
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static char *Comment(char *comm, const char *card)
#else	/* old style */
static char *Comment(comm, card)	/* copy comment field from card */
char *comm;	/* comment field */
const char *card;	/* card image */
#endif
{
	int i, j = 8;
	if (card[j] == '=') {	/* is it a comment card? */
		j = 10;
		if (card[j] == '\'') {	/* check if character field */
			for (++j; (j < 80) && (card[j] != '\''); j++)
				;	/* skip until ' */
			if (j < 80)	/* skip ', if found */
				j++;
		}
		for (; (j < 80) && (card[j] != '/'); j++)
			;	/* skip until '/' */
		if (j < 80)	/* skip '/', if found */
			j++;
	}
	for (; (j < 80) & (card[j] == ' '); j++)
		;	/* skip leading spaces */
	for (i = 0; j < 80; i++, j++)	/* copy comment field */
		comm[i] = card[j];
	for (--i; (i >= 0) && (comm[i] == ' '); i--)
		;	/* skip trailing spaces */
	comm[++i] = '\0';	/* terminate comment field */
	return(comm);
}
/*--------------------------------------------------------------------*/
static char ifile[PSIZE] = {(char)0};	/* image file path */
static char ipath[PSIZE] = {(char)0};	/* image directory path */
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void SetPath(const int i)
#else	/* old style */
static void SetPath(i)	/* assign image path from environment */
const int i;	/* select i'th path name */
#endif
{
	char **p;
	int n;
	if((p = getpath("IMGPATH", &n)) == (char **)0)
		strcpy(ipath, ".");	/* copy current directory path */
	else if(i < n)
		strcpy(ipath, p[i]);	/* copy i'th environment path */
	else
		strcpy(ipath, ".");	/* copy current directory path */
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void AssignImageFile(const char *file)
#else	/* old style */
static void AssignImageFile(file)	/* cat image file to path name */
const char *file;	/* image file */
#endif
{
	strcpy(ifile, ipath);	/* image directory path */
	if(strlen(ifile) > 0)
		strcat(ifile, "/");	/* append "/" */
	strcat(ifile, file);	/* append image file */
}
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
static void ReOrder(void *buffer,
const int nrec,
const int type,
const double scale,
const double zero,
const double Blank)
#else	/* old style */
static void ReOrder(buffer, nrec, type, scale, zero, Blank)
void *buffer;	/* DATA buffer */
const int nrec;	/* number of FITS records */
const int type;	/* bits per pixel */
const double scale;	/* BSCALE value */
const double zero;	/* BZERO value */
const double Blank;	/* BLANK value */
#endif
{
	int scaling = (scale == 0.0 || scale == 1.0)? 0: 1;
	switch (type) {
		case 2:
			if (scaling) {	/* scaling of short to short data */
				short w, m = 0xFF;
				short b = (short) Blank;
				short *p = (short *) buffer;
				short *e = p + 1440 * nrec;
				char  *q = (char *) buffer;
				for (; p < e; p++) {
					w = *q++;	/* high order byte */
					w <<= 8;
					w |= (*q++ & m);	/* low order byte */
					if (w != b) {
						double d = w;
						d *= scale;
						d += zero;
						*p = (short) ((d < 0.0)? d - 0.5: d + 0.5);
					} else *p = w;
				}
			} else {	/* no scaling of short data */
				short w, m = 0xFF;
				short b = (short) Blank;
				short z = (short) ((zero < 0.0)? zero-0.5: zero+0.5);
				short *p = (short *) buffer;
				short *e = p + 1440 * nrec;
				char  *q = (char *) buffer;
				for (; p < e; p++) {
					w = *q++;	/* high order byte */
					w <<= 8;
					w |= (*q++ & m);	/* low order byte */
					*p = (w != b)? w + z: w;
				}
			}
			break;
		case 3:
			if (scaling) {	/* scaling of LONG to float data */
				LONG w, m = 0xFF;
				LONG b = (LONG) Blank;
				float *p = (float *) buffer;
				float *e = p + 720 * nrec;
				char *q = (char *) buffer;
				for (; p < e; p++) {
					w = *q++;	/* most significant byte */
					w <<= 24;
					w |= (*q++ & m) << 16;
					w |= (*q++ & m) << 8;
					w |= (*q++ & m);	/* least significant byte */
					*p = (float) ((w != b)? scale*w + zero: FLT_BLANK);
				}
			} else {	/* no scaling of float data */
				LONGWORD w;
				LONG m = 0xFF;
				float *p = (float *) buffer;
				float *e = p + 720 * nrec;
				char *q = (char *) buffer;
				for (; p < e; p++) {
					w.l = *q++;	/* most significant byte */
					w.l <<= 24;
					w.l |= (*q++ & m) << 16;
					w.l |= (*q++ & m) << 8;
					w.l |= (*q++ & m);	/* least significant byte */
					*p = w.f;
				}
			}
			break;
		case 4:
			if (scaling) {	/* scaling of LONG to Long data */
				LONG w, m = 0xFF;
				LONG b = (LONG) Blank;
				LONG *p = (LONG *) buffer;
				LONG *e = p + 720 * nrec;
				char *q = (char *) buffer;
				for (; p < e; p++) {
					w = *q++;	/* most significant byte */
					w <<= 24;
					w |= (*q++ & m) << 16;
					w |= (*q++ & m) << 8;
					w |= (*q++ & m);	/* least significant byte */
					if (w != b) {
						double d = w;
						d *= scale;
						d += zero;
						*p = (LONG) ((d < 0.0)? d - 0.5: d + 0.5);
					} else *p = w;
				}
			} else {	/* no scaling of Long data */
				LONG w, m = 0xFF;
				LONG b = (LONG) Blank;
				LONG z = (LONG) ((zero < 0.0)? zero-0.5: zero+0.5);
				LONG *p = (LONG *) buffer;
				LONG *e = p + 720 * nrec;
				char *q = (char *) buffer;
				for (; p < e; p++) {
					w = *q++;	/* most significant byte */
					w <<= 24;
					w |= (*q++ & m) << 16;
					w |= (*q++ & m) << 8;
					w |= (*q++ & m);	/* least significant byte */
					*p = (w != b)? w + z: w;
				}
			}
			break;
		default:
			break;
	}
}
/*--------------------------------------------------------------------*/
static char keyw[9];	/* FITS keyword plus terminating '\0' */
static char field[73];	/* value field plus terminating '\0' */
static char comm[73];	/* comment field plus terminating '\0' */
static char text[73];	/* text string plus terminating '\0' */
static char *tmp = "itemp";
static char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
static char ext[5] = {'.', 'f', 'i', 't', '\0'};
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;

	int option, opt_err = 0, n_arg = 0;
	int image_key = 0;	/* use IMAGE keyword */
	int swap_on = 1;	/* swap words and bytes according to FITS */
	int scale_on = 1; /* data is scaled according to BSCALE and BZERO */
	char **arg;

	char **itab, (*record)[80];
	char *image, *im, *fitsfile, name[FNSIZE];
	char *fitspath = getenv("FITSPATH");
	void *rec = (void *)0, *ptr = (void *)0;
	FILE *istream, *ostream, *tstream;
	long isize;
	unsigned asize, rem, mrec, mr, maxr, fitsfilelength;
	int bpix, ndim, ncol, nrow, type;
	int narray, t, i, k, n, ntab;
	int rw = 1;	/* write only */
	int cpf = -1;	/* create and extend */
	double scale, zero, Blank;

	if (fitspath == NULL || strlen(fitspath) == 0)
		fitsfilelength = FNSIZE;
	else
		fitsfilelength = strlen(fitspath) + 1 + FNSIZE;
	if ((fitsfile = (char *)calloc(fitsfilelength, sizeof(char))) == NULL) {
		fprintf(stderr, "rdfits: calloc: Can't allocate file name\n");
		exit(EXIT_FAILURE);
	}

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "rdfits: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "nise:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'n':	swap_on = 0; break;	/* no word or byte swap */
			case 'i':	image_key++; break;	/* use IMAGE keyword */
         case 's':   scale_on = 0; break; /* no data scaling */
			case 'e':	strncpy(ext+1, optarg, 3); break;	/* extension */
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 1) {
		printf(
		"Usage: rdfits -n -i -s -e ext {image1|@file} [image2] [image3]...\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	if (*arg[0] == '@') {	/* read list of FITS files from file */
		fitsfile[0] = '\0';
		if (fitspath != NULL && strlen(fitspath) > 0) {
			strcat(fitsfile, fitspath);
			strcat(fitsfile, "/");
		}
		strncat(fitsfile, arg[0]+1, FNSIZE-1);
		if ((tstream = fopen(fitsfile, "r")) == (FILE *)0) {	/* text file */
			fprintf(stderr, "rdfits: can't open %s\n", fitsfile);
			exit(EXIT_FAILURE);
		}
		if ((itab = (char **)calloc(NFILES, sizeof(char *))) == (char **)0) {
			fprintf(stderr, "rdfits: calloc: Allocation failed\n");
			exit(EXIT_FAILURE);
		}
		for (n = 0; n < NFILES && fscanf(tstream, "%50s", fitsfile) != EOF; ) {
			k = strlen(fitsfile);
			if ((k -= (int)strlen(ext)) > 0 && strcmp(fitsfile+k, ext) == 0) {
				fitsfile[k] = '\0';	/* truncate extension */
				if ((itab[n] = (char *)calloc(strlen(fitsfile)+1,
						sizeof(char))) == (char *)0) {
					fprintf(stderr, "rdfits: calloc: Allocation failed\n");
					exit(EXIT_FAILURE);
				}
				strcpy(itab[n], fitsfile);
				n++;
			}
		}
		fclose(tstream);	/* close text file */
	} else {	/* read list of FITS files from command line */
		n = n_arg;	/* number of FITS files */
		if ((itab = (char **)calloc(n, sizeof(char *))) == (char **)0) {
			fprintf(stderr, "rdfits: calloc: Allocation failed\n");
			exit(EXIT_FAILURE);
		}
		for (i = 0; i < n; ++i)
			itab[i] = arg[i];
	}
	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "rdfits: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}
	for (ntab = 0, i = 0; i < n; ++i) {
		image = truncname(itab[i]);
		if (strpbrk(image, punct) != (char *)0)
			printf("%s is not a legal image name\n", image);
		else
			itab[ntab++] = image;
	}
	maxr = UINT_Max/RSIZE;	/* maximum possible number of records */
	if (maxr > MAXREC)	/* but not more than MAXREC records */
		maxr = MAXREC;
	asize = 0;	/* array size in bytes */
	for (mrec = maxr; mrec > 0; mrec--) {	/* find big array */
		asize = mrec*RSIZE;	/* try this size */
		if ((rec = (void *)malloc(asize)) != (void *)0)
			break;	/* big array allocated */
	}
	if (mrec <= 0) {	/* array has not been allocated */
		fprintf(stderr, "rdfits: malloc: Buffer allocation failed\n");
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "rdfits: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (mrec < maxr) {	/* no space for maximum size */
		mr = mrec;
		mr++;
		mr >>= 1;
		asize = mr*RSIZE;	/* array size reduced by 50% */
		if ((ptr = (void *)realloc(rec, asize)) == (void *)0)
			asize = mrec*RSIZE;	/* reallocation failed */
		else {	/* successful reallocation */
			mrec = mr;	/* number of records reallocated */
			rec = ptr;
		}
	}
	record = (char (*)[80])rec;
	for (n = 0; n < ntab; ++n) {	/* begin converting FITS files */
		image = itab[n];	/* next image name */
		fitsfile[0] = '\0';
		if (fitspath != NULL && strlen(fitspath) > 0) {
			strcat(fitsfile, fitspath);
			strcat(fitsfile, "/");
		}
		strcat(fitsfile, image);
		strcat(fitsfile, ext);	/* FITS file extension */
		if ((istream = fopen(fitsfile, O_RO)) == (FILE *)0) {
			printf("can't open %s\n", fitsfile);
			goto Next_Image;
#if	BUF_SIZ	/* set internal buffer size */
		} else if (setvbuf(istream, NULL, _IOFBF, buffsize())) {
			printf("setvbuf() failed for %s\n", fitsfile);
			fclose(istream);	/* close FITS file */
			goto Next_Image;
#endif
		}
		if (fread(rec, 1, RSIZE, istream) != RSIZE) {	/* 1'st record */
			printf("Error reading FITS header from %s\n", fitsfile);
			fclose(istream);	/* close FITS file */
			goto Next_Image;
		}
		KeyWord(keyw, record[0]);	/* 1'st keyword */
		if (strcmp(keyw, "SIMPLE") != 0) {
			printf("1'st keyword is %s in header of %s\n", keyw, fitsfile);
			fclose(istream);	/* close FITS file */
			goto Next_Image;
		}
		Logical(field, record[0]);
		if (strcmp(field, "T") != 0) {
			printf("SIMPLE  = %s in header of %s\n", field, fitsfile);
			fclose(istream);	/* close FITS file */
			goto Next_Image;
		}
		KeyWord(keyw, record[1]);	/* 2'nd keyword */
		if (strcmp(keyw, "BITPIX") != 0) {
			printf("2'nd keyword is %s in header of %s\n", keyw, fitsfile);
			fclose(istream);	/* close FITS file */
			goto Next_Image;
		}
		Number(field, record[1]);
		sscanf(field, "%d", &bpix);
		KeyWord(keyw, record[2]);	/* 3'rd keyword */
		if (strcmp(keyw, "NAXIS") != 0) {
			printf("3'rd keyword is %s in header of %s\n", keyw, fitsfile);
			fclose(istream);	/* close FITS file */
			goto Next_Image;
		}
		Number(field, record[2]);
		sscanf(field, "%d", &ndim);
		if (ndim < 1 || ndim > 2) {
			printf("NAXIS   = %s in header of %s\n", field, fitsfile);
			fclose(istream);	/* close FITS file */
			goto Next_Image;
		}
		KeyWord(keyw, record[3]);	/* 4'th keyword */
		if (strcmp(keyw, "NAXIS1") != 0) {
			printf("4'th keyword is %s in header of %s\n", keyw, fitsfile);
			fclose(istream);	/* close FITS file */
			goto Next_Image;
		}
		Number(field, record[3]);
		sscanf(field, "%d", &ncol);
		if (ncol < 1) {
			printf("NAXIS1  = %s in header of %s\n", field, fitsfile);
			fclose(istream);	/* close FITS file */
			goto Next_Image;
		}
		if (ndim == 2) {
			KeyWord(keyw, record[4]);	/* 5'th keyword */
			if (strcmp(keyw, "NAXIS2") != 0) {
				printf("5'th keyword is %s in header of %s\n", keyw, fitsfile);
				fclose(istream);	/* close FITS file */
				goto Next_Image;
			}
			Number(field, record[4]);
			sscanf(field, "%d", &nrow);
			if (nrow < 1) {
				printf("NAXIS2  = %s in header of %s\n", field, fitsfile);
				fclose(istream);	/* close FITS file */
				goto Next_Image;
			}
		}
		if (xist(tmp))
			if (idel(tmp) < 0)	/* delete image tmp, if existing */
				rmi(tmp);	/* remove tmp header, if not deleted */
		if (setimagefile(tmp, "") == 0) {	/* erase FILE value */
			fprintf(stderr, "rdfits: FILE value not erased in %s\n", tmp);
			fclose(istream);	/* close FITS file */
			rmi(tmp);	/* remove tmp header */
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "rdfits: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		switch(bpix) {	/* assign pixel data type from BITPIX */
			case   8:	type = 1;	break;	/* unsigned char */
			case  16:	type = 2;	break;	/* short int */
			case -32:	type = 3;	break;	/* float */
			case  32:	type = 4;	break;	/* Long int */
			case -64:	type = 5;	break;	/* complex */
			default :	type = 0;	break;	/* undefined */
		}
		if (setpixtype(tmp, type) == 0) {	/* set pixel type */
			fprintf(stderr, "rdfits: PIXTYPE value not entered in %s\n", tmp);
			fclose(istream);	/* close FITS file */
			rmi(tmp);	/* remove tmp header */
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "rdfits: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		while(1) {	/* enter keywords from FITS header */
			for (k = 0; k < 36; k++) {
				KeyWord(keyw, record[k]);	/* get keyword from card image */
				if (strcmp(keyw, "END") == 0)	/* end of FITS header */
					goto End_Header;
				switch (t = KeyType(keyw)) {	/* get value field */
					case LOGICAL:
						Logical(field, record[k]);
						break;
					case INTEGER:
					case REAL:
						Number(field, record[k]);
						break;
					case CHARACTER:
						Character(field, record[k]);
						break;
					default:
						field[0] = '\0';
						break;
				}
				Comment(comm, record[k]);	/* get comment field */
				if (strcmp(keyw, "FILE") == 0) {	/* special case */
					strcpy(keyw, "IMAGE");
					strcpy(text, FileName(field));
				} else
					strcpy(text, field);	/* value field */
				if (strlen(field) > 0 && strlen(comm) > 0)
					strcat(text, "\t\\ ");
				strcat(text, comm);	/* comment field */
				if (t > 0)
					if (putk(tmp, keyw, text) == (char *)0) {
						fprintf(stderr,
							"rdfits: %s = %s not entered in %s\n",
								keyw, text, tmp);
						fclose(istream);	/* close FITS file */
						rmi(tmp);	/* remove tmp header */
						if (epilog() < 1)	/* update disk headers */
							fprintf(stderr,
								"rdfits: epilog: Disk header(s) not updated\n");
						exit(EXIT_FAILURE);
					}
			}
			if (fread(rec, 1, RSIZE, istream) != RSIZE) {	/* FITS record */
				printf("Error reading FITS header from %s\n", fitsfile);
				fclose(istream);	/* close FITS file */
				rmi(tmp);	/* remove tmp header */
				goto Next_Image;
			}
		}	/* while(1) */
		End_Header:
		Blank = blank(tmp);	/* get BLANK value */
		scale = bscale(tmp);	/* get BSCALE value */
		zero = bbzero(tmp);	/* get BZERO value */
		rmk(tmp, "BSCALE");	/* remove BSCALE keyword */
		rmk(tmp, "BZERO");	/* remove BZERO keyword */
      if (scale_on == 0) { /* no data scaling */
         scale = 1.0;
         zero = 0.0;
      }
		if (type == 4)	/* 32-bit integer */
			if (!(scale == 0.0 || scale == 1.0)) {
				type = 3;	/* 32-bit float */
				setpixtype(tmp, type);
				setbitpix(tmp, -32);	/* Long is converted to float */
				setblank(tmp, FLT_BLANK);
			}
		/* should we get image name from "IMAGE" key */
		if (image_key && (im = iname(tmp)) != (char *)0) {
			strcpy(name, im);	/* copy image name from field buffer */
			for (i = 0; name[i] != 0; i++)	/* convert to lowercase */
				name[i] = tolower(name[i]);
			if (strpbrk(truncname(name), punct) != (char *)0) {
				printf("%s is not a legal image name\n", name);
				fclose(istream);	/* close FITS file */
				rmi(tmp);	/* remove tmp header */
				goto Next_Image;
			} else if (xist(name)) {	/* check if image exists */
				printf("%s is an existing image\n", name);
				fclose(istream);	/* close FITS file */
				rmi(tmp);	/* remove tmp header */
				goto Next_Image;
			} else
				image = name;	/* use image name from header */
		} else if (xist(image)) {	/* image name not from header */
			printf("%s is an existing image\n", image);
			fclose(istream);	/* close FITS file */
			rmi(tmp);	/* remove tmp header */
			goto Next_Image;
		}
		sethistory(tmp, "rdfits");
		sethistory(tmp, image);
		setipath(0);	/* use first token in IMGPATH */
		if ((i = icopy(tmp, image, rw, cpf)) < 0) {
			fprintf(stderr,
				"rdfits: Image file for %s was not created\n", image);
			fclose(istream);	/* close FITS file */
			rmi(tmp);	/* remove tmp header */
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "rdfits: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		} else {	/* image file was created */
			rmi(tmp);	/* remove tmp header */
			iclose(i);	/* close image file */
		}
		SetPath(0);	/* in order to get direct access to image file */
		AssignImageFile(imagefile(image));	/* assign private image file */
		if ((ostream = fopen(ifile, C_WO)) == (FILE *)0) {
			fprintf(stderr,
				"rdfits: can't open image file %s for %s\n",
				ifile, image);
			fclose(istream);	/* close FITS file */
			if (idel(image) < 0)	/* try to delete header and image file */
				rmi(image);	/* remove image header, if not deleted */
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "rdfits: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
#if	BUF_SIZ	/* set internal buffer size */
		} else if (setvbuf(ostream, NULL, _IOFBF, buffsize())) {
			fprintf(stderr,
				"rdfits: %s: setvbuf() failed for image file %s\n",
				image, ifile);
			fclose(istream);	/* close FITS file */
			if (idel(image) < 0)	/* try to delete header and image file */
				rmi(image);	/* remove image header, if not deleted */
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr, "rdfits: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
#endif
		}
		isize = imagesize(image);	/* image size in bytes */
		narray = (int)(isize/asize);	/* number of full buf arrays */
		rem  = (unsigned)(isize%asize);	/* number of bytes in last array */
		for (k = 0; k < narray; k++) {	/* copy data arrays */
			if (fread(rec, 1, asize, istream) != asize) {	/* DATA */
				printf("Error reading DATA buffer from %s\n", fitsfile);
				fclose(istream);	/* close FITS file */
				fclose(ostream);	/* close image file */
				idel(image);	/* delete image header and image file */
				goto Next_Image;
			}
			if (swap_on)	/* reorder bytes and words */
				ReOrder(rec, mrec, type, scale, zero, Blank);
			if (fwrite(rec, 1, asize, ostream) != asize) {	/* DATA */
				printf("Error writing image file %s for %s\n", ifile, image);
				fclose(istream);	/* close FITS file */
				fclose(ostream);	/* close image file */
				idel(image);	/* delete image header and image file */
				goto Next_Image;
			}
		}
		if (rem > 0) {	/* remaining bytes in final array */
			if (fread(rec, 1, rem, istream) != rem) {	/* DATA array */
				printf("Error reading DATA buffer from %s\n", fitsfile);
				fclose(istream);	/* close FITS file */
				fclose(ostream);	/* close image file */
				idel(image);	/* delete image header and image file */
				goto Next_Image;
			}
			if (swap_on)	/* reorder bytes and words */
				ReOrder(rec, (rem-1)/RSIZE+1, type, scale, zero, Blank);
			if (fwrite(rec, 1, rem, ostream) != rem) {	/* DATA record */
				printf("Error writing image file %s for %s\n", ifile, image);
				fclose(istream);	/* close FITS file */
				fclose(ostream);	/* close image file */
				idel(image);	/* delete image header and image file */
				goto Next_Image;
			}
		}
		fclose(istream);	/* close FITS file */
		fclose(ostream);	/* close image file */
		if (update() == 0)	/* update new disk header */
			fprintf(stderr, "Disk header %s.HDR not created\n", image);
		Next_Image: ;
	}
	free(rec);	/* remember to deallocate big array */
	if (epilog() < 1) {	/* update disk headers; remove image list */
		fprintf(stderr, "rdfits: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
