#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "fft2.h"	/* function declarations for fft2.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	float FltSub = (float)0;	/* substitute for blank input values */
	int n_arg = 0;
	char **arg;
	void *buffer;
	char *image1, *image2;
	float *a;
	int i, x, y, ncols, direction;
	int id1, npix1, nlin1, type1;
	int id2, npix2, nlin2, type2;
	double Blank1;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "rfft: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "b:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'b':	sscanf(optarg, "%f", &FltSub); break;
			case '?':	opt_err++; break;	/* unknown option */
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 2) {
		printf("Usage: rfft -b FltSub image1 image2\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image1 = truncname(arg[0]);	/* image name */
	if (strpbrk(image1, punct) != (char *)0) {
		fprintf(stderr, "rfft: Illegal image name: %s\n", image1);
		exit(EXIT_FAILURE);
	}

	image2 = truncname(arg[1]);	/* image name */
	if (strpbrk(image2, punct) != (char *)0) {
		fprintf(stderr, "rfft: Illegal image name: %s\n", image2);
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "rfft: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	if ((id1 = iopen(image1, 0, &type1, &npix1, &nlin1)) < 0) {
		fprintf(stderr, "rfft: Can't open %s\n", image1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"rfft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type1 < 1 || type1 > 5) {
		fprintf(stderr,
			"rfft: %s has illegal type = %d\n", image1, type1);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"rfft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type1 != 5) {	/* forward Fourier transform */
		if (npix1 != nlin1) {
			fprintf(stderr, "rfft: %s has npix = %d and nlin = %d\n",
				image1, npix1, nlin1);
			iclose(id1);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr,
					"rfft: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
		type2 = 5;	/* complex output image */
		npix2 = npix1/2 + 1;
		nlin2 = nlin1;
		ncols = npix1 + 2;
		direction = 1;
	} else {	/* inverse Fourier transform */
		type2 = 3;	/* float output image */
		ncols = npix1 + npix1;
		npix2 = ncols - 2;
		nlin2 = nlin1;
		direction = -1;
		if (npix2 != nlin2) {
			fprintf(stderr, "rfft: %s has npix = %d and nlin = %d\n",
				image1, npix1, nlin1);
			iclose(id1);
			if (epilog() < 1)	/* update disk headers */
				fprintf(stderr,
					"rfft: epilog: Disk header(s) not updated\n");
			exit(EXIT_FAILURE);
		}
	}

	for (i = 1; i < nlin1; i <<= 1)
		;

	if (i > nlin1) {
		fprintf(stderr,
			"rfft: %s has illegal nlin = %d\n", image1, nlin1);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"rfft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (nlin1 > max_dim()) {
		fprintf(stderr,
			"rfft: %s has too big nlin = %d\n", image1, nlin1);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"rfft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	Blank1 = blank(image1);

	if ((buffer = (void *)calloc(npix1, bytepix(type1))) == (void *)0) {
		fprintf(stderr, "rfft: Buffer allocation failed\n");
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "rfft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (xist(image2)) {
		fprintf(stderr, "rfft: %s: Existing image\n", image2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "rfft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (cpi(image1, image2) < 1) {	/* copy image header */
		fprintf(stderr,
			"rfft: %s.hdr -> %s.hdr copy failed\n", image1, image2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "rfft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (setimagefile(image2, "") == 0) {	/* erase FILE value */
		fprintf(stderr,
			"rfft: Can't erase old FILE value for %s\n", image2);
		rmi(image2);	/* remove image header */
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "rfft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((id2 = icreat(image2, 1, type2, npix2, nlin2)) < 0) {
		fprintf(stderr, "rfft: Can't create %s\n", image2);
		rmi(image2);	/* remove image header */
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "rfft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "rfft: update: Disk header not created\n");

	if ((a = (float *)calloc(nlin1, ncols*sizeof(float))) == (float *)0) {
		fprintf(stderr, "rfft: Array allocation failed\n");
		iclose(id2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"rfft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	for (i = 0, y = 0; y < nlin1; i += ncols, ++y) {
		getline(id1, y, buffer);	/* fetch new row */
		switch (type1) {
			case 1:	/* unsigned char */
				{
					const UCHAR *q = (UCHAR *)buffer;
					for (x = 0; x < npix1; ++x)
						a[i+x] = (float)q[x];
				}
				break;
			case 2:	/* short int */
				{
					const short *q = (short *)buffer;
					const short b = (short)Blank1;
					for (x = 0; x < npix1; ++x)
						a[i+x] = (q[x] == b)? FltSub: (float)q[x];
				}
				break;
			case 3:	/* float */
				{
					const float *q = (float *)buffer;
					const float h = (float) (0.9*Blank1);
					for (x = 0; x < npix1; ++x)
						a[i+x] = (q[x] > h)? FltSub: q[x];
				}
				break;
			case 4:	/* LONG int */
				{
					const LONG *q = (LONG *)buffer;
					const LONG b = (LONG)Blank1;
					for (x = 0; x < npix1; ++x)
						a[i+x] = (q[x] == b)? FltSub: (float)q[x];
				}
				break;
			case 5:	/* complex */
				{
					const float *q = (float *)buffer;
					const float h = (float) (0.9*Blank1);
					for (x = 0; x < ncols; ++x)
						a[i+x] = (q[x] > h)? FltSub: q[x];
				}
				break;
		}
	}

	fft2(ncols, a, nlin1, -1, direction);

	for (i = 0, y = 0; y < nlin2; i += ncols, ++y)
		putline(id2, y, &a[i]);

	free(a);
	iclose(id2);
	free(buffer);
	iclose(id1);

	sethistory(image2, "rfft");

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "rfft: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
