#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int softedge = 0;	/* Soft Gaussian edge? */
	double sdev = 10.0;	/* Standard deviation of edge fall-off */
	double value = 1.0;
	double a = 0.5/(sdev*sdev);
	double theta1, theta2, zmax, emax;
	double xcen, ycen, pa1, pa2, width;
	int type, npix, nlin;
	void *buffer;
	char *image;
	int flag, id, rw = 2;	/* read and write */
	const float Zero = (float)0;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr,
			"setcross: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "s:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 's':	softedge++; sscanf(optarg, "%lf", &sdev); break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 6) {
		printf(
		"Usage: setcross -s std image Xc Yc PA1 PA2 Width [value]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image = truncname(arg[0]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "setcross: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	sscanf(arg[1], "%lf", &xcen);
	sscanf(arg[2], "%lf", &ycen);
	sscanf(arg[3], "%lf", &pa1);
	sscanf(arg[4], "%lf", &pa2);
	sscanf(arg[5], "%lf", &width);
	if (n_arg > 6) sscanf(arg[6], "%lf", &value);

	theta1 = atan(1.0)*pa1/45.0;	/* position angle in radians */
	theta2 = atan(1.0)*pa2/45.0;	/* position angle in radians */

	zmax = 0.5*width;
	emax = zmax;

	if (softedge) {
		emax += 5.0*sdev;
		a = 0.5/(sdev*sdev);
	}

	/* read image names from header directory */
	if ((flag = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "setcross: prolog: Error = %d\n", flag);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	setbufsiz(4096);
	if ((id = iopen(image, rw, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "setcross: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
			"setcross: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 5) {
		fprintf(stderr,
			"setcross: Type = %d is not implemented\n", type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
			"setcross: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((buffer = (void *)calloc(npix, bytepix(type))) == (void *)0) {
		fprintf(stderr, "setcross: allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
			"setcross: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	switch(type) {
		case 1:
			{
				UCHAR *pix = (UCHAR *)buffer;
				UCHAR val = (value < 0.0)? 0: (UCHAR)(value + 0.5);
				int y;
				for (y = 0; y < nlin; ++y) {
					double dy = y - ycen;
					int x;
					getline(id, y, buffer);
					for (x = 0; x < npix; ++x) {
						double dx = x - xcen;
						double theta = atan2(dy, dx);
						double r = sqrt(dx*dx + dy*dy);
						double z1 = r * fabs(sin(theta - theta1));
						double z2 = r * fabs(sin(theta - theta2));
						double z = MIN(z1,z2);
						double d;
						if (z <= zmax)
							pix[x] = val;
						else if (z > zmax && z <= emax) {
							d = z - zmax;
							d = value*exp(-a*(d*d));
							pix[x] = (d < 0.0)? 0: (UCHAR)(d + 0.5);
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 2:
			{
				short *pix = (short *)buffer;
				short val = (short)((value<0)? value-0.5: value+0.5);
				int y;
				for (y = 0; y < nlin; ++y) {
					double dy = y - ycen;
					int x;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						double dx = x - xcen;
						double theta = atan2(dy, dx);
						double r = sqrt(dx*dx + dy*dy);
						double z1 = r * fabs(sin(theta - theta1));
						double z2 = r * fabs(sin(theta - theta2));
						double z = MIN(z1,z2);
						double d;
						if (z <= zmax)
							pix[x] = val;
						else if (z > zmax && z <= emax) {
							d = z - zmax;
							d = value*exp(-a*(d*d));
							pix[x] = (short)((d < 0.0)? d - 0.5: d + 0.5);
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 3:
			{
				float *pix = (float *)buffer;
				float val = (float)value;
				int y;
				for (y = 0; y < nlin; ++y) {
					double dy = y - ycen;
					int x;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						double dx = x - xcen;
						double theta = atan2(dy, dx);
						double r = sqrt(dx*dx + dy*dy);
						double z1 = r * fabs(sin(theta - theta1));
						double z2 = r * fabs(sin(theta - theta2));
						double z = MIN(z1,z2);
						double d;
						if (z <= zmax)
							pix[x] = val;
						else if (z > zmax && z <= emax) {
							d = z - zmax;
							d = value*exp(-a*(d*d));
							pix[x] = (float) d;
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 4:
			{
				LONG *pix = (LONG *)buffer;
				LONG val = (LONG)((value<0)? value-0.5: value+0.5);
				int y;
				for (y = 0; y < nlin; ++y) {
					double dy = y - ycen;
					int x;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						double dx = x - xcen;
						double theta = atan2(dy, dx);
						double r = sqrt(dx*dx + dy*dy);
						double z1 = r * fabs(sin(theta - theta1));
						double z2 = r * fabs(sin(theta - theta2));
						double z = MIN(z1,z2);
						double d;
						if (z <= zmax)
							pix[x] = val;
						else if (z > zmax && z <= emax) {
							d = z - zmax;
							d = value*exp(-a*(d*d));
							pix[x] = (LONG)((d < 0.0)? d - 0.5: d + 0.5);
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 5:
			{
				COMPLEX *pix = (COMPLEX *)buffer;
				float val = (float)value;
				int y;
				for (y = 0; y < nlin; ++y) {
					double dy = y - ycen;
					int x;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						double dx = x - xcen;
						double theta = atan2(dy, dx);
						double r = sqrt(dx*dx + dy*dy);
						double z1 = r * fabs(sin(theta - theta1));
						double z2 = r * fabs(sin(theta - theta2));
						double z = MIN(z1,z2);
						double d;
						if (z <= zmax) {
							pix[x].x = val;
							pix[x].y = Zero;
						} else if (z > zmax && z <= emax) {
							d = z - zmax;
							d = value*exp(-a*(d*d));
							pix[x].x = (float) d;
							pix[x].y = Zero;
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
	}
	free(buffer);
	iclose(id);
	sethistory(image, "setcross");
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "setcross: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
