#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#include <math.h>
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int softedge = 0;	/* Soft Gaussian edge? */
	double sdev = 10.0;	/* Standard deviation of edge fall-off */
	double pa = 0.0;	/* Position angle (in deg) of major axis */
	double eps = 0.0;	/* Ellipticity */
	double value = 1.0;
	double a = 0.5/(sdev*sdev);
	double theta, q, cos_theta, sin_theta, sqrt_q;
	double xcen, ycen, rmin, rmax, emin, emax;
	int type, npix, nlin;
	void *buffer;
	char *image;
	int flag, id, rw = 2;	/* read and write */
	const float Zero = (float)0;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "setring: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "s:p:e:")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 's':	softedge++; sscanf(optarg, "%lf", &sdev); break;
			case 'p':	sscanf(optarg, "%lf", &pa); break;
			case 'e':	sscanf(optarg, "%lf", &eps); break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}

	if (opt_err || n_arg < 5) {
		printf(
		"Usage: setring -s std -p pa -e eps image Xc Yc Rmin Rmax [value]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	theta = atan(1.0)*pa/45.0;	/* position angle in radians */

	if (eps < 0.0 || eps >= 1.0) {
		fprintf(stderr, "setring: Ellipticity = %f is illegal\n", eps);
		exit(EXIT_FAILURE);
	}
	q = 1.0 - eps;	/* ratio of the minor to the major axis */

	cos_theta = cos(theta);
	sin_theta = sin(theta);
	sqrt_q = sqrt(q);

	image = truncname(arg[0]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "setring: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	sscanf(arg[1], "%lf", &xcen);
	sscanf(arg[2], "%lf", &ycen);
	sscanf(arg[3], "%lf", &rmin);
	if (rmin < 0.0) rmin = 0.0;
	sscanf(arg[4], "%lf", &rmax);
	if (rmax < rmin) rmax = rmin;
	if (n_arg > 5) sscanf(arg[5], "%lf", &value);

	emin = rmin;
	emax = rmax;
	if (softedge) {
		emin -= 5.0*sdev;
		if (emin < 0.0) emin = 0.0;
		emax += 5.0*sdev;
		a = 0.5/(sdev*sdev);
	}

	/* read image names from header directory */
	if ((flag = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "setring: prolog: Error = %d\n", flag);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	setbufsiz(4096);
	if ((id = iopen(image, rw, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "setring: Can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "setring: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 5) {
		fprintf(stderr, "setring: Type = %d is not implemented\n", type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "setring: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((buffer = (void *)calloc(npix, bytepix(type))) == (void *)0) {
		fprintf(stderr, "setring: allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"setring: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	switch(type) {
		case 1:
			{
				double rmin2 = rmin * rmin;
				double rmax2 = rmax * rmax;
				double emin2 = emin * emin;
				double emax2 = emax * emax;
				UCHAR *pix = (UCHAR *)buffer;
				UCHAR val = (value < 0.0)? 0: (UCHAR)(value + 0.5);
				int y;
				for (y = 0; y < nlin; ++y) {
					double dy_cos = (y - ycen)*cos_theta;
					double dy_sin = (y - ycen)*sin_theta;
					int x;
					getline(id, y, buffer);
					for (x = 0; x < npix; ++x) {
						double dx = ((x - xcen)*cos_theta + dy_sin)*sqrt_q;
						double dy = (dy_cos - (x - xcen)*sin_theta)/sqrt_q;
						double rad2 = dx*dx + dy*dy;
						double d;
						if (rad2 <= rmax2 && rad2 >= rmin2)
							pix[x] = val;
						else if (rad2 > rmax2 && rad2 <= emax2) {
							d = sqrt(rad2) - rmax;
							d = value*exp(-a*(d*d));
							pix[x] = (d < 0.0)? 0: (UCHAR)(d + 0.5);
						} else if (rad2 < rmin2 && rad2 >= emin2) {
							d = rmin - sqrt(rad2);
							d = value*exp(-a*(d*d));
							pix[x] = (d < 0.0)? 0: (UCHAR)(d + 0.5);
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 2:
			{
				double rmin2 = rmin * rmin;
				double rmax2 = rmax * rmax;
				double emin2 = emin * emin;
				double emax2 = emax * emax;
				short *pix = (short *)buffer;
				short val = (short)((value<0)? value-0.5: value+0.5);
				int y;
				for (y = 0; y < nlin; ++y) {
					double dy_cos = (y - ycen)*cos_theta;
					double dy_sin = (y - ycen)*sin_theta;
					int x;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						double dx = ((x - xcen)*cos_theta + dy_sin)*sqrt_q;
						double dy = (dy_cos - (x - xcen)*sin_theta)/sqrt_q;
						double rad2 = dx*dx + dy*dy;
						double d;
						if (rad2 <= rmax2 && rad2 >= rmin2)
							pix[x] = val;
						else if (rad2 > rmax2 && rad2 <= emax2) {
							d = sqrt(rad2) - rmax;
							d = value*exp(-a*(d*d));
							pix[x] = (short)((d < 0.0)? d - 0.5: d + 0.5);
						} else if (rad2 < rmin2 && rad2 >= emin2) {
							d = rmin - sqrt(rad2);
							d = value*exp(-a*(d*d));
							pix[x] = (short)((d < 0.0)? d - 0.5: d + 0.5);
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 3:
			{
				double rmin2 = rmin * rmin;
				double rmax2 = rmax * rmax;
				double emin2 = emin * emin;
				double emax2 = emax * emax;
				float *pix = (float *)buffer;
				float val = (float)value;
				int y;
				for (y = 0; y < nlin; ++y) {
					double dy_cos = (y - ycen)*cos_theta;
					double dy_sin = (y - ycen)*sin_theta;
					int x;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						double dx = ((x - xcen)*cos_theta + dy_sin)*sqrt_q;
						double dy = (dy_cos - (x - xcen)*sin_theta)/sqrt_q;
						double rad2 = dx*dx + dy*dy;
						double d;
						if (rad2 <= rmax2 && rad2 >= rmin2)
							pix[x] = val;
						else if (rad2 > rmax2 && rad2 <= emax2) {
							d = sqrt(rad2) - rmax;
							d = value*exp(-a*(d*d));
							pix[x] = (float) d;
						} else if (rad2 < rmin2 && rad2 >= emin2) {
							d = rmin - sqrt(rad2);
							d = value*exp(-a*(d*d));
							pix[x] = (float) d;
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 4:
			{
				double rmin2 = rmin * rmin;
				double rmax2 = rmax * rmax;
				double emin2 = emin * emin;
				double emax2 = emax * emax;
				LONG *pix = (LONG *)buffer;
				LONG val = (LONG)((value<0)? value-0.5: value+0.5);
				int y;
				for (y = 0; y < nlin; ++y) {
					double dy_cos = (y - ycen)*cos_theta;
					double dy_sin = (y - ycen)*sin_theta;
					int x;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						double dx = ((x - xcen)*cos_theta + dy_sin)*sqrt_q;
						double dy = (dy_cos - (x - xcen)*sin_theta)/sqrt_q;
						double rad2 = dx*dx + dy*dy;
						double d;
						if (rad2 <= rmax2 && rad2 >= rmin2)
							pix[x] = val;
						else if (rad2 > rmax2 && rad2 <= emax2) {
							d = sqrt(rad2) - rmax;
							d = value*exp(-a*(d*d));
							pix[x] = (LONG)((d < 0.0)? d - 0.5: d + 0.5);
						} else if (rad2 < rmin2 && rad2 >= emin2) {
							d = rmin - sqrt(rad2);
							d = value*exp(-a*(d*d));
							pix[x] = (LONG)((d < 0.0)? d - 0.5: d + 0.5);
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
		case 5:
			{
				double rmin2 = rmin * rmin;
				double rmax2 = rmax * rmax;
				double emin2 = emin * emin;
				double emax2 = emax * emax;
				COMPLEX *pix = (COMPLEX *)buffer;
				float val = (float)value;
				int y;
				for (y = 0; y < nlin; ++y) {
					double dy_cos = (y - ycen)*cos_theta;
					double dy_sin = (y - ycen)*sin_theta;
					int x;
					getline(id, y, buffer);
					for(x = 0; x < npix; ++x) {
						double dx = ((x - xcen)*cos_theta + dy_sin)*sqrt_q;
						double dy = (dy_cos - (x - xcen)*sin_theta)/sqrt_q;
						double rad2 = dx*dx + dy*dy;
						double d;
						if (rad2 <= rmax2 && rad2 >= rmin2) {
							pix[x].x = val;
							pix[x].y = Zero;
						} else if (rad2 > rmax2 && rad2 <= emax2) {
							d = sqrt(rad2) - rmax;
							d = value*exp(-a*(d*d));
							pix[x].x = (float) d;
							pix[x].y = Zero;
						} else if (rad2 < rmin2 && rad2 >= emin2) {
							d = rmin - sqrt(rad2);
							d = value*exp(-a*(d*d));
							pix[x].x = (float) d;
							pix[x].y = Zero;
						}
					}
					putline(id, y, buffer);
				}
			}
			break;
	}
	free(buffer);
	iclose(id);
	sethistory(image, "setring");
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "setring: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	free(arg);
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
