#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	char *image;
	int xmin, xmax, ymin, ymax;
	int stx, sty, extx, exty;
	int type, npix, nlin;
	int x, y, flag, id, rw = 2;	/* read and write */
	double value;
	void *buffer;
	const float Zero = (float)0;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if (argc < 2) {
		printf("Usage: setwin image [value] [Xmin] [Xmax] [Ymin] [Ymax]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}
	image = truncname(argv[1]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "setwin: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}
	/* read image name from header directory */
	if ((flag = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "setwin: prolog: Error = %d\n", flag);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	setbufsiz(4096);
	if ((id = iopen(image, rw, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "setwin: can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "setwin: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if	(argc > 2) sscanf(argv[2], "%lf", &value); else value = 0.0;
	if (argc > 3) sscanf(argv[3], "%d", &xmin); else xmin = 0;
	if (argc > 4) sscanf(argv[4], "%d", &xmax); else xmax = npix - 1;
	if (argc > 5) sscanf(argv[5], "%d", &ymin); else ymin = 0;
	if (argc > 6) sscanf(argv[6], "%d", &ymax); else ymax = nlin - 1;

	stx = xmin;
	sty = ymin;
	extx = xmax - xmin + 1;
	exty = ymax - ymin + 1;
	if (stx < 0) {
		extx += stx;
		stx = 0;
	}
	if (sty < 0) {
		exty += sty;
		sty = 0;
	}
	if (extx <= 0 || exty <= 0) {
		fprintf(stderr, "setwin: Window is placed above/left of image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "setwin: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (stx >= npix || sty >= nlin) {
		fprintf(stderr, "setwin: Window is placed below/right of image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "setwin: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (stx + extx > npix)
		extx = npix - stx;
	if (sty + exty > nlin)
		exty = nlin - sty;
	if (!window(id, stx, sty, extx, exty)) {
		fprintf(stderr, "setwin: Window is not inside image\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "setwin: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type < 1 || type > 5) {
		fprintf(stderr, "setwin: Type = %d is not implemented\n", type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "setwin: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if ((buffer = (void *)calloc(extx, bytepix(type))) == (void *)0) {
		fprintf(stderr, "setwin: allocation of line buffer failed\n");
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr,
				"setwin: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	switch(type) {
		case 1:
			{
				UCHAR *pix = (UCHAR *)buffer;
				const UCHAR pval = (UCHAR )value;
				for(x = 0; x < extx; x++)
					pix[x] = pval;
			}
			break;
		case 2:
			{
				short *pix = (short *)buffer;
				const short pval = (short)value;
				for(x = 0; x < extx; x++)
					pix[x] = pval;
			}
			break;
		case 3:
			{
				float *pix = (float *)buffer;
				const float pval = (float)value;
				for(x = 0; x < extx; x++)
					pix[x] = pval;
			}
			break;
		case 4:
			{
				LONG *pix = (LONG *)buffer;
				const LONG pval = (LONG)value;
				for(x = 0; x < extx; x++)
					pix[x] = pval;
			}
			break;
		case 5:
			{
				COMPLEX *pix = (COMPLEX *)buffer;
				const float pval = (float)value;
				for(x = 0; x < extx; x++) {
					pix[x].x = pval;
					pix[x].y = Zero;
				}
			}
			break;
	}
	for (y = 0; y < exty; y++)
		putline(id, y, buffer);
	free(buffer);
	iclose(id);
	sethistory(image, "setwin");
	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "setwin: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	fflush(stdout);
	exit(EXIT_SUCCESS);
}
