#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <stdlib.h>	/* for ANSI compatibility */
#include <string.h>
#if	defined(SYSV) || defined(BSD)	/* Linux */
#include <unistd.h>
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "bufsiz.h"	/* function declarations for bufsiz.c */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "fbuf.h"	/* constants and prototypes for fbuf.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	char *image, *buffer;
	int id, block, i, j, k;
	int npix, nlin, pag_buf, bsize = buffsize();
	int type, rw = 0;	/* read only */
	int frame = 1;	/* default frame buffer number */
	int page = 0;	/* default window page number */

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if (argc < 2) {
		printf("Usage: wrframe image [frame] [page]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

#if   defined(SYSV) || defined(BSD)	/* Linux */
   if (iopl(3) != 0) {
		fprintf(stderr, "wrframe: Can't get I/O permission\n");
		exit(EXIT_FAILURE);
	}
#endif

	image = truncname(argv[1]);
	if (strpbrk(image, punct) != (char *)0) {
		fprintf(stderr, "wrframe: Illegal image name: %s\n", image);
		exit(EXIT_FAILURE);
	}

	if(argc > 2) {
		sscanf(argv[2], "%d", &frame);	/* frame number */
		if (frame < 1 || frame > 2) {
			fprintf(stderr, "wrframe: Frame number = %d\n", frame);
			exit(EXIT_FAILURE);
		}
	}

	if(argc > 3) {
		sscanf(argv[3], "%d", &page);	/* page number */
		if (page < 0 || page >= RAM_PAG) {
			fprintf(stderr, "wrframe: Page number = %d\n", page);
			exit(EXIT_FAILURE);
		}
	}

	if ((buffer = (char *)malloc(bsize)) == (char *)0) {
		fprintf(stderr, "rdframe: Can't allocate input buffer\n");
		exit(EXIT_FAILURE);
	}

	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "wrframe: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}

	setipath(0);	/* select first token in IMGPATH */

	if ((id = iopen(image, rw, &type, &npix, &nlin)) < 0) {
		fprintf(stderr, "wrframe: can't open %s\n", image);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "wrframe: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (type != 2) {
		fprintf(stderr,
			"wrframe: %s has illegal image type = %d\n", image, type);
		iclose(id);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "wrframe: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	pag_buf = (int)(PAG_SIZ/bsize);	/* number of blocks per page */
	map_fram(pag_buf);	/* map frame buffer RAM */

	block = 0;	/* image block number */
	for (k = page; k < RAM_PAG; k++) {	/* step through pages */
		set_fwin(frame, k);	/* select frame and page */
		for (j = 0; j < pag_buf; j++) {	/* step through blocks */
			if (bread(id, block, buffer) > 0)	/* read block */
				switch (frame) {
					case 1:
						for (i = 0; i < bsize; i++)
							fb1_win[j][i] = buffer[i];
						break;
					case 2:
						for (i = 0; i < bsize; i++)
							fb2_win[j][i] = buffer[i];
						break;
				} else
				goto End;	/* exit after last block */
			block++;
		}
	}
	End:

	close_map();
	iclose(id);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "wrframe: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	free(buffer);

	fflush(stdout);
	exit(EXIT_SUCCESS);
}
