#include "host.h"	/* defines MSC, HC1, HC2, WC, BSD or SYSV */
#include <stdio.h>	/* declarations for stream I/O functions */
#include <string.h>	/* declarations for string functions */
#if	defined(ANSI)	/* ANSI-C */
#include <stdlib.h>
#else	/* not ANSI-C */
#if	defined(BSD)	/* BSD-UNIX */
extern char *calloc();
extern char *malloc();
extern char *realloc();
#endif
#if	defined(SYSV)	/* UNIX System V */
extern void *calloc();
extern void *malloc();
extern void *realloc();
#endif
#endif
/*--------------------------------------------------------------------*/
#include "exitret.h"	/* define EXIT values, if not defined */
#include "nonansi.h"	/* const and void redefined if not ANSI-C */
#include "imhdr.h"	/* function declarations for imhdr.c */
#include "imkw.h"	/* function declarations for imkw.c */
#include "imio.h"	/* function declarations for imio.c */
#include "opxtract.h"	/* function declaration for opxtract.c */
#include "getopt.h"	/* declarations for getopt.c */
#include "args.h"	/* function declaration for args.c */
/*--------------------------------------------------------------------*/
#include "vmalloc.h"	/* function declarations for vmalloc.c */
#include "chkheap.h"	/* redefine malloc functions */
/*--------------------------------------------------------------------*/
char *punct = "!\"#%&\'()*+,-./:;<=>?[\\]^{|}~";
/*--------------------------------------------------------------------*/
#if	defined(ANSI)	/* ANSI-C */
main(int Argc, char **Argv)
#else	/* old style */
main(Argc, Argv)
int Argc;
char **Argv;
#endif
{
	int argc;
	char **argv;
	int option, opt_err = 0;
	int n_arg = 0;
	char **arg;
	int un_signed = 0;
	int i, xmin, xmax, ymin, ymax;
	int stx, sty, extx, exty;
	int id1, type1, npix1, nlin1, rw1 = 0;
	int id2, type2, npix2, nlin2, rw2 = 1;
	double Blank1, Blank2;
	char *image1, *image2, *buf1, *buf2;

#if	defined(ARGS)
	/* substitute some arguments (: or !) from previous command line */
	argv = args(Argc, Argv, &argc);
#else	/* no substitution */
	argc = Argc;
	argv = Argv;
#endif

	if ((arg = (char **)calloc(argc, sizeof (char *))) == (char **)0) {
		fprintf(stderr, "xtract: calloc: Can't allocate argument table\n");
		exit(EXIT_FAILURE);
	}

	while (((option = getopt(argc, argv, "u")) != NONOPT)
			|| (optarg != NULL)) {
		switch(option) {
			case 'u':	un_signed++; break;
			case '?':	opt_err++; break;
			case NONOPT:	arg[n_arg++] = optarg; break;
			default:		break;
		}
	}
	if (opt_err || n_arg < 2) {
		printf(
		"Usage: xtract -u image1 image2 [type] [Xmin] [Xmax] [Ymin] [Ymax]\n");
		fflush(stdout);
		exit(EXIT_FAILURE);
	}

	image1 = truncname(arg[0]);
	if (strpbrk(image1, punct) != (char *)0) {
		fprintf(stderr, "xtract: Illegal image name: %s\n", image1);
		exit(EXIT_FAILURE);
	}
	image2 = truncname(arg[1]);
	if (strpbrk(image2, punct) != (char *)0) {
		fprintf(stderr, "xtract: Illegal image name: %s\n", image2);
		exit(EXIT_FAILURE);
	}
	/* read image names from header directory */
	if ((i = prolog(0)) < 1) {	/* select first token in HDRPATH */
		fprintf(stderr, "xtract: prolog: Error = %d\n", i);
		exit(EXIT_FAILURE);
	}
	setipath(0);	/* select first token in IMGPATH */
	if ((id1 = iopen(image1, rw1, &type1, &npix1, &nlin1)) < 0) {
		fprintf(stderr, "xtract: can't open %s\n", image1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	if (type1 < 1 || type1 > 5) {
		fprintf(stderr,
			"xtract: %s has illegal type = %d\n", image1, type1);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}
	Blank1 = blank(image1);

	if (n_arg > 2) sscanf(arg[2], "%d", &type2); else type2 = type1;
	if (n_arg > 3) sscanf(arg[3], "%d", &xmin); else xmin = 0;
	if (n_arg > 4) sscanf(arg[4], "%d", &xmax); else xmax = npix1 - 1;
	if (n_arg > 5) sscanf(arg[5], "%d", &ymin); else ymin = 0;
	if (n_arg > 6) sscanf(arg[6], "%d", &ymax); else ymax = nlin1 - 1;

	if (type2 < 1 || type2 > 5) {
		fprintf(stderr,
			"xtract: %s has illegal type = %d\n", image2, type2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	stx = xmin;
	sty = ymin;
	extx = xmax - xmin + 1;
	exty = ymax - ymin + 1;

	if (stx < 0) {
		extx += stx;
		stx = 0;
	}

	if (sty < 0) {
		exty += sty;
		sty = 0;
	}

	if (extx <= 0 || exty <= 0) {
		fprintf(stderr, "xtract: window is placed above/left of image\n");
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (stx >= npix1 || sty >= nlin1) {
		fprintf(stderr, "xtract: window is placed below/right of image\n");
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (stx + extx > npix1)
		extx = npix1 - stx;
	if (sty + exty > nlin1)
		exty = nlin1 - sty;

	if (!window(id1, stx, sty, extx, exty)) {
		fprintf(stderr, "xtract: window is not inside image\n");
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((buf1 = (char *)calloc(extx, bytepix(type1))) == (char *)0) {
		fprintf(stderr, "xtract: Allocation of buffer #1 failed\n");
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if ((buf2 = (char *)calloc(extx, bytepix(type2))) == (char *)0) {
		fprintf(stderr, "xtract: Allocation of buffer #2 failed\n");
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (xist(image2)) {
		fprintf(stderr, "xtract: %s: Existing image\n", image2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (cpi(image1, image2) < 1) {	/* copy image header */
		fprintf(stderr,
			"xtract: %s.hdr -> %s.hdr copy failed\n", image1, image2);
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	} else if (setimagefile(image2, "") == 0) {	/* erase FILE value */
		fprintf(stderr,
			"xtract: Can't erase old FILE value for %s\n", image2);
		rmi(image2);	/* remove image header */
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	npix2 = extx;
	nlin2 = exty;

	if((id2 = icreat(image2, rw2, type2, npix2, nlin2)) < 0) {
		fprintf(stderr, "xtract: Can't create %s\n", image2);
		rmi(image2);	/* remove image header */
		iclose(id1);
		if (epilog() < 1)	/* update disk headers */
			fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	if (update() == 0)	/* update disk headers */
		fprintf(stderr, "xtract: update: Disk header not created\n");

	Blank2 = blank(image2);

	if (un_signed > 0 && type1 == 2 & type2 > 2)	/* copy unsigned short */
		type1 = -2;

	for (i = 0; i < exty; i++) {
		getline(id1, i, buf1);
		bufcpy(buf2, buf1, extx, type2, type1, Blank2, Blank1);
		putline(id2, i, buf2);
	}

	iclose(id2);
	free(buf2);
	free(buf1);
	iclose(id1);

	setiname(image2);
	sethistory(image2, "xtract");
	sethistory(image2, image2);

	if (epilog() < 1) {	/* update disk headers */
		fprintf(stderr, "xtract: epilog: Disk header(s) not updated\n");
		exit(EXIT_FAILURE);
	}

	fflush(stdout);
	exit(EXIT_SUCCESS);
}
